/*
 * Console pseudo-filesystem
 * Copyright (c) 2004, David H. Hovemeyer <daveho@cs.umd.edu>
 * $Revision: 1.11 $
 *
 * This is free software.  You are permitted to use,
 * redistribute, and modify it as specified in the file "COPYING".
 */

#include <geekos/errno.h>
#include <geekos/kassert.h>
#include <geekos/vfs.h>
#include <geekos/malloc.h>
#include <geekos/screen.h>
#include <geekos/keyboard.h>
#include <geekos/consfs.h>

/* ----------------------------------------------------------------------
 * Private data and functions
 * ---------------------------------------------------------------------- */
 
//#define DEBUG_CONSFS

/*
 * Read from console (keyboard).
 * Returns number of bytes read, or error code.
 */
static int Console_Read(struct File *file, void *buf, ulong_t numBytes)
{
	#ifdef DEBUG_CONSFS
	Print("Reading from console %ld bytes, mode: %x\n", numBytes, file->mode);
	#endif
	
	//If not allowd to read
    if( !(file->mode & O_READ) )
    {
    	#ifdef DEBUG_CONSFS
    	Print("Reading from console not allowed\n");
    	#endif
    	
    	return EUNSUPPORTED;
    }
	
	//The buffer of keycodes
	char* keybuf = buf;
	
	//Read key for key
	int i;
    for(i=0; i<numBytes; i++)
    {       	
    	//The new keycode
    	Keycode code = 0x0;
    	
    	//Get keycode until the highest bit is not set
    	//This bit is cleared when key is pressed and
    	//set when released 
    	while( (code & KEY_RELEASE_FLAG) == 0)
    		code = Wait_For_Key();
    	
    	//Get the lower code which contains the ASCII code
    	keybuf[i] = code & 0x00ff;    	
    }
    	
    
    return numBytes;
}

/*
 * Write to console (screen).
 * Returns number of bytes written.
 */
static int Console_Write(struct File *file, void *buf, ulong_t numBytes)
{	
	#ifdef DEBUG_CONSFS
	Print("Write to console %ld bytes, mode: %x\n", numBytes, file->mode);
	#endif
	
	//If not allowd to write
    if( !(file->mode & O_WRITE) )
    {
    	#ifdef DEBUG_CONSFS
    	Print("Writing to console not allowed\n");
    	#endif
    	
    	return EUNSUPPORTED;
    }
    
    //Print buffer to screen
    Put_Buf(buf, numBytes);
    
    return numBytes;    	
}

/*
 * Close console input or output file.
 * Returns 0 if successful, error code if not.
 */
static int Console_Close(struct File *file)
{
	//Anything else to do?
    return 0;
}

/*
 * Clone a console File object.
 */
static int Console_Clone(struct File *file, struct File **pClone)
{
    struct File *clone;

    if (file->ops->Write == 0)
	clone = Open_Console_Input();
    else
	clone = Open_Console_Output();

    if (clone == 0)
	return ENOMEM;
    else {
	*pClone = clone;
	return 0;
    }
}

/*
 * File_Ops for console input.
 */
static struct File_Ops s_consInputFileOps = {
    0,			/* FStat() */
    &Console_Read,
    0,			/* Write() */
    0,			/* Seek() */
    &Console_Close,
    0,			/* Read_Entry() */
    &Console_Clone,
};

/*
 * File_Ops for console output.
 */
static struct File_Ops s_consOutputFileOps = {
    0,			/* FStat() */
    0,			/* Read() */
    &Console_Write,
    0,			/* Seek() */
    &Console_Close,
    0,			/* Read_Entry() */
    &Console_Clone,
};

/*
 * Create a File object for console input or output.
 * Returns null pointer if File can't be created.
 */
static struct File *Do_Open(struct File_Ops *ops, int mode)
{
    return Allocate_File(ops, 0, 0, 0, mode, 0);
}

/* ----------------------------------------------------------------------
 * Public functions
 * ---------------------------------------------------------------------- */

/*
 * Open a File which will read from the console (keyboard).
 */
struct File *Open_Console_Input(void)
{
    struct File* file = Do_Open(&s_consInputFileOps, O_READ);
    
    //Set buffer to big number
    if( file )
    	file->endPos = 1<<30;
    
    return file;
}

/*
 * Open a File which will write to the console (screen).
 */
struct File *Open_Console_Output(void)
{
    return Do_Open(&s_consOutputFileOps, O_WRITE);
}

