/*
 * ELF executable loading
 * Copyright (c) 2003, Jeffrey K. Hollingsworth <hollings@cs.umd.edu>
 * Copyright (c) 2003, David H. Hovemeyer <daveho@cs.umd.edu>
 * $Revision: 1.29 $
 * 
 * This is free software.  You are permitted to use,
 * redistribute, and modify it as specified in the file "COPYING".
 */

#include <geekos/errno.h>
#include <geekos/kassert.h>
#include <geekos/ktypes.h>
#include <geekos/screen.h>  /* for debug Print() statements */
#include <geekos/pfat.h>
#include <geekos/malloc.h>
#include <geekos/string.h>
#include <geekos/user.h>
#include <geekos/fileio.h>
#include <geekos/elf.h>


//Types used by the ELF definition pdf
typedef unsigned long Elf32_Addr;
typedef unsigned short Elf32_Half;
typedef unsigned long Elf32_Off;
typedef long Elf32_Sword;
typedef unsigned long Elf32_Word;



#define B(x) ( 0x000000ff & (x))
/** ELF data to host data, half word
 * @a array to get data from
 * @i start index to get data from
 * @data Data entry in ELF header: MSB or LSB */
#define ELFTOHOSTH(a, i, data) ( (data==1) ? (B(a[i]) | B(a[i+1])<<8) : (B(a[i])<<8 | B(a[i+1])) )
/** ELF data to host data, full word
 * @a array to get data from
 * @i start index to get data from
 * @data Data entry in ELF header: MSB or LSB */
#define ELFTOHOSTW(a, i, data) ( (data==1) ? (B(a[i]) | B(a[i+1])<<8 | B(a[i+2])<<16 | B(a[i+3])<<24) : (B(a[i])<<24 | B(a[i+1])<<16 | B(a[i+2])<<8 | B(a[i+3])) )


    
int Parse_ELF_SegmentTable(char *exeFileData, ulong_t exeFileLength, struct Exe_Format *exeFormat,
 Elf32_Half data, Elf32_Off shoff, Elf32_Half shentsize, Elf32_Half shnum);
int Parse_ELF_ProgramTable(char *exeFileData, ulong_t exeFileLength, struct Exe_Format *exeFormat,
 Elf32_Half data, Elf32_Off shoff, Elf32_Half shentsize, Elf32_Half shnum);



/**
 * From the data of an ELF executable, determine how its segments
 * need to be loaded into memory. This function has never been tested
 * with MSB encoding!
 * @param exeFileData buffer containing the executable file
 * @param exeFileLength length of the executable file in bytes
 * @param exeFormat structure describing the executable's segments
 *   and entry address; to be filled in
 * @return 0 if successful, < 0 on error. Error codes are in elf.h
 */
int Parse_ELF_Executable(char *exeFileData, ulong_t exeFileLength,
    struct Exe_Format *exeFormat)
{
	//Return value of subroutines
	int ret;	
	
	//You may want to use the unix tool 'readelf'
	//The ELF file begins with an ELF header. 
	//This header has the following structure:
	//16 bytes containing clas, data, version	
	Elf32_Half Class, data, fileversion;
	//Linear alignment of the following values;
	Elf32_Half type, machine;
	Elf32_Word version;
	Elf32_Addr entry;
	Elf32_Off phoff, shoff;
	Elf32_Word flags;
	Elf32_Half ehsize, phentsize, phnum, shentsize, shnum, shstrndx;	
	
	
	//If we now the values above we know the position of the
	//section table (shoff) and the position of the programm
	//table (phoff). 	
	
	
	//Check the first 4 bytes if they look like the magic bytes.
	if( exeFileData[0] != 0x7f ||
		exeFileData[1] != 'E' ||
		exeFileData[2] != 'L' ||
		exeFileData[3] != 'F')
		return ELFERR_MAGIC;
		
	//File class: 0=Invalid, 1=32bit, 2=64bit
	Class = exeFileData[4];
	if( Class != 1 )
		return ELFERR_CLASS;
		
	//Encoding: 0=Invalid, 1=LSB, 2=MSB
	data = exeFileData[5];
	if( data < 1 || data > 2 )
		return ELFERR_DATA;
		
	//Version: Must be 1, 0 is invalid
	fileversion = exeFileData[6];
	if( fileversion != 1 )
		return ELFERR_FILEVERSION;
		
	//Padding to 16 bytes with zeros
	//After that the ELF ident ends
	//and the remaining bytes of the header
	//come along.
	if( exeFileData[7] != 0 ||
		exeFileData[8] != 0 ||
		exeFileData[9] != 0 ||
		exeFileData[10] != 0 ||
		exeFileData[11] != 0 ||
		exeFileData[12] != 0 ||
		exeFileData[13] != 0 ||
		exeFileData[14] != 0 ||
		exeFileData[15] != 0)
		return ELFERR_PADDING;
		
		
	//Object file type
	//0=Notype, 1=Relocateable file, 2=exectable file,
	//3=shared object, 4=core file, 
	//[0xff00, 0xffff]=processor specific
	//We only support executeables, i think
	type = ELFTOHOSTH(exeFileData, 16, data);
	if( type != 2)
		return ELFERR_TYPE;
		
	//Architecture
	//0=None, 1=M32, 2=SPARC, 3=386,
	//4=68K, 5=88K, 7=860, 8=MIPS,
	//10=MIPS-RS4, 11-16=Reserved
	//We only support IA-32, here 386
	machine = ELFTOHOSTH(exeFileData, 18, data);
	if( machine != 3)
		return ELFERR_MACHINE;
				
	//Object file version
	//1 is the only one now
	version = ELFTOHOSTW(exeFileData, 20, data);
	if( version != 1)
		return ELFERR_VERSION;
		
	//Entry point
	entry = ELFTOHOSTW(exeFileData, 24, data);
	exeFormat->entryAddr = entry;

	
	//Programm header offset
	phoff = ELFTOHOSTW(exeFileData, 28, data);
	//Segment header offset
	shoff = ELFTOHOSTW(exeFileData, 32, data);
	//Machine flags	
	flags = ELFTOHOSTW(exeFileData, 36, data);
	//ELF header size
	ehsize = ELFTOHOSTH(exeFileData, 40, data);
	//Programm header entry size
	phentsize = ELFTOHOSTH(exeFileData, 42, data);
	//Number of programm header entries
	phnum = ELFTOHOSTH(exeFileData, 44, data);
	//Segment header entry size
	shentsize = ELFTOHOSTH(exeFileData, 46, data);
	//Number of segment header entries
	shnum = ELFTOHOSTH(exeFileData, 48, data);
	//Index of the stringtable segment
	shstrndx = ELFTOHOSTH(exeFileData, 50, data);
	
	
	
	
	//Print("== ELF Header == \n");	
	//Print("class: %d\n", Class);
	//Print("data: %d\n", data);
	//Print("fileversion: %d\n", fileversion);
	//Print("type: %d\n", type);
	//Print("machine: %d\n", machine);
	//Print("version: %ld\n", version);
	//Print("entry: 0x%lx\n", entry);
	//Print("phoff: 0x%lx = %ld\n", phoff, phoff);
	//Print("shoff: 0x%lx = %ld\n", shoff, shoff);
	//Print("flags: 0x%lx\n", flags);
	//Print("ehsize: %d\n", ehsize);
	//Print("phentsize: %d\n", phentsize);
	//Print("phnum: %d\n", phnum);
	//Print("shentsize: %d\n", shentsize);
	//Print("shnum: %d\n", shnum);
	//Print("shstrndx: %d\n", shstrndx);
	//Print("\n");
	
		
	//Parse segment table
	ret = Parse_ELF_SegmentTable(exeFileData, exeFileLength, exeFormat,
		data, shoff, shentsize, shnum);		
	//Error occoured
	if(ret < 0)
		return ret;
		
		
	//Parse program table
	ret = Parse_ELF_ProgramTable(exeFileData, exeFileLength, exeFormat,
		data, phoff, phentsize, phnum);
	//Error occoured
	if(ret < 0)
		return ret;
		
		
	//ATTENTION:	We do not make use of the program table which
	//				lies at offset phoff. For now its enough
	//				to get the <=3 segments from the elf file.		
		   
    //Everything went ok
    return 0;
}



/**
 * Parse the segment table (and all sections).
 * 
 * @param exeFileData buffer containing the executable file
 * @param exeFileLength length of the executable file in bytes
 * @param exeFormat structure describing the executable's segments
 *   and entry address; to be filled in
 * @param data Encoding format: MSB=2, LSB=1
 * @param shoff Segment header offset
 * @param shentsize Segment header entry size
 * @param shnum Number of entries
 * @return 0 if successful, < 0 on error. Error codes are in elf.h
 */
int Parse_ELF_SegmentTable(char *exeFileData, ulong_t exeFileLength, struct Exe_Format *exeFormat,
     Elf32_Half data, Elf32_Off shoff, Elf32_Half shentsize, Elf32_Half shnum)
{
	//The section header table begins at index 'shoff' and has 'shnum' many
	//section header entries, each of them 'shentsize' big. Every entry
	//in this array has the following structure:
	Elf32_Word name, type, flags;
	Elf32_Addr addr;
	Elf32_Off offset;
	Elf32_Word size, link, info, addralign, entsize;
	
	//The current entry in table
	Elf32_Off curr;
	
	for(curr=0; curr<shnum; curr++)
	{
		//Base address of this entry
		Elf32_Off base = shoff + curr*shentsize;
		
		//String table index to the name of the section
		name = ELFTOHOSTW(exeFileData, base+0, data);
		//Type and semantic of section
		//0=None, 1=Progbits, 2=Symtable, 3=Strtable,
		//6=dynamic, 8=nobits and many more
		type = ELFTOHOSTW(exeFileData, base+4, data);
		//Some attributes of this section
		flags = ELFTOHOSTW(exeFileData, base+8, data);
		//Offset address of this section in memory
		addr = ELFTOHOSTW(exeFileData, base+12, data);
		//Offset of data in ELF file
		offset = ELFTOHOSTW(exeFileData, base+16, data);
		//Size of section in ELF file and memory
		size = ELFTOHOSTW(exeFileData, base+20, data);
		//Section header table index link (depends on type)
		link = ELFTOHOSTW(exeFileData, base+24, data);
		//Extra info on section (depends on type)
		info = ELFTOHOSTW(exeFileData, base+28, data);
		//Address aligning of section in memory
		addralign = ELFTOHOSTW(exeFileData, base+32, data);
		//If section contains table of fixed sized entries
		//then this value represents the size of each entry.
		//If section does not contain such entries, this val is 0
		entsize = ELFTOHOSTW(exeFileData, base+36, data);
				
		
		//Print("== %ld. Section == \n", curr);
		//Print("name: %ld\n", name);
		//Print("type: %ld\n", type);
		//Print("flags: 0x%lx\n", flags);
		//Print("addr: 0x%lx\n", addr);
		//Print("offset: 0x%lx\n", offset);
		//Print("size: %ld\n", size);
		//Print("links: %ld\n", link);
		//Print("addralign: %ld\n", addralign); 
		//Print("entsize: %ld\n", entsize); 
		//Print("\n");
			
	}	
	
	
	return 0;	
}


/**
 * Parse the program table (and all segments).
 * 
 * @param exeFileData buffer containing the executable file
 * @param exeFileLength length of the executable file in bytes
 * @param exeFormat structure describing the executable's segments
 *   and entry address; to be filled in
 * @param data Encoding format: MSB=2, LSB=1
 * @param phoff Program header offset
 * @param phentsize Program header entry size
 * @param phnum Number of entries
 * @return 0 if successful, < 0 on error. Error codes are in elf.h
 */
int Parse_ELF_ProgramTable(char *exeFileData, ulong_t exeFileLength, struct Exe_Format *exeFormat,
     Elf32_Half data, Elf32_Off phoff, Elf32_Half phentsize, Elf32_Half phnum)
{
	//The program header table begins at index 'phoff' and has 'phnum' many
	//header entries, each of them 'phentsize' big. Every entry
	//in this array has the following structure:
	Elf32_Word type;
	Elf32_Off offset;
	Elf32_Addr vaddr, paddr;
	Elf32_Word filesz, memsz, flags, align;
	
	
	//The current entry in table
	Elf32_Off curr;
	//Currently we have zero segments
	exeFormat->numSegments = 0;		
	
	for(curr=0; curr<phnum; curr++)
	{
		//Base address of this entry
		Elf32_Off base = phoff + curr*phentsize;
		
		//Type of segment
		//0=Null, 1=Load, 2=dynamic, 3=interpreter, ...
		type = ELFTOHOSTW(exeFileData, base+0, data);
		//Offset in ELF file where the data lies at
		offset = ELFTOHOSTW(exeFileData, base+4, data);
		//virtual address where the segments lies at
		vaddr = ELFTOHOSTW(exeFileData, base+8, data);
		//physical address where the segments lies at
		//This is not useful for us...
		paddr = ELFTOHOSTW(exeFileData, base+12, data);
		//Size of segment in ELF file
		filesz = ELFTOHOSTW(exeFileData, base+16, data);
		//Size of segment in memory
		memsz = ELFTOHOSTW(exeFileData, base+20, data);
		//Flags like read, execute and so on
		flags = ELFTOHOSTW(exeFileData, base+24, data);
		//Aligning of bytes in memory
		align = ELFTOHOSTW(exeFileData, base+28, data);
		
		////Print("== %ld. Segment == \n", curr);
		////Print("type: 0x%lx\n", type); 
		////Print("offset: 0x%lx\n", offset); 
		////Print("vaddr: 0x%lx\n", vaddr); 
		////Print("paddr: 0x%lx\n", paddr); 
		////Print("filesz: %ld\n", filesz); 
		////Print("memsz: %ld\n", memsz); 
		////Print("flags: 0x%lx\n", flags); 
		////Print("align: %ld\n", align);
		////Print("\n");		
		
		
		//We found programm bits
		//We only support EXE_MAX_SEGMENTS (=3) many segments, .text and
		//.data should be enough
		if( type == 1 && exeFormat->numSegments < EXE_MAX_SEGMENTS )
		{
			//Set values
			exeFormat->segmentList[ exeFormat->numSegments ].lengthInFile = filesz;
			exeFormat->segmentList[ exeFormat->numSegments ].offsetInFile = offset;
			exeFormat->segmentList[ exeFormat->numSegments ].protFlags = flags;
			exeFormat->segmentList[ exeFormat->numSegments ].sizeInMemory = memsz;
			exeFormat->segmentList[ exeFormat->numSegments ].startAddress = vaddr;			
			//Increment amount of segments
			exeFormat->numSegments++;			 
		}
	}	
	
	
	return 0;	
}



