/*
 * GeekOS file system
 * Copyright (c) 2004, David H. Hovemeyer <daveho@cs.umd.edu>
 * $Revision: 1.54 $
 *
 * This is free software.  You are permitted to use,
 * redistribute, and modify it as specified in the file "COPYING".
 */

#include <limits.h>
#include <geekos/errno.h>
#include <geekos/kassert.h>
#include <geekos/screen.h>
#include <geekos/malloc.h>
#include <geekos/string.h>
#include <geekos/bitset.h>
#include <geekos/synch.h>
#include <geekos/bufcache.h>
#include <geekos/gosfs.h>
#include <geekos/int.h>



//Uncomment this to print debug messages
//#define DEBUG_GOSFS


/* ----------------------------------------------------------------------
 * Private data and functions
 * ---------------------------------------------------------------------- */
 
 
struct GOSFS_Instance;
struct GOSFS_File;
 
/**
 * This struct represents an instance of GOSFS. The VFS takes an object
 * of this type to communicate with the GOSFS implementation. To reach this
 * the mount points fsdata member points to this struct.
 */
struct GOSFS_Instance
{
	/** How many blocks do we have? */
	unsigned cntblocks;
	/** For mutual access to filesystem management structures. */
    struct Mutex lock;
    
    /** A buffer which represents the used-block-bitmap */
    void* usedbitmap;  
    
    /** The corresponding mount point */
    struct Mount_Point *mountPoint;
    
    /** A list of all open files. */
	struct GOSFS_File* openFiles[GOSFS_NUM_MAX_OPEN_FILES];
};


/**
 * This structure represents a open file of GOSFS. If multiple processes
 * are opening a single file, they all point to the same instance of
 * this structure via the File.fsData member.
 */
struct GOSFS_File
{
	/** How many references are there to this file? A file is
	 * "really" closed if ref gets zero. */
	int ref;
		
	/** The only cache which are all processes using */
	struct FS_Buffer_Cache* cache;
	
	/** The only directory entry */
	struct GOSFS_Dir_Entry* entry;
	
	/** The path of the file */
	char* path;
};


/** The magic number of GOSFS */
#define GOSFS_MAGICNUMBER 0x687277a5

/** Calc the amount of bitmap blocks needed for cntblocks-many blocks */
#define GOSFS_BITMAP_BLOCKS( cntblocks ) ( ((cntblocks)/8 + GOSFS_FS_BLOCK_SIZE-1) / GOSFS_FS_BLOCK_SIZE) 


/*
static ulong_t Get_Physical_Blocknr( struct Mount_Point *mountPoint, 
	struct GOSFS_Dir_Entry* file, int blocknr);
	
static int Get_Dir_Entry_Child_Entry( struct Mount_Point *mountPoint, 
	struct GOSFS_Dir_Entry* parent, const char* child, struct GOSFS_Dir_Entry* entry);

static int Find_Dir_Entry( struct Mount_Point *mountPoint, const char* path,
	struct GOSFS_Dir_Entry* entry);
	 
static int Add_Dir_Entry( struct Mount_Point *mountPoint,
	struct GOSFS_Dir_Entry* parent, struct GOSFS_Dir_Entry* entry);
		
static int Find_Free_Dir_Entry_Slot( struct Mount_Point *mountPoint,
	struct GOSFS_Dir_Entry* parent);
	
static int Add_Dir_Entry_Slot( struct Mount_Point *mountPoint,
	struct GOSFS_Dir_Entry* parent);
	
static int Add_Block_To_Dir_Slot( struct Mount_Point *mountPoint,
	struct GOSFS_Dir_Entry* entry);
*/


void Flush_Bitmap( struct GOSFS_Instance* gosfs )
{
	#ifdef DEBUG_GOSFS	
	Print("\t\tWrite down bitmap.\n");
	#endif
	
	//Check if FS is locked
	KASSERT( gosfs->lock.state == MUTEX_LOCKED );
	
	//Create a buffer cache
	struct FS_Buffer_Cache* cache = Create_FS_Buffer_Cache(gosfs->mountPoint->dev, GOSFS_FS_BLOCK_SIZE);
	struct FS_Buffer* buf=0;
	KASSERT( cache != 0 );
	KASSERT( gosfs->mountPoint != 0 );
	
	//Size of bitmap in bytes and number of blocks
	const unsigned bitmapBlocks = GOSFS_BITMAP_BLOCKS( gosfs->cntblocks ); 
			
				
	//Write bitmap to block device
	int i; 
	for(i=0; i<bitmapBlocks; i++)	
	{
		//Get the superblock
		Get_FS_Buffer(cache, i+1, &buf);
		
		//Copy data to block buffer
		memset(buf->data, 0, GOSFS_FS_BLOCK_SIZE);
		memcpy(buf->data, gosfs->usedbitmap + GOSFS_FS_BLOCK_SIZE*i, 
			MIN(GOSFS_FS_BLOCK_SIZE, gosfs->cntblocks/8 - GOSFS_FS_BLOCK_SIZE*i));
			
	
		//Release buffer	
		Modify_FS_Buffer(cache, buf);
		Sync_FS_Buffer(cache, buf);
		Release_FS_Buffer(cache, buf);
	}
	
	
	//Destroy the cache which is not needed any more
	Destroy_FS_Buffer_Cache(cache);
}


/**
 * Add a new block to a directory entry. Used when a we want to write to
 * a file or directory and need more space.
 * @param entry The entry which should get a new block.
 * @return 0 for succ, <0 for an error (like no free mem) 
 * */	
int Add_Block_To_Dir_Entry( struct Mount_Point *mountPoint,
	struct GOSFS_Dir_Entry* entry)
{
	// Stefans hints:
	// Look at the code of Get physical block nr and Get_Dir_Entry_Child_Entry
	// to determine the last block. Then get a new block which is free
	// for this you have to use the mountPoint where fsdata points to our
	// GOSFS implementation and therefore the bitmap.
	// This is the only point where new blocks are allocated for a file/dir.
	
	//This is the amount of blocks we use now
	const ulong_t cntblocks = (entry->size + GOSFS_FS_BLOCK_SIZE-1)/GOSFS_FS_BLOCK_SIZE;
	int ret = 0;
	
	#ifdef DEBUG_GOSFS
	Print("\tAdd a new block to %s which has %lu Bytes / %lu blocks.\n",	
		entry->filename, entry->size, cntblocks);
	#endif
	
	//Get the gosfs implementation
	struct GOSFS_Instance* gosfs = (struct GOSFS_Instance*)mountPoint->fsData;	
	KASSERT( gosfs != 0 );	
	KASSERT( gosfs->lock.state == MUTEX_LOCKED );
	
	
	//File reached maximal length
	if( cntblocks >= GOSFS_NUM_DIRECT_BLOCKS + GOSFS_NUM_PTRS_PER_BLOCK*GOSFS_NUM_INDIRECT_BLOCKS)
	{
		#ifdef DEBUG_GOSFS
		Print("\t\tError: Try to add block to %s but has reached max. size!\n",
			entry->filename);
		#endif
		
		return ENOSPACE;
	}
	
	
	//This is the first free block we see
	int freeblocknr = Find_First_Free_Bit(gosfs->usedbitmap, gosfs->cntblocks);	
	
	
	
	//There is no free block!
	if( freeblocknr < 0)
	{
		#ifdef DEBUG_GOSFS
		Print("\t\tError: Try to add block to %s but there is no space on disc!\n",
			entry->filename);
		#endif
		
		return ENOSPACE;		
	}
	
	
	//Set new block to used
	Set_Bit(gosfs->usedbitmap, freeblocknr);	
	
	
	
	//Add direct block
	if( cntblocks < GOSFS_NUM_DIRECT_BLOCKS )
	{
		//Enter entry
		entry->blockList[cntblocks] = freeblocknr;
	}
	
	//Otherwise we maybe add a indirect block
	else if( cntblocks < GOSFS_NUM_DIRECT_BLOCKS + GOSFS_NUM_PTRS_PER_BLOCK*GOSFS_NUM_INDIRECT_BLOCKS)
	{
		//The index of the block where the indirect-blocknr are saved
		const ulong_t indblock = (cntblocks-GOSFS_NUM_DIRECT_BLOCKS) / GOSFS_NUM_PTRS_PER_BLOCK;
		KASSERT( indblock < GOSFS_NUM_INDIRECT_BLOCKS);
		
		
		//We have to add the indirect block
		if( cntblocks == GOSFS_NUM_DIRECT_BLOCKS )
		{
			//This is the first free block we see
			int freeindblocknr = Find_First_Free_Bit(gosfs->usedbitmap, gosfs->cntblocks);
			
			//No free block for the indirect block
			if( freeindblocknr < 0 )
			{
				#ifdef DEBUG_GOSFS
				Print("\t\tError: Try to add indirect block to %s but there is no space on disc!\n",
					entry->filename);
				#endif
				
				ret = ENOSPACE;		
			}
			
			
			//Set new block to used and enter the block nr
			Set_Bit(gosfs->usedbitmap, freeindblocknr);	
			entry->blockList[indblock+GOSFS_NUM_DIRECT_BLOCKS] = freeindblocknr;
			
			#ifdef DEBUG_GOSFS
			Print("\t\tAdded a indirect block %d to %s.\n", freeindblocknr, entry->filename);
			#endif
		}
		
		//We can go on...
		if( ret == 0)
		{		
			//This is the phy. block number where all indirect blocknr. are saved.
			const ulong_t idxblock = entry->blockList[indblock+GOSFS_NUM_DIRECT_BLOCKS];			
			KASSERT( idxblock != 0);
			
			//Create a buffer cache
			struct FS_Buffer_Cache* cache = Create_FS_Buffer_Cache(mountPoint->dev, GOSFS_FS_BLOCK_SIZE);
			struct FS_Buffer* buf=0;		
			//No memory available
			KASSERT( cache != 0);
			
			//Get the block where the indirect block indices are saved
			Get_FS_Buffer(cache, idxblock, &buf);
			KASSERT( buf!=0 );
			
			//Here are all indices
			ulong_t* idxbuf = (ulong_t*) buf->data;
			
			//Set the free blocknr
			idxbuf[ (cntblocks-GOSFS_NUM_DIRECT_BLOCKS) % GOSFS_NUM_PTRS_PER_BLOCK] = freeblocknr;
			
			//Write block and release it
			Modify_FS_Buffer(cache, buf);
			Sync_FS_Buffer(cache, buf);
			Release_FS_Buffer(cache, buf);	
			Destroy_FS_Buffer_Cache(cache);			
		}
	}
	else
	{
		TODO("Implement inindirect blocks.\n");
		return ENOSPACE;
	}
	
	
	
	//Reset the bit of the block we wanted
	if( ret < 0)
	{
		Clear_Bit(gosfs->usedbitmap, freeblocknr);
	}
	//So here is all fine
	else
	{
		#ifdef DEBUG_GOSFS
		Print("\t\tAdded block %d to %s.\n", freeblocknr, entry->filename);
		#endif
		
		//Flush down the bitmap
		Flush_Bitmap(gosfs);		
	}
		
	return ret;
}



/**
 * Get the physical blocknr of the blocknr-th block of the file given by a 
 * directory entry. This is done from the given mountPoint.
 * @return The physical block number.
 */
ulong_t Get_Physical_Blocknr( struct Mount_Point *mountPoint, struct GOSFS_Dir_Entry* file,
						int blocknr)
{
	#ifdef DEBUG_GOSFS
	Print("\tGet phy. blocknr of %d-th block of \"%s\" of size %ld.\n", blocknr, file->filename, file->size);
	#endif
	
	//If we want to get this block than filesize must be at least this size
	//The equation case is necessary because we want to get the block when we
	//write in it and therefore we get exactly equivalence.
	KASSERT( (file->size + GOSFS_FS_BLOCK_SIZE-1)/GOSFS_FS_BLOCK_SIZE >= blocknr);
			
	//Get the gosfs implementation
	struct GOSFS_Instance* gosfs = (struct GOSFS_Instance*)mountPoint->fsData;	
	KASSERT( gosfs != 0 );	
	KASSERT( gosfs->lock.state == MUTEX_LOCKED );
	
	const unsigned maxblocknr = (file->size + GOSFS_FS_BLOCK_SIZE -1)/GOSFS_FS_BLOCK_SIZE;	
	KASSERT( blocknr <= maxblocknr );
		
	//Block is a direct block
	if( blocknr < GOSFS_NUM_DIRECT_BLOCKS )
	{
		const ulong_t retblocknr = file->blockList[blocknr];
		
		#ifdef DEBUG_GOSFS
		Print("\t\tGot index as direct block: %lu\n", retblocknr);
		#endif
		
		//Check if values are correct
		KASSERT( retblocknr < gosfs->cntblocks);
		KASSERT( Is_Bit_Set( gosfs->usedbitmap, retblocknr) );
		return retblocknr;
	}
	
	
	//Block is a indirect block
	if( blocknr < GOSFS_NUM_DIRECT_BLOCKS 
				+ GOSFS_NUM_INDIRECT_BLOCKS*GOSFS_NUM_PTRS_PER_BLOCK )
	{				
		//The index of the block where the indirect-blocknr are saved
		const ulong_t indblock = (blocknr-GOSFS_NUM_DIRECT_BLOCKS) / GOSFS_NUM_PTRS_PER_BLOCK;
		KASSERT( indblock < GOSFS_NUM_INDIRECT_BLOCKS);
		
		//This is the phy. block number where all indirect blocknr. are saved.
		const ulong_t idxblock = file->blockList[indblock+GOSFS_NUM_DIRECT_BLOCKS];
		
		#ifdef DEBUG_GOSFS
		Print("\t\tUsing indirect block retrieving. Phy. Idxblock: %lu\n", idxblock);
		#endif
		
		KASSERT( idxblock < gosfs->cntblocks);
		KASSERT( Is_Bit_Set( gosfs->usedbitmap, idxblock) );
		
		//Create a buffer cache
		struct FS_Buffer_Cache* cache = Create_FS_Buffer_Cache(mountPoint->dev, GOSFS_FS_BLOCK_SIZE);
		struct FS_Buffer* buf=0;		
		//No memory available
		KASSERT( cache != 0);
		
		//Get the block where the indirect block indices are saved
		Get_FS_Buffer(cache, idxblock, &buf);
		KASSERT( buf!=0 );
		
		//Here are all indices
		ulong_t* idxbuf = (ulong_t*) buf->data;
		
		//The index where the block number lies and the block number we wanted
		const ulong_t idx = (blocknr-GOSFS_NUM_DIRECT_BLOCKS) % GOSFS_NUM_PTRS_PER_BLOCK;
		const ulong_t retblocknr = idxbuf[idx];
		
		Release_FS_Buffer(cache, buf);
		Destroy_FS_Buffer_Cache(cache);
		
		
		#ifdef DEBUG_GOSFS
		Print("\t\tGet index at pos %lu: %lu\n", idx, retblocknr);
		#endif
					
		//Final check if values are correct
		KASSERT( retblocknr < gosfs->cntblocks);	
		KASSERT( Is_Bit_Set( gosfs->usedbitmap, retblocknr) );
		return retblocknr;		
	}
	
	TODO("Implement 2-level indirect block retrieving.\n");		
}



/**
 * Free all blocks which belong to the directory entry
 */
int Free_Blocks_Of_Dir_Entry( struct Mount_Point *mountPoint,
	struct GOSFS_Dir_Entry* entry)
{
	//This is the amount of blocks we use now
	const ulong_t cntblocks = (entry->size + GOSFS_FS_BLOCK_SIZE-1)/GOSFS_FS_BLOCK_SIZE;
	
	//Get the gosfs implementation
	struct GOSFS_Instance* gosfs = (struct GOSFS_Instance*)mountPoint->fsData;	
	KASSERT( gosfs != 0 );	
	KASSERT( gosfs->lock.state == MUTEX_LOCKED );
	
	
	//Mark all blocks as unused
	int currblock = 0;
	for( currblock=0; currblock<cntblocks; currblock++)
	{
		//Get the physical block
		int phyblocknr = Get_Physical_Blocknr(mountPoint, entry, currblock);
		KASSERT( phyblocknr > 1 );
		
		//Mark block as unused
		Clear_Bit( gosfs->usedbitmap, phyblocknr );
	}
	
	
	//Remoe the indirect blocknr block 
	if( cntblocks >= GOSFS_NUM_DIRECT_BLOCKS )
	{	
		//The max. number of indirect blocks
		const ulong_t cntindirblocks = (cntblocks-GOSFS_NUM_DIRECT_BLOCKS)/GOSFS_NUM_PTRS_PER_BLOCK;
			
		//Remove all this blocks
		for(currblock=0; currblock < MIN(cntindirblocks, GOSFS_NUM_INDIRECT_BLOCKS); currblock++)
		{
			//Get the phy. block of indir. block
			int phyblocknr = entry->blockList[currblock + GOSFS_NUM_DIRECT_BLOCKS];
			KASSERT( phyblocknr > 1 );
						
			//Mark block as unused
			Clear_Bit( gosfs->usedbitmap, phyblocknr );
		}
	}
	
	//In-in direct blocks
	if( cntblocks >= GOSFS_NUM_DIRECT_BLOCKS + GOSFS_NUM_PTRS_PER_BLOCK*GOSFS_NUM_INDIRECT_BLOCKS)
	{
		TODO("Implement deleting 2-level indirect blocks.");			
	}
	
		
		
	//Flush down the bitmap
	Flush_Bitmap(gosfs);	
	
	return 0;
}


/** 
 * We want to get the directory entry of a child in a parent directory entry.
 * @param mountPoint The mount point of the filesystem we want to search in
 * @param parent The parent directory we want to search the child in
 * @param child The name of the child
 * @param entry The directory entry where we copy the found directory entry 
 */
int Get_Dir_Entry_Child_Entry( struct Mount_Point *mountPoint, struct GOSFS_Dir_Entry* parent,
	const char* child, struct GOSFS_Dir_Entry** entry,
	struct FS_Buffer_Cache* cache, struct FS_Buffer** buffer )
{			
	//Garuantee that we have place for such many directory entries in block.
	KASSERT( GOSFS_DIRENTRIES_PER_BLOCK * sizeof(struct GOSFS_Dir_Entry*) < GOSFS_FS_BLOCK_SIZE);	
	//Check if parent entry is a directory
	KASSERT( parent->flags & GOSFS_DIRENTRY_ISDIRECTORY );	
	//If this is a directory then the size must be a multiple of blocksize/entryperblock
	KASSERT( parent->size % ( GOSFS_FS_BLOCK_SIZE/GOSFS_DIRENTRIES_PER_BLOCK ) == 0);
	
	//Get the gosfs implementation
	struct GOSFS_Instance* gosfs = (struct GOSFS_Instance*)mountPoint->fsData;	
	KASSERT( gosfs != 0 );	
	KASSERT( gosfs->lock.state == MUTEX_LOCKED );
	
	//The amount of entries in directory
	const ulong_t cntentries = parent->size / ( GOSFS_FS_BLOCK_SIZE/GOSFS_DIRENTRIES_PER_BLOCK );
	
	
	#ifdef DEBUG_GOSFS
	Print("\tSearch for child %s in \"%s\" with %ld entries.\n", child, parent->filename, cntentries);
	#endif		
	
	
	//Run through all entries until we found the entry we search for	
	ulong_t entryidx;
	for( entryidx=0; entryidx < cntentries; entryidx++)
	{
		//Read the new block
		if( entryidx % GOSFS_DIRENTRIES_PER_BLOCK == 0)
		{
			//The number of block we have to read to check entry
			ulong_t virtblock = entryidx / GOSFS_DIRENTRIES_PER_BLOCK;
			ulong_t phyblock = Get_Physical_Blocknr( mountPoint, parent, virtblock);
			
			
			KASSERT( *buffer != 0);
			//Buffer was in use before
			Release_FS_Buffer(cache, *buffer);			
			//Get the block where the indirect block indices are saved in
			Get_FS_Buffer(cache, phyblock, buffer);
			KASSERT( *buffer != 0 );			
		}
		
		//For now the buffer must be in use!
		KASSERT( *buffer != 0  &&  (*buffer)->flags & FS_BUFFER_INUSE );		
		//The index of the entry in the block
		ulong_t idxinblock = entryidx % GOSFS_DIRENTRIES_PER_BLOCK;
		
		//The current entry
		struct GOSFS_Dir_Entry* currentry = (struct GOSFS_Dir_Entry*)((*buffer)->data + 
				idxinblock*GOSFS_FS_BLOCK_SIZE/GOSFS_DIRENTRIES_PER_BLOCK);
		
		#ifdef DEBUG_GOSFS
		Print("\t\tInspect file %s\n", currentry->filename);
		#endif	
			
			
		//Entry is used and names are equal!
		if( currentry->flags & GOSFS_DIRENTRY_USED  &&
			strcmp(currentry->filename, child) == 0)
		{
			#ifdef DEBUG_GOSFS
			Print("\t\tFound file %s\n", child);
			#endif
			
			//Set current entry
			*entry = currentry;			
			break;
		}		
	}
	
		
	//Could not find file
	if (entryidx >= cntentries)	
		return ENOTFOUND;		
		
	//If we return successfuly there must be a valid buffer
	KASSERT( *buffer != 0  &&  (*buffer)->flags & FS_BUFFER_INUSE );
		
	return 0;
}

/** 
 * We traverse the directory tree on the disc and find the corresponding directory entry
 * of the file given by path. The found entry is pointered to the data structure. This
 * structure is hold in the buffer which is also given.
 * @param mountPoint The mount point of the file system
 * @param path The path of the file whose dir-entry we want to retrieve
 * @param entry Target struct where we copy the data we found
 * @return 0 if success, <>0 otherwise 
 */
int Find_Dir_Entry( struct Mount_Point *mountPoint, const char* path, 
	struct GOSFS_Dir_Entry** entry, 
	struct FS_Buffer_Cache* cache, struct FS_Buffer** buffer  )
{
	#ifdef DEBUG_GOSFS
	Print("Find dir entry of \"%s\"\n", path);
	#endif
	
	//Get the gosfs implementation
	struct GOSFS_Instance* gosfs = (struct GOSFS_Instance*)mountPoint->fsData;	
	KASSERT( gosfs != 0 );	
	KASSERT( gosfs->lock.state == MUTEX_LOCKED );
			
	//Length of path
	const int pathlen = strlen(path);	
	//First sign must be a '/'
	KASSERT( pathlen==0 || path[0] == '/' );
	
	
	
	//The startindex of the subdirectory name in path
	int startidx=0;
	//The endidx of the subdirectory name in path. So
	//path[endidx] == '/'
	int endidx=0;
	//The subdir name
	char dirname[GOSFS_FILENAME_MAX+1];
	
				
	//initial entry for entry
	Get_FS_Buffer(cache, 0, buffer);
	KASSERT( (*buffer) != 0 );
			
	struct GOSFS_Superblock* superblock = (*buffer)->data;
	*entry = &(superblock->root);	
	KASSERT( (*entry)->flags & GOSFS_DIRENTRY_ISDIRECTORY );
	KASSERT( superblock->magicnumber == GOSFS_MAGICNUMBER );
		
	//So path is "/" --> already finished :-)
	if( pathlen <= 1 )
	{
		#ifdef DEBUG_GOSFS
		Print("\tReturn root dir.\n");
		#endif
			
		return 0;
	}
		
	
	//Split up path and traverse tree
	while( true )
	{		
		//loop invariant: at endidx is the slash
		KASSERT( path[endidx] == '/' );
		
		//The initial values for the next subdir name
		startidx = endidx+1;
		endidx = startidx+1;
			
		//So there have been a '/' at the end --> this is not a path
		//to a file
		if( endidx > pathlen )
		{
			#ifdef DEBUG_GOSFS
			Print("\tError: Path ending with '/'c!\n");
			#endif
			
			return EINVALIDFS;
		}
		
		
		//Search for the next '/'
		while( path[endidx] != '/' && endidx<pathlen)
			endidx++;
			
			
		#ifdef DEBUG_GOSFS
		Print("\tLoop finding with startidx %d, endidx %d, substr: %s\n", startidx, endidx, &path[startidx]);
		#endif	
			
			
		//If this would be 1, then subdirs name is ""
		if( endidx - startidx < 1)
		{
			#ifdef DEBUG_GOSFS
			Print("\tError: Empty subdir name!\n");
			#endif
			
			return EINVALIDFS;
		}
			
		//Check size of dir name
		if( endidx - startidx - 1 >= GOSFS_FILENAME_MAX )
		{
			#ifdef DEBUG_GOSFS
			Print("\tError: To long subdir name!\n");
			#endif
			
			return ENAMETOOLONG;
		}
		
		//Copy directory name
		memcpy(dirname, &path[startidx], endidx-startidx);
		dirname[endidx-startidx] = '\0';

		//Go on to subdirectory
		int ret = Get_Dir_Entry_Child_Entry( mountPoint, *entry, dirname, entry,
					 cache, buffer);

		//Error at subroutine
		if( ret != 0)
			return ret;
		
		//We have finished for now!
		if( endidx >= pathlen )
		{
			#ifdef DEBUG_GOSFS
			Print("\tFound file.\n");
			#endif
			
			KASSERT( *buffer != 0 );
			return 0;
		}
			
		//Does not work for files, only for directories:
		//The last character is '/' --> finished too
		//if( endidx == pathlen-1 && path[endidx]=='/')
		//	return 0;			
	}
			
	return ENOTFOUND;
}


/**
 * Extend a file so that its at least size-big
 */
int Extend_File(struct Mount_Point *mountPoint,
	struct GOSFS_Dir_Entry* entry, 	int size)
{
	//This is the amount of blocks we use now
	const ulong_t oldblockcnt = (entry->size + GOSFS_FS_BLOCK_SIZE-1)/GOSFS_FS_BLOCK_SIZE;
	//This is the amount of blocks we must have
	const ulong_t newblockcnt = (size + GOSFS_FS_BLOCK_SIZE-1)/GOSFS_FS_BLOCK_SIZE;
	
	
	//Easy case: finished...
	if( oldblockcnt >= newblockcnt )
	{
		entry->size = MAX(size, entry->size);
		return 0;
	}
	
	//Otherwise we add blocks until we reach the right size
	
	//Current block amount
	ulong_t currbl = oldblockcnt;
	int ret=0;
	
	//Add blocks until we have reached the size
	while( currbl < newblockcnt  &&  ret==0 )
	{
		//Add a block to direntry
		ret = Add_Block_To_Dir_Entry(mountPoint, entry);
		
		//Worked good...
		if( ret == 0)
		{			
			//Goto next block
			currbl++;
			
			//Current size is set to a value in the block.
			//If this is not done and we abort while we not get ready
			//we do not know that these blocks also belong to the file
			entry->size = (currbl-1)*GOSFS_FS_BLOCK_SIZE + size%GOSFS_FS_BLOCK_SIZE;			
		}		
	}
	
	//So we get at least this size
	KASSERT( entry->size == size  || ret!=0 );
		
	return ret;
}


/**
 * Get the slot number of the entrynmb-th used directory entry of parent.
 * @return -1 if not existant, the slot otherwise (beginning with 0)
 */
int Get_Dir_Entry_Slot( struct Mount_Point *mountPoint,
	struct GOSFS_Dir_Entry* parent, ulong_t entrynmb)
{	
	//So we assume the size of the directory is always a multiple of a block
	KASSERT( parent->size % GOSFS_FS_BLOCK_SIZE == 0 );	
	//Check if parent entry is a directory
	KASSERT( parent->flags & GOSFS_DIRENTRY_ISDIRECTORY );
	
	//Get the gosfs implementation
	struct GOSFS_Instance* gosfs = (struct GOSFS_Instance*)mountPoint->fsData;	
	KASSERT( gosfs != 0 );	
	KASSERT( gosfs->lock.state == MUTEX_LOCKED );
	
	//This is the amount of blocks we use now
	const ulong_t cntblocks = (parent->size + GOSFS_FS_BLOCK_SIZE-1)/GOSFS_FS_BLOCK_SIZE;
	
	#ifdef DEBUG_GOSFS
	Print("\tFind a free slot under %ld possible.\n", cntblocks);	 
	#endif
	
	//Create a buffer cache
	struct FS_Buffer_Cache* cache = Create_FS_Buffer_Cache(mountPoint->dev, GOSFS_FS_BLOCK_SIZE);
	struct FS_Buffer* buf=0;		
	//No memory available
	KASSERT( cache != 0);
	
	//The current slot
	int currnmbofslots=-1;
	//This is our target	
	int physlot = -1;
	
	//Run through all blocks until we have the entrynmb-th entry
	ulong_t currblock;	
	for( currblock=0; currblock<cntblocks && physlot<0; currblock++)
	{
		//The phy blocknr we want to open
		const ulong_t phyblocknr = Get_Physical_Blocknr(mountPoint, parent, currblock);
				
		//Get the block where the indirect block indices are saved
		Get_FS_Buffer(cache, phyblocknr, &buf);
		KASSERT( buf!=0 );
				
		//Scan all entries in block
		int idx;
		for(idx=0; idx<GOSFS_DIRENTRIES_PER_BLOCK && physlot<0; idx++)
		{
			#ifdef DEBUG_GOSFS
			Print("\tCheck slot %d in block %ld\n", idx, currblock);	 
			#endif
			
			//The entry we want to check
			struct GOSFS_Dir_Entry* entry = (buf->data + idx*GOSFS_FS_BLOCK_SIZE/GOSFS_DIRENTRIES_PER_BLOCK);
						
			//Found a used entry
			if( entry->flags & GOSFS_DIRENTRY_USED )
				currnmbofslots++;
			
			//Found it!	
			if( currnmbofslots == entrynmb)
				physlot = currnmbofslots;	
		}
		
			
		//Release buffer		
		Release_FS_Buffer(cache, buf);		
	}
	
	//Destroy caches
	Destroy_FS_Buffer_Cache(cache);	
	
	return physlot;
}


/**
 *  Find a free directory slot (therefore the GOSFS_DIRENTRY_USED is
 * not used).
 * @param parent The parent directory entry where the slot should be found in.
 * @return <0 if there is no such slot. Otherwise the index of the slot. */
int Find_Free_Dir_Entry_Slot( struct Mount_Point *mountPoint,
	struct GOSFS_Dir_Entry* parent)
{
	// Stefans hints:
	// just like Get_Dir_Entry_Child_Entry run through all blocks and
	// find a unused slot.
	
	//So we assume the size of the directory is always a multiple of a block
	KASSERT( parent->size % GOSFS_FS_BLOCK_SIZE == 0 );	
	//Check if parent entry is a directory
	KASSERT( parent->flags & GOSFS_DIRENTRY_ISDIRECTORY );
	
	//Get the gosfs implementation
	struct GOSFS_Instance* gosfs = (struct GOSFS_Instance*)mountPoint->fsData;	
	KASSERT( gosfs != 0 );	
	KASSERT( gosfs->lock.state == MUTEX_LOCKED );
	
	//This is the amount of blocks we use now
	const ulong_t cntblocks = (parent->size + GOSFS_FS_BLOCK_SIZE-1)/GOSFS_FS_BLOCK_SIZE;
	int ret = -1;
	
	#ifdef DEBUG_GOSFS
	Print("\tFind a free slot under %ld possible.\n", cntblocks);	 
	#endif
	
	//Create a buffer cache
	struct FS_Buffer_Cache* cache = Create_FS_Buffer_Cache(mountPoint->dev, GOSFS_FS_BLOCK_SIZE);
	struct FS_Buffer* buf=0;		
	//No memory available
	KASSERT( cache != 0);
	
	
	//Run through all blocks and search for unused directory entries
	ulong_t currblock;	
	for( currblock=0; currblock<cntblocks && ret<0; currblock++)
	{
		//The phy blocknr we want to open
		const ulong_t phyblocknr = Get_Physical_Blocknr(mountPoint, parent, currblock);
				
		//Get the block where the indirect block indices are saved
		Get_FS_Buffer(cache, phyblocknr, &buf);
		KASSERT( buf!=0 );
				
		//Scan all entries in block
		int idx;
		for(idx=0; idx<GOSFS_DIRENTRIES_PER_BLOCK && ret<0; idx++)
		{
			#ifdef DEBUG_GOSFS
			Print("\tCheck slot %d in block %ld\n", idx, currblock);	 
			#endif
			
			//The entry we want to check
			struct GOSFS_Dir_Entry* entry = (buf->data + idx*GOSFS_FS_BLOCK_SIZE/GOSFS_DIRENTRIES_PER_BLOCK);
						
			//Found our entry
			if( (entry->flags & GOSFS_DIRENTRY_USED) == 0 )
			{			
				#ifdef DEBUG_GOSFS
				Print("\tSlot is unused.\n");	 
				#endif					
				ret = currblock*GOSFS_DIRENTRIES_PER_BLOCK + idx;
			}
		}
		
			
		//Release buffer		
		Release_FS_Buffer(cache, buf);		
	}
	
	//Destroy caches
	Destroy_FS_Buffer_Cache(cache);	
	
	return ret;
}
	
/**
 * Add a new directory entry slot. A new block is added.
 * @param parent The parent directory entry for which the slot should added.
 * @return <0 for an error (like no free mem). 0 for succ.
 */
int Add_Dir_Entry_Slot( struct Mount_Point *mountPoint,
	struct GOSFS_Dir_Entry* parent)
{
	// Stefans hints:
	// Add a block and return the index of the slot
	
	//So we assume the size of the directory is always a multiple of a block
	KASSERT( parent->size % GOSFS_FS_BLOCK_SIZE == 0 );	
	//Check if parent entry is a directory
	KASSERT( parent->flags & GOSFS_DIRENTRY_ISDIRECTORY );
	
	//Get the gosfs implementation
	struct GOSFS_Instance* gosfs = (struct GOSFS_Instance*)mountPoint->fsData;	
	KASSERT( gosfs != 0 );	
	KASSERT( gosfs->lock.state == MUTEX_LOCKED );
	
	//This is the amount of blocks we use now
	const ulong_t cntblocks = (parent->size + GOSFS_FS_BLOCK_SIZE-1)/GOSFS_FS_BLOCK_SIZE;
	int ret;
	
	
	#ifdef DEBUG_GOSFS
	Print("\tAdd a new block to dir \"%s\" which has %lu Bytes / %lu blocks.\n",	
		parent->filename, parent->size, cntblocks);
	#endif
	
		
	//Add a block to directory
	ret = Add_Block_To_Dir_Entry( mountPoint, parent );
	
	if( ret!=0 )
	{
		#ifdef DEBUG_GOSFS
		Print("\t\t Error when adding block to dir.\n");
		#endif
		
		return ret;
	}
	
	
	//So increment the size of the directory
	parent->size += GOSFS_FS_BLOCK_SIZE;
	
	//Get the phy block which we have added before
	const ulong_t phyblocknr = Get_Physical_Blocknr(mountPoint, parent, cntblocks);
	KASSERT( phyblocknr != 0 );
		
	
	//Create a buffer cache
	struct FS_Buffer_Cache* cache = Create_FS_Buffer_Cache(mountPoint->dev, GOSFS_FS_BLOCK_SIZE);
	struct FS_Buffer* buf=0;		
	//No memory available
	KASSERT( cache != 0);
	
	//Get the block where the indirect block indices are saved
	Get_FS_Buffer(cache, phyblocknr, &buf);
	KASSERT( buf!=0 );
	
	//Reset content of file
	memset(buf->data, 0, GOSFS_FS_BLOCK_SIZE);
		
	//Write block and release it
	Modify_FS_Buffer(cache, buf);
	Sync_FS_Buffer(cache, buf);
	Release_FS_Buffer(cache, buf);			
	Destroy_FS_Buffer_Cache(cache);
	
			
	return ret;
}


/** Add the directory entry 'entry' to the directory which is given
 * by the direntry 'parentdir' */
int Add_Dir_Entry( struct Mount_Point *mountPoint,
	struct GOSFS_Dir_Entry* parent, struct GOSFS_Dir_Entry* entry)
{	
	// Stefans hints:
	// Find a free slot. If there is no free slot add a new one.
	// Afterwards copy the data.
	// Do not forget to set the used flag in slot!
	
	
	#ifdef DEBUG_GOSFS
	Print("Add dir entry %s into \"%s\".\n", entry->filename, parent->filename);
	#endif
	
	//Get the gosfs implementation
	struct GOSFS_Instance* gosfs = (struct GOSFS_Instance*)mountPoint->fsData;	
	KASSERT( gosfs != 0 );	
	KASSERT( gosfs->lock.state == MUTEX_LOCKED );
	
	
	//Check if parent entry is a directory
	KASSERT( parent->flags & GOSFS_DIRENTRY_ISDIRECTORY );
		
	
	//Get a free slot in parent directory
	int freeslot = Find_Free_Dir_Entry_Slot( mountPoint, parent );
	
	//If there is no free slot, add some
	if( freeslot < 0 )
	{
		//Try to add them
		int ret = Add_Dir_Entry_Slot(mountPoint, parent);
		
		//Error while adding a slot
		if( ret < 0 )
		{
			#ifdef DEBUG_GOSFS
			Print("\tError while adding dir slot.\n");
			#endif
			
			return ret;
		}
		
		//Now find again a free slot
		freeslot = Find_Free_Dir_Entry_Slot( mountPoint, parent );
	
		#ifdef DEBUG_GOSFS
		Print("\tNew slot is at %d.\n", freeslot);
		#endif		
	}
	
	//For now there must be a freeslot
	KASSERT( freeslot >= 0 );
	
	//Some position informations
	const ulong_t virtblocknr = freeslot / GOSFS_DIRENTRIES_PER_BLOCK;
	const ulong_t idxinblock = freeslot % GOSFS_DIRENTRIES_PER_BLOCK;
	const ulong_t phyblocknr = Get_Physical_Blocknr(mountPoint, parent, virtblocknr);
	
	KASSERT( phyblocknr > 0);
	
	
	//Create a buffer cache
	struct FS_Buffer_Cache* cache = Create_FS_Buffer_Cache(mountPoint->dev, GOSFS_FS_BLOCK_SIZE);
	struct FS_Buffer* buf=0;		
	//No memory available
	KASSERT( cache != 0);
	
	//Get the block where the indirect block indices are saved
	Get_FS_Buffer(cache, phyblocknr, &buf);
	KASSERT( buf!=0 );
	
	#ifdef DEBUG_GOSFS
	Print("\tAdd entry at phy-block %ld, slot-in-block %ld\n", phyblocknr, idxinblock);
	#endif	
	
	//Reset content of file
	memcpy(buf->data + idxinblock*GOSFS_FS_BLOCK_SIZE/GOSFS_DIRENTRIES_PER_BLOCK,
		 entry,
		 sizeof(struct GOSFS_Dir_Entry) );
		 
		
	//Write block and release it
	Modify_FS_Buffer(cache, buf);
	Sync_FS_Buffer(cache, buf);
	Release_FS_Buffer(cache, buf);			
	Destroy_FS_Buffer_Cache(cache);
			
	return 0;
}


/**
 * Add a file given by 'path' to the FS.
 */
int Add_File_To_FS( struct Mount_Point* mountPoint, const char* path)
{
	#ifdef DEBUG_GOSFS
	Print("Add the file %s to FS.\n", path);
	#endif
	
	//Get the gosfs implementation
	struct GOSFS_Instance* gosfs = (struct GOSFS_Instance*)mountPoint->fsData;	
	KASSERT( gosfs != 0 );	
	KASSERT( gosfs->lock.state == MUTEX_LOCKED );
	
	//Error code
	int ret = 0;
	
	//The parent directory
	char parentdir[GOSFS_FILENAME_MAX];
	strcpy(parentdir, path);
	
	//This is not a valid path
	if( parentdir[ strlen(parentdir)-1 ] == '/' )
	{
		#ifdef DEBUG_GOSFS
		Print("Invalid path %s: ends with '/'.\n", path);
		#endif
	
		return EINVALIDFS;
	}
	
		
	//Cut of file name
	int parentlen = strlen(parentdir)-1;
	while( parentdir[parentlen]!='/' && parentlen>=0 )
	{
		parentdir[parentlen] = '\0';
		parentlen--;
	}	
	
	//Set last terminating string
	parentdir[parentlen] = '\0';	
	KASSERT( parentlen < strlen(path) );
	
	//There is no '/'
	if( parentlen < 0)
	{
		#ifdef DEBUG_GOSFS
		Print("Error: In %s is no '/'!\n", path);
		#endif
					
		return EINVALIDFS;
	}
	
	
	#ifdef DEBUG_GOSFS
	Print("\tParent directory: %s, filename: %s.\n", parentdir, &path[parentlen+1]);
	#endif
	
	
	//Create a buffer cache
	struct FS_Buffer_Cache* cache = Create_FS_Buffer_Cache(mountPoint->dev, GOSFS_FS_BLOCK_SIZE);
	struct FS_Buffer* buffer=0;		
	//No memory available
	KASSERT( cache != 0);	
	
	
	//Try to find parent directory entry
	struct GOSFS_Dir_Entry* parentdirentry = 0;
	ret = Find_Dir_Entry(mountPoint, parentdir, &parentdirentry,
			cache, &buffer);				
			
	
	//Error occoured when trying to find parent directory
	if( ret != 0 )
	{
		#ifdef DEBUG_GOSFS
		Print("Error: Parent directory %s does not exist.\n", parentdir);
		#endif
		
		//Release buffer
		if( buffer != 0  &&  buffer->flags & FS_BUFFER_INUSE)
			Release_FS_Buffer(cache, buffer);
	}
	//Create file
	else
	{			
		//Check that parentdir really exists
		KASSERT( parentdirentry!=0  &&  parentdirentry->flags & GOSFS_DIRENTRY_ISDIRECTORY);
		
		//Get the corresponding directory entry
		struct GOSFS_Dir_Entry newentry;
		
		//Fill out entry structure
		memset( &newentry, 0, sizeof(struct GOSFS_Dir_Entry) );
		strcpy( newentry.filename, &path[parentlen+1]);
		newentry.flags |= GOSFS_DIRENTRY_USED;
					
		//Add the entry
		ret = Add_Dir_Entry(mountPoint, parentdirentry, &newentry);
					
		
		//Error when adding new entry
		if( ret != 0 )
		{
			#ifdef DEBUG_GOSFS
			Print("Error: Cannnot add directory entry.\n");
			#endif							
		}
		
		else
		{	
			KASSERT( buffer != 0 );
											
			//Because the parentdirentry structure may have been
			//changed we have to write it down again
			Modify_FS_Buffer(cache, buffer);
			Sync_FS_Buffer(cache, buffer);
		}			

		//Release, no matter if add dir entry failed
		Release_FS_Buffer(cache, buffer);			
	}		
	
		
	//Destroy buffer
	Destroy_FS_Buffer_Cache(cache);
		
	return ret;
} 

/** Flush the file down to disc */
void Flush_GOSFS_File(struct Mount_Point* mountPoint, struct GOSFS_File *file)
{		
	//Get the gosfs implementation
	struct GOSFS_Instance* gosfs = (struct GOSFS_Instance*)mountPoint->fsData;	
	KASSERT( gosfs != 0 );
	KASSERT( gosfs->lock.state == MUTEX_LOCKED );
	
	//Write down buffers of file itself
	Sync_FS_Buffer_Cache(file->cache);
	
	//Create a buffer cache
	struct FS_Buffer_Cache* cache = Create_FS_Buffer_Cache(mountPoint->dev, GOSFS_FS_BLOCK_SIZE);
	struct FS_Buffer* buffer=0;		
	//No memory available
	KASSERT( cache != 0);
	
	//Get the corresponding directory entry
	struct GOSFS_Dir_Entry* entry;
	int ret = Find_Dir_Entry(mountPoint, file->path, &entry, cache, &buffer);
	KASSERT( ret == 0 );
	
	
	//Copy the entry
	memcpy(entry, file->entry, sizeof(struct GOSFS_Dir_Entry));
	
	//Write down actual directory entry
	Modify_FS_Buffer(cache, buffer);
	Sync_FS_Buffer(cache, buffer);
	Release_FS_Buffer(cache, buffer);
	Destroy_FS_Buffer_Cache(cache);
}


/* ----------------------------------------------------------------------
 * Implementation of VFS operations
 * ---------------------------------------------------------------------- */

/*
 * Get metadata for given file.
 */
static int GOSFS_FStat(struct File *file, struct VFS_File_Stat *stat)
{	
	//Get corresponding GOSFS file
	struct GOSFS_File* gosfile = file->fsData;	
	KASSERT( gosfile != 0 );
	
	#ifdef DEBUG_GOSFS
	Print("Called stat for open file/dir %s., flags: %lx\n", gosfile->path, gosfile->entry->flags);
	#endif
		
	//Set members of stat
	stat->isDirectory = (gosfile->entry->flags & GOSFS_DIRENTRY_ISDIRECTORY)!=0;
	stat->isSetuid = (gosfile->entry->flags & GOSFS_DIRENTRY_SETUID)!=0;
	stat->size = gosfile->entry->size;
	
	//Clear acls
	memset( stat->acls, 0, VFS_MAX_ACL_ENTRIES * sizeof(struct VFS_ACL_Entry));
	stat->acls[0].permission = O_READ | O_WRITE | O_EXCL | O_CREATE;
	stat->acls[0].uid = 0;
	
	#ifdef DEBUG_GOSFS
	Print("\t Ready with stat, size: %d.\n", stat->size);
	#endif	
		
	return 0;	
}

/*
 * Read data from current position in file.
 */
static int GOSFS_Read(struct File *file, void *buf, ulong_t numBytes)
{
    //If nothing is to write, exit
	if( numBytes == 0)
		return 0;
		
		
	//Get corresponding GOSFS file
	struct GOSFS_File* gosfile = file->fsData;	
	KASSERT( gosfile != 0 );
		
	//The mount point
	struct Mount_Point* mountPoint = file->mountPoint;
		
	//Get the gosfs implementation
	struct GOSFS_Instance* gosfs = (struct GOSFS_Instance*)mountPoint->fsData;	
	KASSERT( gosfs != 0 );
	
	
	//The starting block where we want to write in
	const ulong_t startpos = file->filePos;
	const ulong_t startblock = (startpos) / GOSFS_FS_BLOCK_SIZE;
	//The last block in which we write
	const ulong_t lastblock = (startpos + numBytes -1) / GOSFS_FS_BLOCK_SIZE;
	
		
	
	#ifdef DEBUG_GOSFS
	Print("Reading from file from virtual block %ld to block %ld \n", startblock, lastblock);
	Print("Starting at %ld and reading %ld bytes. \n", startpos, numBytes);
	#endif	
	
	//Hopefully this is not a directory!
	KASSERT( (gosfile->entry->flags & GOSFS_DIRENTRY_ISDIRECTORY) == 0 );
	
	
	//Want to read more than possible
	if( startpos + numBytes > gosfile->entry->size )
	{
		#ifdef DEBUG_GOSFS
		Print("\tError: Reading more data than file has.\n");
		#endif	
		
		return EINVALID;
	}
	
	//Not allowed to read
	if( (file->mode & O_READ) == 0  )
	{
		#ifdef DEBUG_GOSFS
		Print("\tError: Reading from not-readable file. Mode %x\n", file->mode );
		#endif	
		
		return EINVALID;
	}
	
	
	
	//Running variables
	ulong_t currpos = startpos;
	
	//The buffer which we use to write
	struct FS_Buffer* buffer = 0;
	
	//Until we do not have written all data - go on
	while( true )
	{									
		//We use FS management structures
		Mutex_Lock( &gosfs->lock );
			//Current virtual position
			const ulong_t currblock = currpos / GOSFS_FS_BLOCK_SIZE;				
			//The phy block in which we write around		
			const ulong_t phyblocknr = Get_Physical_Blocknr(mountPoint, gosfile->entry, currblock);
			const ulong_t posinbuf = currpos - startpos;
		//We use FS management structures
		Mutex_Unlock( &gosfs->lock );			
		
				
		#ifdef DEBUG_GOSFS
		Print("\tReading from block %ld, phy %ld; currpos %ld, posinbuf: %ld\n", 
			currblock, phyblocknr, currpos, posinbuf);
		#endif
		
		KASSERT( phyblocknr > 1 );	
		
		//Get the physical buffer
		int ret = Get_FS_Buffer(gosfile->cache, phyblocknr, &buffer);		
		if( ret != 0)
		{
			#ifdef DEBUG_GOSFS
			Print("\tError while getting buffer, %d\n", ret);
			#endif
			break;
		}		
		KASSERT(buffer != 0 );
		
		
		//ATTENTION: NEARLY THE SAME CODE WHICH IS HERE IS THE SAME
		//IN THE WRITE SECTION. SO IF THERE IS SOMETHING WRONG
		//IT MIGHT BE GOOD TO LOOK IN WRITE AS WELL.
		
		//We are in last block
		if( currblock == lastblock)
		{			
			//Starblock is endblock
			if( currblock == startblock )	
			{
				#ifdef DEBUG_GOSFS
				Print("\tRead from %ld, %ld bytes at first=last block\n", currpos%GOSFS_FS_BLOCK_SIZE, numBytes);
				#endif
									
				//Copy data. of last block
				memcpy( buf + posinbuf, 
					buffer->data + currpos%GOSFS_FS_BLOCK_SIZE,					
					numBytes );
			}
			//Not last block
			else
			{		
				#ifdef DEBUG_GOSFS
				Print("\tRead from %ld, %ld bytes at last!=first block\n", currpos%GOSFS_FS_BLOCK_SIZE,
					(startpos+numBytes)%GOSFS_FS_BLOCK_SIZE);
				#endif
								
				//Copy data. of last block
				memcpy( buf + posinbuf,
					buffer->data + currpos%GOSFS_FS_BLOCK_SIZE, 
					(startpos+numBytes)%GOSFS_FS_BLOCK_SIZE );
			}
							
								
			//Set current position
			currpos = (startpos+numBytes);				
		}
		else
		{
			#ifdef DEBUG_GOSFS
			Print("\tRead from %ld, %ld bytes to !=last block\n", currpos%GOSFS_FS_BLOCK_SIZE,
				GOSFS_FS_BLOCK_SIZE - currpos%GOSFS_FS_BLOCK_SIZE);
			#endif
			
			//Copy data up to full block
			memcpy( buf + posinbuf,
				buffer->data + currpos%GOSFS_FS_BLOCK_SIZE, 
				GOSFS_FS_BLOCK_SIZE - currpos%GOSFS_FS_BLOCK_SIZE);
							
				
			//Wrote data up to end of block
			currpos = (currblock+1)*GOSFS_FS_BLOCK_SIZE;
		}		
		
		
		//Nothing to do anymore
		KASSERT( buffer->flags & FS_BUFFER_INUSE );		
		Release_FS_Buffer(gosfile->cache, buffer);		
		
		
		//We have finished
		if( currblock == lastblock)
			break;
	}
	
	
	//Warning, have not the right currpos - no space?
	if( startpos + numBytes != currpos )
	{
		#ifdef DEBUG_GOSFS
		Print("\tWARNING: startpos %ld + numBytes %ld != currpos %ld\n",
			startpos, numBytes, currpos);
		#endif
	}
	 
	
	 	
	#ifdef DEBUG_GOSFS
		Print("\tReady with reading. size: %ld, currpos: %ld, read: %ld\n\n",
			gosfile->entry->size, file->filePos, currpos-startpos);
	#endif
	
	
	//Return the number of written bytes
	return currpos - startpos;
}

/*
 * Write data to current position in file.
 */
static int GOSFS_Write(struct File *file, void *buf, ulong_t numBytes)
{
	//If nothing is to write, exit
	if( numBytes == 0)
		return 0;
		
	//Get corresponding GOSFS file
	struct GOSFS_File* gosfile = file->fsData;	
	KASSERT( gosfile != 0 );
		
	//The mount point
	struct Mount_Point* mountPoint = file->mountPoint;
		
	//Get the gosfs implementation
	struct GOSFS_Instance* gosfs = (struct GOSFS_Instance*)mountPoint->fsData;	
	KASSERT( gosfs != 0 );
	
	
	//The starting block where we want to write in
	const ulong_t startpos = file->filePos;
	const ulong_t startblock = (startpos) / GOSFS_FS_BLOCK_SIZE;
	//The last block in which we write
	const ulong_t lastblock = (startpos + numBytes -1) / GOSFS_FS_BLOCK_SIZE;
		
	//The old size of the file and the esteminated new one
	//This size has to be reset if an error occoured
	const ulong_t oldsize = gosfile->entry->size;
	
	
	#ifdef DEBUG_GOSFS
	Print("Writing to file from virtual block %ld to block %ld \n", startblock, lastblock);
	Print("Starting at %ld and writting %ld bytes. \n", startpos, numBytes);
	#endif	
	
	//Hopefully this is not a directory!
	KASSERT( (gosfile->entry->flags & GOSFS_DIRENTRY_ISDIRECTORY) == 0 );
	
	
	
	//Not allowed to read
	if( (file->mode & O_WRITE) == 0  )
	{
		#ifdef DEBUG_GOSFS
		Print("\tError: Writing to not-writable file. Mode %x\n", file->mode);
		#endif	
		
		return EINVALID;
	}
	
	
	//Running variables
	ulong_t currpos = startpos;
	
	//The buffer which we use to write
	struct FS_Buffer* buffer = 0;
	
	//Until we do not have written all data - go on
	while( true ) 
	{					
		//Current virtual position
		const ulong_t currblock = currpos / GOSFS_FS_BLOCK_SIZE;
				
		
		//We have to add a block. This is only done if the currpos is as big
		//as the old size and we are at the start of a block.
		//If we add a block, we are at the start of a block
		if( currpos >= oldsize  &&  currpos % GOSFS_FS_BLOCK_SIZE == 0 )
		{			
			//At least one byte more such that add block registers
			//the pos of the new block and later get phy block
			//has the correct asserts
			gosfile->entry->size = currpos;
			
			
			//We use FS management structures
			Mutex_Lock( &gosfs->lock );			
				//Add the new block
				int ret =  Add_Block_To_Dir_Entry(mountPoint, gosfile->entry);
				//If we are interrupted later, than we remember that this
				//block has already be assigned	
				gosfile->entry->size++;		
			//We use FS management structures
			Mutex_Unlock( &gosfs->lock );
						
			if(ret != 0)
			{
				#ifdef DEBUG_GOSFS
				Print("\tError while addid block %ld: %d\n", currblock, ret);
				#endif
				
				break;
			}
		}
		
		
		//We use FS management structures
		Mutex_Lock( &gosfs->lock );				
			//The phy block in which we write around		
			const ulong_t phyblocknr = Get_Physical_Blocknr(mountPoint, gosfile->entry, currblock);
			const ulong_t posinbuf = currpos - startpos;
		//We use FS management structures
		Mutex_Unlock( &gosfs->lock );			
		
				
		#ifdef DEBUG_GOSFS
		Print("\tWriting to block %ld, phy %ld; currpos %ld, posinbuf: %ld\n", 
			currblock, phyblocknr, currpos, posinbuf);
		#endif
		
		KASSERT( phyblocknr > 1 );	
		
		//Get the physical buffer
		int ret = Get_FS_Buffer(gosfile->cache, phyblocknr, &buffer);		
		if( ret != 0)
		{
			#ifdef DEBUG_GOSFS
			Print("\tError while getting buffer, %d\n", ret);
			#endif
			break;
		}		
		KASSERT(buffer != 0 );
		
		
		//ATTENTION: NEARLY THE SAME CODE WHICH IS HERE IS THE SAME
		//IN THE READ SECTION. SO IF THERE IS SOMETHING WRONG
		//IT MIGHT BE GOOD TO LOOK IN READ AS WELL.
				
		
		//We are in last block
		if( currblock == lastblock)
		{			
			//Starblock is endblock
			if( currblock == startblock )	
			{
				#ifdef DEBUG_GOSFS
				Print("\tWrite from %ld, %ld bytes at first=last block\n", currpos%GOSFS_FS_BLOCK_SIZE, numBytes);
				#endif
									
				//Copy data. of last block
				memcpy( buffer->data + currpos%GOSFS_FS_BLOCK_SIZE, 
					buf + posinbuf,
					numBytes );
			}
			//Not last block
			else
			{		
				#ifdef DEBUG_GOSFS
				Print("\tWrite from %ld, %ld bytes at last!=first block\n", currpos%GOSFS_FS_BLOCK_SIZE,
					(startpos+numBytes)%GOSFS_FS_BLOCK_SIZE);
				#endif
								
				//Copy data. of last block
				memcpy( buffer->data + currpos%GOSFS_FS_BLOCK_SIZE, 
					buf + posinbuf,
					(startpos+numBytes)%GOSFS_FS_BLOCK_SIZE );
			}
							
								
			//Set current position
			currpos = (startpos+numBytes);				
		}
		else
		{
			#ifdef DEBUG_GOSFS
			Print("\tWrite from %ld, %ld bytes to !=last block\n", currpos%GOSFS_FS_BLOCK_SIZE,
				GOSFS_FS_BLOCK_SIZE - currpos%GOSFS_FS_BLOCK_SIZE);
			#endif
			
			//Copy data up to full block
			memcpy( buffer->data + currpos%GOSFS_FS_BLOCK_SIZE, 
				buf + posinbuf,
				GOSFS_FS_BLOCK_SIZE - currpos%GOSFS_FS_BLOCK_SIZE);
				
			
				
			//Wrote data up to end of block
			currpos = (currblock+1)*GOSFS_FS_BLOCK_SIZE;
		}		
		
		
		KASSERT( buffer->flags & FS_BUFFER_INUSE );		
		//And write down file
		Modify_FS_Buffer(gosfile->cache, buffer);
		Sync_FS_Buffer(gosfile->cache, buffer);
		Release_FS_Buffer(gosfile->cache, buffer);		
		
		
		//We have finished
		if( currblock == lastblock)
			break;
	}
	
	
	//Warning, have not the right currpos - no space?
	if( startpos + numBytes != currpos )
	{
		#ifdef DEBUG_GOSFS
		Print("\tWARNING: startpos %ld + numBytes %ld != currpos %ld\n",
			startpos, numBytes, currpos);
		#endif
	}
	 
	
	 
	//Set effictive written bytes
	gosfile->entry->size = MAX(oldsize, currpos);	
	
	
	#ifdef DEBUG_GOSFS
		Print("\tReady with writting. size: %ld, currpos: %ld, wrote: %ld\n\n",
			gosfile->entry->size, file->filePos, currpos-startpos);
	#endif
	
	
	//Return the number of written bytes
	return currpos - startpos;
}

/*
 * Seek to a position in file.
 */
static int GOSFS_Seek(struct File *file, ulong_t pos)
{
    //Get corresponding GOSFS file
	struct GOSFS_File* gosfile = file->fsData;	
	KASSERT( gosfile != 0 );
	
	//The mount point
	struct Mount_Point* mountPoint = file->mountPoint;
	
	//Get the gosfs implementation
	struct GOSFS_Instance* gosfs = (struct GOSFS_Instance*)mountPoint->fsData;	
	KASSERT( gosfs != 0 );
	
	#ifdef GOSFS_DEBUG
	Print("Called seeking to pos %ld, flags: %lx.\n", pos, gosfile->entry->flags);
	#endif
	
	int ret=0;

	//We need FS management structures
	Mutex_Lock( &gosfs->lock );
	

	//Invalid positions...
	//Here the case of a directory
	if( gosfile->entry->flags & GOSFS_DIRENTRY_ISDIRECTORY )
	{				
		//Get the directory slot
		int physlot = Get_Dir_Entry_Slot(mountPoint, gosfile->entry, pos);	
				
		//Could not find the pos-th directory entry
		if( physlot < 0)
		{
			#ifdef GOSFS_DEBUG
			Print("Error when seeking in dir: %ld-th dir entry does not exist.\n", pos);
			#endif
			
			ret = EINVALID;
		}
	}
	//And the file
	else
	{	
		//If we do have writting perm. then extend file to this position
		if( file->mode & O_WRITE )
		{
			//Extend file
			ret = Extend_File( mountPoint, gosfile->entry, pos);
			//So if it went good, finish
			KASSERT( gosfile->entry->size >= pos  ||  ret!=0 );
		}
		
		//If position is beyond the file size -- quit
		else if( gosfile->entry->size < pos )
		{
			#ifdef GOSFS_DEBUG
			Print("Error when seeking in file: Invalid position %ld, filesize: %ld.\n", pos, gosfile->entry->size);
			#endif
			
			ret  = EINVALID;
		}
	}
		
		
	//Otherwise just set position	
	if( ret == 0)
	{
		file->filePos = pos;
	}
	
	
	Mutex_Unlock( &gosfs->lock );
		
		
	return ret;	
}

/*
 * Close a file.
 */
static int GOSFS_Close(struct File *file)
{
	//Get corresponding GOSFS file
	struct GOSFS_File* gosfile = file->fsData;	
	KASSERT( gosfile != 0 );
	
	
	//The mount point
	struct Mount_Point* mountPoint = file->mountPoint;
	
	//Get the gosfs implementation
	struct GOSFS_Instance* gosfs = (struct GOSFS_Instance*)mountPoint->fsData;	
	KASSERT( gosfs != 0 );
	
	//Use FS management structures
	Mutex_Lock( &gosfs->lock );
	
	#ifdef DEBUG_GOSFS
	Print("Close GOSFS file, ref-cnt before close: %d\n", gosfile->ref);
	#endif
	
	//Decrement reference
	gosfile->ref--;
	KASSERT( gosfile->ref >= 0);	
	
	//If nobody is refering now, delete gosfsfile and flush buffers
	if( gosfile->ref == 0)
	{	
		#ifdef DEBUG_GOSFS
		Print("Ref count to 0 -- flush file.\n");
		#endif
		
		//Flush down the file
		Flush_GOSFS_File( mountPoint, gosfile );
		Destroy_FS_Buffer_Cache(gosfile->cache);
		
		//Delete entry in open file table 
		int idx;
		for(idx=0; idx<GOSFS_NUM_MAX_OPEN_FILES; idx++)
		{
			//Found entry and reset it
			if( gosfs->openFiles[idx] == gosfile)
				gosfs->openFiles[idx] = 0;
		}
		
		//Free the structure
		Free( gosfile ); 
	}	
	
	
	//Use FS management structures
	Mutex_Unlock( &gosfs->lock );
		
	return 0;	
}

/*
 * Clone a file.
 * The clone will access the same underlying file as the
 * original, but read/write/seek operations, etc. will be distinct.
 * The filepos values are preserved.
 * Returns 0 if successful, or an error code if unsuccessful.
 */
static int GOSFS_Clone(struct File *file, struct File **pClone)
{
	//Create mem for file
	*pClone = Malloc( sizeof(struct File) );
	
	//No mem available
	if( *pClone == 0)
		return ENOMEM;
	
	//copy memory
	memcpy(*pClone, file, sizeof(struct File) );
	
	//Get corresponding GOSFS file
	struct GOSFS_File* gosfile = file->fsData;	
	KASSERT( gosfile != 0 );
	
	
	#ifdef DEBUG_GOSFS
	Print("Clone GOSFS file, before clone refcnt: %d\n", gosfile->ref);
	#endif
	
	KASSERT( gosfile->ref > 0);
	//Increment reference counter
	gosfile->ref++;
		
	return 0;
}

/*static*/ struct File_Ops s_gosfsFileOps = {
    &GOSFS_FStat,
    &GOSFS_Read,
    &GOSFS_Write,
    &GOSFS_Seek,
    &GOSFS_Close,
    0, /* Read_Entry */
    &GOSFS_Clone,
};

/*
 * Stat operation for an already open directory.
 */
static int GOSFS_FStat_Directory(struct File *dir, struct VFS_File_Stat *stat)
{
	//FIXME: Anything special to do here??
	//Maybe check if dir is really a directory. But then GOSFS_FStat
	//should check if realy should handle a file
    return GOSFS_FStat(dir, stat);
}

/*
 * Directory Close operation.
 */
static int GOSFS_Close_Directory(struct File *dir)
{
	//FIXME: Any differences??
	//Maybe check if dir is really a directory. But then GOSFS_Close
	//should check if realy should handle a file
	return GOSFS_Close(dir);
}

/*
 * Read a directory entry from an open directory.
 */
static int GOSFS_Read_Entry(struct File *dir, struct VFS_Dir_Entry *entry)
{
	//Get mount point
    struct Mount_Point *mountPoint = dir->mountPoint;
	
	//Get the gosfs implementation
	struct GOSFS_Instance* gosfs = (struct GOSFS_Instance*)mountPoint->fsData;	
	KASSERT( gosfs != 0 );
	
	//We use FS management structures - thus we need mutex...
	Mutex_Lock( &gosfs->lock );
	//Pointer to the special data for gosfs
	struct GOSFS_File *gosfile = (struct GOSFS_File*)dir->fsData;
	
	int ret = EUNSPECIFIED;

	//Find the physical slot according to filePos
	int slot = Get_Dir_Entry_Slot(mountPoint, gosfile->entry, dir->filePos);
	if ( slot == -1 )
	{
		//-1 means that the slot to filePos was not found...
		ret = VFS_NO_MORE_DIR_ENTRIES;
	}
	else
	{
		//Get the buffer cache
		struct FS_Buffer_Cache* cache = gosfile->cache;
		struct FS_Buffer* buffer=0;
		//No memory available
		KASSERT( cache != 0);
	
		ulong_t blocknmb    = slot/GOSFS_DIRENTRIES_PER_BLOCK;
		ulong_t slotinblock = slot%GOSFS_DIRENTRIES_PER_BLOCK;
		
		const ulong_t phyblocknr = Get_Physical_Blocknr(mountPoint, gosfile->entry, blocknmb);
		
		//Get the block where the indirect block indices are saved
		Get_FS_Buffer(cache, phyblocknr, &buffer);
		KASSERT( buffer != 0 );

		//The entry we want to check
		struct GOSFS_Dir_Entry* currdirentry = (buffer->data + slotinblock*GOSFS_FS_BLOCK_SIZE/GOSFS_DIRENTRIES_PER_BLOCK);
					
		//Found an used entry
		if( (currdirentry->flags & GOSFS_DIRENTRY_USED) == 1 )
		{			
			ret = 0; //Everything seems ok

			//Copy the filename
			strcpy(entry->name, currdirentry->filename);
			
			//Copy stats
			entry->stats.size = currdirentry->size;
		    entry->stats.isDirectory = 0;
		    if ( currdirentry->flags & GOSFS_DIRENTRY_ISDIRECTORY )
		    {
		    	entry->stats.isDirectory = 1;
		    }
		    
		    entry->stats.isSetuid = 0;
		    if ( currdirentry->flags & GOSFS_DIRENTRY_SETUID )
		    {
		    	entry->stats.isSetuid = 1;
		    }
		}

		//Release buffer
		if ( buffer->flags & FS_BUFFER_INUSE )
			Release_FS_Buffer(cache, buffer);
	}

	Mutex_Unlock( &gosfs->lock );

    return ret;
}

/*static*/ struct File_Ops s_gosfsDirOps = {
    &GOSFS_FStat_Directory,
    0, /* Read */
    0, /* Write */
    &GOSFS_Seek, /* Seek */
    &GOSFS_Close_Directory,
    &GOSFS_Read_Entry,
};

/*
 * Open a file named by given path.
 */
int GOSFS_Open(struct Mount_Point *mountPoint, const char *path, int mode, struct File **pFile)
{		
	#ifdef DEBUG_GOSFS
	Print("Called open for file %s.\n", path);
	#endif
	
	//Get the gosfs implementation
	struct GOSFS_Instance* gosfs = (struct GOSFS_Instance*)mountPoint->fsData;	
	KASSERT( gosfs != 0 );
	
	
	//We use FS management structures
	Mutex_Lock( &gosfs->lock );
	
	
	//Search for file if it is already open
	int idx;
	for(idx=0; idx<GOSFS_NUM_MAX_OPEN_FILES; idx++)
	{
		//Found file as already open
		if( gosfs->openFiles[idx] != 0 &&
			strcmp( gosfs->openFiles[idx]->path, path) == 0 &&
			(gosfs->openFiles[idx]->entry->flags & GOSFS_DIRENTRY_ISDIRECTORY) == 0)
		{
			#ifdef DEBUG_GOSFS
			Print("Easy opening: File was already open at slot %d.\n", idx);
			#endif
			
			//Allocate new file object with this gos-file			
			*pFile = Allocate_File( &s_gosfsFileOps,
					0, gosfs->openFiles[idx]->entry->size,
					gosfs->openFiles[idx],
					mode, mountPoint);
					
			//Increment reference
			gosfs->openFiles[idx]->ref++;
			
			
			#ifdef DEBUG_GOSFS
			Print("Now ref-coutner is %d.\n", gosfs->openFiles[idx]->ref);
			#endif
			
			//So there must be at least two
			KASSERT( gosfs->openFiles[idx]->ref >= 2);
			
			//We use FS management structures
			Mutex_Unlock( &gosfs->lock );								
			//Finished
			return 0;
		}
	}
	
	
	//Otherwise search for an empty slot in open files table
	for(idx=0; idx<GOSFS_NUM_MAX_OPEN_FILES; idx++)
	{
		//Found file as already open
		if( gosfs->openFiles[idx] == 0)
			break;
	}
	
	//No free slot
	if( idx >= GOSFS_NUM_MAX_OPEN_FILES)
	{
		#ifdef DEBUG_GOSFS
		Print("Error: GOSFS can not open more files.\n");
		#endif
			
		//We use FS management structures
		Mutex_Unlock( &gosfs->lock );	
		return EMFILE;
	}		
		
	//Otherwise save the gosfs file
	struct GOSFS_File** gosfile = & (gosfs->openFiles[idx]);
	
		
	//Create a buffer cache
	struct FS_Buffer_Cache* cache = Create_FS_Buffer_Cache(mountPoint->dev, GOSFS_FS_BLOCK_SIZE);
	struct FS_Buffer* buffer=0;		
	//No memory available
	KASSERT( cache != 0);
	
	//Get the corresponding directory entry
	struct GOSFS_Dir_Entry* entry;
	int ret = Find_Dir_Entry(mountPoint, path, &entry, cache, &buffer);
	
	//File exists but is a directory
	if( ret == 0  &&  entry->flags & GOSFS_DIRENTRY_ISDIRECTORY )
	{
		ret = EINVALID;
	}	
	
	//File does not exist and we have to create it
	else if( ret == ENOTFOUND  &&  (mode & O_CREATE) )
	{
		//Release buffer
		if( buffer != 0 )
			Release_FS_Buffer(cache, buffer);		
		//Destroy old cache			
		Destroy_FS_Buffer_Cache(cache);
			
		
		//Add the file	
		ret = Add_File_To_FS( mountPoint, path);
		
		//Recreate new cache. We have to this procedure because AddFile
		//has changed things on file system
		cache = Create_FS_Buffer_Cache(mountPoint->dev, GOSFS_FS_BLOCK_SIZE);
				
		if( ret == 0 )
		{									
			//And search again...
			ret = Find_Dir_Entry(mountPoint, path, &entry, cache, &buffer);			
			//But for now we must find the entry
			KASSERT( ret == 0 );
		}		
	}
		
	
	//Successfully created
	if( ret == 0 )
	{
		KASSERT(entry != 0);
		
		//The file is a directory!
		if( entry->flags & GOSFS_DIRENTRY_ISDIRECTORY )
		{
			ret =  EUNSUPPORTED;
		}
		else
		{						
			//Create the file structure
			struct File* file = Malloc( sizeof(struct File) );
			memset(file, 0, sizeof(struct File));
			*pFile = file;
		
			//Fill out file structure	
			file->ops  = &s_gosfsFileOps;
			file->filePos = 0;
			file->endPos = entry->size;
			file->mode = mode;
			file->mountPoint = mountPoint;
			
			//We can add some data at fsData member if we need
			//this for managing files, especially if several
			//processes open the same files.
			//We save the filename here
			*gosfile = (struct GOSFS_File*) Malloc( sizeof(struct GOSFS_File));
			file->fsData =  (struct GOS_File*) *gosfile;
			
			//Create GOSFS file
			(*gosfile)->cache = Create_FS_Buffer_Cache(mountPoint->dev, GOSFS_FS_BLOCK_SIZE);
			(*gosfile)->entry = Malloc( sizeof(struct GOSFS_Dir_Entry));
			memcpy( (*gosfile)->entry, entry, sizeof(struct GOSFS_Dir_Entry));
			(*gosfile)->path = Malloc( strlen(path) );
			strcpy((*gosfile)->path, path);
			(*gosfile)->ref = 1;						
		}
	}
	
	
	//Release block again
	if( buffer != 0  &&  buffer->flags & FS_BUFFER_INUSE )	
		Release_FS_Buffer(cache, buffer);			
	Destroy_FS_Buffer_Cache(cache);	
	
	
	//We use FS management structures
	Mutex_Unlock( &gosfs->lock );
	//Return the return-value
	return ret;
}

/*
 * Create a directory named by given path.
 * get the last directory of the path
 * lookup the whole path (excl. the last [new] directory) in our directory tree
 * if does not exist this means that multiple directories would be created
 * otherwise append this new directory to the directory tree if not already in the directory
 */
static int GOSFS_Create_Directory(struct Mount_Point *mountPoint, const char *path)
{
	if ( strlen(path) > VFS_MAX_PATH_LEN )
	{
		Print("gosfs create_dir: path [%s] too long!\n", path);
		return ENAMETOOLONG;
	}

	//Create_Directory of the vfs implementation should already take care of this...
	KASSERT ( path[0]=='/' ); 
	
	//Find the last slash
	int ridx = strlen(path)-1;
	while ( ridx>0 && path[ridx]!='/' )
	{
		ridx--;
	}
	
	//We really should have found a slash!
	KASSERT ( path[ridx] == '/' );
	
	//The "parent" is everything till (incl.) the last (or first) slash!
	//i.e. path=/a/foo -> parent=/a/
	//     path=/a     -> parent=/
	//FIXME: should we Malloc here? VFS_MAX_PATH_LEN might be some waste...
	char parentdir[VFS_MAX_PATH_LEN+1];	
	memset(parentdir, 0, sizeof(char) * (VFS_MAX_PATH_LEN+1));

	//Copy everything till the last slash:
	int idx;
	for ( idx=0; idx<=ridx; idx++ )
	{
		parentdir[idx] = path[idx];
	}
	
	#ifdef DEBUG_GOSFS
	Print("gosfs create_dir: resolved parent directory: [%s]\n", parentdir);
	#endif
	
	//Resolve the new directories name
	//Therefore we just use the pointer of path to ridx+1 (+1 because without last slash)
	char *newdir = (char*)(path+ridx+1);
	
	//Make sure the new directory name doesn't exceed the max filename
	if ( strlen(newdir) > GOSFS_FILENAME_MAX )
	{
		Print("gosfs create_dir: directory name [%s] too long!\n", newdir);
		return ENAMETOOLONG;
	}
	
	#ifdef DEBUG_GOSFS
	Print("gosfs create_dir: resolved new directory: [%s]\n", newdir);
	#endif
	
	//Create a buffer cache
	struct FS_Buffer_Cache* cache = Create_FS_Buffer_Cache(mountPoint->dev, GOSFS_FS_BLOCK_SIZE);
	struct FS_Buffer* buffer=0;
	//No memory available
	KASSERT( cache != 0);	

	//Get the gosfs implementation
	struct GOSFS_Instance* gosfs = (struct GOSFS_Instance*)mountPoint->fsData;	
	KASSERT( gosfs != 0 );
	
	//We use FS management structures - thus we need mutex...
	Mutex_Lock( &gosfs->lock );
	
	//Error code
	int ret = 0;
	
	#ifdef DEBUG_GOSFS
	Print("trying to find directory %s in tree.\n", path);
	#endif

	//Find_Dir_Entry disallows a slash at the end of parendir!
	//so remove it

	//Make sure the directory does not exist already!
	struct GOSFS_Dir_Entry* direntry = 0;
	ret = Find_Dir_Entry(mountPoint, path, &direntry,
			cache, &buffer);
	if ( ret == 0 )
	{
		//Release the buffer if acquired and used
		KASSERT( buffer != 0  &&  buffer->flags & FS_BUFFER_INUSE );

		Release_FS_Buffer(cache, buffer);

		#ifdef DEBUG_GOSFS
		Print("Error: Directory %s already exists.\n", path);
		#endif

		ret = EEXIST;
	}
	else
	{
		//Directory/file does not exist already
		
		//Find_Dir_Entry disallows a slash at the end of parendir!
		//so remove it
		if ( parentdir[strlen(parentdir)-1] == '/' )
		{
			parentdir[strlen(parentdir)-1] = '\0';
		}
		
		//Release buffer - otherwise the second Find_Dir_Entry will fail
		if( buffer != 0  &&  buffer->flags & FS_BUFFER_INUSE)
			Release_FS_Buffer(cache, buffer);

		//Try to find parent directory entry
		struct GOSFS_Dir_Entry* parentdirentry = 0;
		ret = Find_Dir_Entry(mountPoint, parentdir, &parentdirentry,
				cache, &buffer);
		//An error occoured when trying to find parent directory
		if( ret != 0 )
		{
			#ifdef DEBUG_GOSFS
			Print("Error: Parent directory %s does not exist.\n", parentdir);
			#endif
			
			//Release buffer
			if( buffer != 0  &&  buffer->flags & FS_BUFFER_INUSE)
				Release_FS_Buffer(cache, buffer);
		}
		//Create directory
		else
		{			
			//Check that parentdir really exists
			KASSERT( parentdirentry!=0  &&  parentdirentry->flags & GOSFS_DIRENTRY_ISDIRECTORY);
			
			//Get the corresponding directory entry
			struct GOSFS_Dir_Entry newentry;
			
			//Fill out entry structure
			memset( &newentry, 0, sizeof(struct GOSFS_Dir_Entry) );
			strcpy( newentry.filename, newdir);
			newentry.size = 0;
			newentry.flags |= GOSFS_DIRENTRY_USED | GOSFS_DIRENTRY_ISDIRECTORY;
						
			//Add the entry
			ret = Add_Dir_Entry(mountPoint, parentdirentry, &newentry);
			
			//Error when adding new entry
			if( ret != 0 )
			{
				#ifdef DEBUG_GOSFS
				Print("Error: Cannnot add directory entry.\n");
				#endif							
			}
			
			else
			{	
				KASSERT( buffer != 0 );

				//Because the parentdirentry structure may have been
				//changed we have to write it down again
				Modify_FS_Buffer(cache, buffer);
				Sync_FS_Buffer(cache, buffer);		
			}

			//Release, no matter if add dir entry failed
			Release_FS_Buffer(cache, buffer);
		}
	}
	
	//release lock
	Mutex_Unlock( &gosfs->lock );
		
	//Destroy buffer
	Destroy_FS_Buffer_Cache(cache);
		
	return ret;
}

/*
 * Open a directory named by given path.
 */
static int GOSFS_Open_Directory(struct Mount_Point *mountPoint, const char *path, struct File **pDir)
{
	#ifdef DEBUG_GOSFS
	Print("Called open for directory [%s].\n", path);
	#endif
	
	//Get the gosfs implementation
	struct GOSFS_Instance* gosfs = (struct GOSFS_Instance*)mountPoint->fsData;	
	KASSERT( gosfs != 0 );
	
	
	//We use FS management structures
	Mutex_Lock( &gosfs->lock );
	
	
	//Search if directory is already open
	int idx;
	for ( idx=0; idx<GOSFS_NUM_MAX_OPEN_FILES; idx++ )
	{
		//Found file as already open
		if (
				gosfs->openFiles[idx] != 0 &&
				strcmp(gosfs->openFiles[idx]->path, path) == 0 &&
				(gosfs->openFiles[idx]->entry->flags & GOSFS_DIRENTRY_ISDIRECTORY) == 1
		   )
		{
			#ifdef DEBUG_GOSFS
			Print("Easy opening: Directory is already open at slot [%d] for this process.\n", idx);
			#endif
			
			//Allocate new file object with this gos-file			
			*pDir = Allocate_File( &s_gosfsDirOps,
					0, gosfs->openFiles[idx]->entry->size,
					gosfs->openFiles[idx],
					0, mountPoint);

			//Increment reference
			gosfs->openFiles[idx]->ref++;
			
			
			#ifdef DEBUG_GOSFS
			Print("Now ref-coutner is %d.\n", gosfs->openFiles[idx]->ref);
			#endif
			
			//So there must be at least two
			KASSERT( gosfs->openFiles[idx]->ref >= 2);
			
			//We use FS management structures
			Mutex_Unlock( &gosfs->lock );	
										
			//Finished
			return 0;
		}
	}
	
	
	//Otherwise search for an empty slot in open files table
	for ( idx=0; idx<GOSFS_NUM_MAX_OPEN_FILES; idx++ )
	{
		//Found file as already open
		if( gosfs->openFiles[idx] == 0)
			break;
	}
	
	//No free slot
	if( idx >= GOSFS_NUM_MAX_OPEN_FILES )
	{
		#ifdef DEBUG_GOSFS
		Print("Error: GOSFS can not open more files/directories.\n");
		#endif
			
		//We use FS management structures
		Mutex_Unlock( &gosfs->lock );	
		return EMFILE;
	}

	//Otherwise save the gosfs directory
	struct GOSFS_File** gosfile = & (gosfs->openFiles[idx]);
	
	
	//Create a buffer cache
	struct FS_Buffer_Cache* cache = Create_FS_Buffer_Cache(mountPoint->dev, GOSFS_FS_BLOCK_SIZE);
	struct FS_Buffer* buffer=0;		
	//No memory available
	KASSERT( cache != 0);
	
	//make sure to remove the last / from path!
	char dirPath[VFS_MAX_PATH_LEN+1];
	memset(dirPath, 0, sizeof(char) * (VFS_MAX_PATH_LEN+1));
	strncpy(dirPath, path, strlen(path));
	
	if ( dirPath[strlen(dirPath)-1] == '/' )
	{
		dirPath[strlen(dirPath)-1] = '\0';
	}
	
	//Get the corresponding directory entry
	struct GOSFS_Dir_Entry* entry;
	int ret = Find_Dir_Entry(mountPoint, dirPath, &entry, cache, &buffer);
	//Directory exists but is a file!
	if( ret == 0  &&  (entry->flags & GOSFS_DIRENTRY_ISDIRECTORY)==0 )
	{
		ret = ENOTDIR;
	}
	
	if( ret == 0 )
	{
		KASSERT(entry != 0);
		
		//The file is a file?!
		if( (entry->flags & GOSFS_DIRENTRY_ISDIRECTORY) == 0 )
		{
			ret =  EUNSUPPORTED;
		}
		else
		{
			//Create the file structure
			struct File* file = Malloc( sizeof(struct File) );
			memset(file, 0, sizeof(struct File));
			*pDir = file;

			//Fill out file structure	
			file->ops  = &s_gosfsDirOps;
			file->filePos = 0;
			file->endPos = entry->size;
			file->mountPoint = mountPoint;
			
			//We can add some data at fsData member if we need
			//this for managing files, especially if several
			//processes open the same files.
			//We save the filename here
			*gosfile = (struct GOSFS_File*) Malloc( sizeof(struct GOSFS_File));
			file->fsData =  (struct GOS_File*) *gosfile;
			
			//Create GOSFS file
			(*gosfile)->cache = Create_FS_Buffer_Cache(mountPoint->dev, GOSFS_FS_BLOCK_SIZE);
			(*gosfile)->entry = Malloc( sizeof(struct GOSFS_Dir_Entry));
			memcpy( (*gosfile)->entry, entry, sizeof(struct GOSFS_Dir_Entry));
			(*gosfile)->path = Malloc( sizeof(char)*strlen(path) );
			strcpy((*gosfile)->path, path);
			(*gosfile)->ref = 1;
		}
	}
	
	//Release block again
	if( buffer != 0  &&  buffer->flags & FS_BUFFER_INUSE )	
		Release_FS_Buffer(cache, buffer);			
	Destroy_FS_Buffer_Cache(cache);	
	
	
	//We use FS management structures
	Mutex_Unlock( &gosfs->lock );
	//Return the return-value
	return ret;
}

/*
 * Delete a file/dir named by given path.
 */
static int GOSFS_Delete(struct Mount_Point *mountPoint, const char *path)
{
	#ifdef DEBUG_GOSFS
	Print("Called delete for file/dir %s.\n", path);
	#endif
	
	//Get the gosfs implementation
	struct GOSFS_Instance* gosfs = (struct GOSFS_Instance*)mountPoint->fsData;	
	KASSERT( gosfs != 0 );
		
	
	//We use FS management structures
	Mutex_Lock( &gosfs->lock );
	
	
	//Search for file if it is already open
	int idx;
	for(idx=0; idx<GOSFS_NUM_MAX_OPEN_FILES; idx++)
	{
		//Found file as already open
		if( gosfs->openFiles[idx] != 0 &&
			strcmp( gosfs->openFiles[idx]->path, path) == 0)
		{
			#ifdef DEBUG_GOSFS
			Print("File/dir was already open at slot %d.\n", idx);
			#endif
									
			//So there must be at least two
			KASSERT( gosfs->openFiles[idx]->ref >= 1);			
			
			
			//We use FS management structures
			Mutex_Unlock( &gosfs->lock );
			return EBUSY;	
		}
	}
	
	//Create a buffer cache
	struct FS_Buffer_Cache* cache = Create_FS_Buffer_Cache(mountPoint->dev, GOSFS_FS_BLOCK_SIZE);
	struct FS_Buffer* delbuffer=0;		
	//No memory available
	KASSERT( cache != 0);
	
	struct GOSFS_Dir_Entry* delentry = 0;
	
	
	//Get the corresponding directory entry		
	int ret = Find_Dir_Entry(mountPoint, path, &delentry, cache, &delbuffer);

	//Could not find file on FS
	if( ret != 0)
	{
		#ifdef DEBUG_GOSFS
		Print("File/dir not present on FS.\n");
		#endif
		
		//We use FS management structures
		Mutex_Unlock( &gosfs->lock );
		return ENOTFOUND;
	}
	
	//So hopefully this entry is used!
	KASSERT( delentry->flags & GOSFS_DIRENTRY_USED );
	
	//If directory, check if there are files in it
	if( delentry->flags & GOSFS_DIRENTRY_ISDIRECTORY )
	{
		int slot = Find_Free_Dir_Entry_Slot(mountPoint, delentry);
		

		//There are files in directory, abort
		//Find_Free_Dir_Entry_Slot returns -1 if there are no blocks to loop through...
		//So if(slot==0 || size==0) ==> directory is empty 
		if( slot!=0 && delentry->size>0 )
		{
			#ifdef DEBUG_GOSFS
			Print("Directory not empty, aborting deleting.\n");
			#endif
			
			//We use FS management structures
			Mutex_Unlock( &gosfs->lock );
			return EINVALID;
		}
	}
	
	//Clear the is used field
	delentry->flags &= ~GOSFS_DIRENTRY_USED;
		
	//Free all blocks which belong to the entry
	Free_Blocks_Of_Dir_Entry(mountPoint, delentry);
	
	
	//We use FS management structures
	Mutex_Unlock( &gosfs->lock );
	return 0;    
}


/*
 * Get metadata (size, permissions, etc.) of file named by given path.
 * Our implementation supports, in contrast to the impl of GeekOS, also
 * directories. 
 */
static int GOSFS_Stat(struct Mount_Point *mountPoint, const char *path, struct VFS_File_Stat *stat)
{
    #ifdef DEBUG_GOSFS
	Print("Called stat for file/dir %s.\n", path);
	#endif
	
	//Get the gosfs implementation
	struct GOSFS_Instance* gosfs = (struct GOSFS_Instance*)mountPoint->fsData;	
	KASSERT( gosfs != 0 );
	
	struct GOSFS_Dir_Entry* entry = 0;
	int ret = 0;
	
	
	//We use FS management structures
	Mutex_Lock( &gosfs->lock );
	
	
	//Search for file if it is already open
	int idx;
	for(idx=0; idx<GOSFS_NUM_MAX_OPEN_FILES; idx++)
	{
		//Found file as already open
		if( gosfs->openFiles[idx] != 0 &&
			strcmp( gosfs->openFiles[idx]->path, path) == 0)
		{
			#ifdef DEBUG_GOSFS
			Print("Easy finding: File/dir was already open at slot %d.\n", idx);
			#endif
									
			//So there must be at least two
			KASSERT( gosfs->openFiles[idx]->ref >= 1);			
			
			//Set the entry
			entry = gosfs->openFiles[idx]->entry;			
		}
	}
	
	
	//Create a buffer cache
	struct FS_Buffer_Cache* cache = Create_FS_Buffer_Cache(mountPoint->dev, GOSFS_FS_BLOCK_SIZE);
	struct FS_Buffer* buffer=0;		
	//No memory available
	KASSERT( cache != 0);
	
	//Could not find the entry, search it on the disc
	if( entry == 0)
	{		
		//Get the corresponding directory entry		
		ret = Find_Dir_Entry(mountPoint, path, &entry, cache, &buffer);
	}
		
	
	//Found directory
	if( ret == 0)
	{
		#ifdef DEBUG_GOSFS
		Print("Found file: Copy stats.\n");
		#endif

		//Set members of stat
		stat->isDirectory = (entry->flags & GOSFS_DIRENTRY_ISDIRECTORY)!=0;
		stat->isSetuid = (entry->flags & GOSFS_DIRENTRY_SETUID)!=0;
		stat->size = entry->size;
		
		//Clear acls
		memset( stat->acls, 0, VFS_MAX_ACL_ENTRIES * sizeof(struct VFS_ACL_Entry));
		stat->acls[0].permission = O_READ | O_WRITE | O_EXCL | O_CREATE;
		stat->acls[0].uid = 0;
			
		#ifdef DEBUG_GOSFS
		Print("\t Ready with stat, size: %d.\n", stat->size);
		#endif	
	}
	
	//Kill buffers
	if( buffer != 0 )
		Release_FS_Buffer(cache, buffer);
	Destroy_FS_Buffer_Cache(cache);		
		
		
	//We use FS management structures
	Mutex_Unlock( &gosfs->lock );	
	return ret;
}

/*
 * Synchronize the filesystem data with the disk
 * (i.e., flush out all buffered filesystem data).
 */
static int GOSFS_Sync(struct Mount_Point *mountPoint)
{
	#ifdef DEBUG_GOSFS
	Print("Called sync for FS.\n");
	#endif
	
	//Get the gosfs implementation
	struct GOSFS_Instance* gosfs = (struct GOSFS_Instance*)mountPoint->fsData;	
	KASSERT( gosfs != 0 );
	
	//Flush all files
	int idx;
	for( idx=0; idx < GOSFS_MAGICNUMBER; idx++)
	{
		//If file open, flush it down to disc
		if( gosfs->openFiles[idx] != 0)
			Flush_GOSFS_File(mountPoint, gosfs->openFiles[idx]);
	}
    
    return 0;
}

/*static*/ struct Mount_Point_Ops s_gosfsMountPointOps = {
    &GOSFS_Open,
    &GOSFS_Create_Directory,
    &GOSFS_Open_Directory,
    &GOSFS_Stat,
    &GOSFS_Sync,
    &GOSFS_Delete,
};

static int GOSFS_Format(struct Block_Device *blockDev)
{
	//Create a superblock structure
	struct GOSFS_Superblock superblock; 
	int i;
	
	//We need interrupts for disc here
	KASSERT(Interrupts_Enabled());
	KASSERT( sizeof(struct GOSFS_Superblock) <= GOSFS_FS_BLOCK_SIZE);
	
	//Get the number of available blocks on device
	int cntblocks = Get_Num_Blocks(blockDev);
	
	//We need more blocks
	if( cntblocks < 3 )
	{
		#ifdef DEBUG_GOSFS
		Print("Error: Formating blockdev with GOSFS with only %d blocks!\n", cntblocks); 
		#endif 
		return -1;
	} 
	
	
	//Fill superblock struct with values
	//First three bytes of magic number form 'hrw'	
	memset(&superblock, 0, sizeof(struct GOSFS_Superblock));
	superblock.magicnumber = GOSFS_MAGICNUMBER;
	superblock.version = 1;
	superblock.cntblocks = cntblocks;
	
	//Set directory entry of root
	superblock.root.size = 0;
	superblock.root.flags |= GOSFS_DIRENTRY_ISDIRECTORY;
	
	
	#ifdef DEBUG_GOSFS
	Print("Formating blockdev with GOSFS with %d blocks @ %d bytes.\n", superblock.cntblocks, GOSFS_FS_BLOCK_SIZE); 
	#endif 
	
	
	//Create a buffer cache
	struct FS_Buffer_Cache* cache = Create_FS_Buffer_Cache(blockDev, GOSFS_FS_BLOCK_SIZE);
	struct FS_Buffer* buf=0;
	
	//No memory available
	if( cache == 0)
		return ENOMEM;
	
	
	//Get the superblock
	Get_FS_Buffer(cache, 0, &buf);
		
	//Copy data to block buffer
	memset(buf->data, 0, GOSFS_FS_BLOCK_SIZE);
	memcpy(buf->data, &superblock, sizeof(superblock));
		
	//Release buffer	
	Modify_FS_Buffer(cache, buf);
	Sync_FS_Buffer(cache, buf);
	Release_FS_Buffer(cache, buf);
	
	
	#ifdef DEBUG_GOSFS
	Print("\tWrote down superblock.\n"); 
	#endif 

	
	
	
	
	//Size of bitmap in bytes and number of blocks
	const unsigned bitmapBlocks = GOSFS_BITMAP_BLOCKS( superblock.cntblocks ); 
	
	//Build bitmap about free blocks, FS_BLOCK_SIZE aligned
	void* freebitmap = Create_Bit_Set(bitmapBlocks * GOSFS_FS_BLOCK_SIZE * 8);
	KASSERT(freebitmap);
	
	//Mark superblock and bitmap-blocks as not-free
	for(i=0; i<=bitmapBlocks; i++)
		Set_Bit(freebitmap, i);
		
	//Write bitmap to block device 
	for(i=0; i<bitmapBlocks; i++)	
	{
		//Get the superblock
		Get_FS_Buffer(cache, i+1, &buf);
		
		//Copy data to block buffer
		memset(buf->data, 0, GOSFS_FS_BLOCK_SIZE);
		memcpy(buf->data, freebitmap + GOSFS_FS_BLOCK_SIZE*i, GOSFS_FS_BLOCK_SIZE);
	
		//Release buffer	
		Modify_FS_Buffer(cache, buf);
		Sync_FS_Buffer(cache, buf);
		Release_FS_Buffer(cache, buf);
	}
	
	
	//Destroy the cache which is not needed any more
	Destroy_FS_Buffer_Cache(cache);
		
		
	#ifdef DEBUG_GOSFS
	Print("\tWrote down %d blocks for bitmap.\n", bitmapBlocks); 
	#endif 
		 
	
	return 0;
}

static int GOSFS_Mount(struct Mount_Point *mountPoint)
{
	//Creatge a new gosfs instance
	struct GOSFS_Instance* instance = Malloc( sizeof(struct GOSFS_Instance) );	
	struct GOSFS_Superblock superblock;
	
	//No mem available
	if( instance == 0 )
		return ENOMEM;
			
	//Clear instance
	memset(instance, 0, sizeof(struct GOSFS_Instance));
	#ifdef DEBUG_GOSFS
	Print("Created GOSFS_Instance object\n"); 
	#endif 
	
	
	//Create a buffer cache
	struct FS_Buffer_Cache* cache = Create_FS_Buffer_Cache(mountPoint->dev, GOSFS_FS_BLOCK_SIZE);
	struct FS_Buffer* buf=0;
	
	//No memory available
	if( cache == 0)
	{
		Free(instance);
		return ENOMEM;
	}
	
	//Get the superblock at block 0
	Get_FS_Buffer(cache, 0, &buf);
	KASSERT( buf!=0 );
		
	//Copy data of first block into superblock structure
	memcpy(&superblock, buf->data, sizeof(struct GOSFS_Superblock));	
	Release_FS_Buffer(cache, buf);
	
	
	//Check magicnumber
	if( superblock.magicnumber != GOSFS_MAGICNUMBER )
	{
		Print("Error: Magicnumber of GOSFS superblock is invalid.\n");
		Free(instance);
		Destroy_FS_Buffer_Cache(cache);
		return -1;
	}
	
	#ifdef DEBUG_GOSFS
	Print("\tGOS_FS has %d bocks.\n", superblock.cntblocks); 
	#endif 
	
	//Fill data of instance variable
	instance->cntblocks = superblock.cntblocks;
	instance->usedbitmap = Create_Bit_Set( superblock.cntblocks );
	instance->mountPoint = mountPoint;	
	KASSERT( instance->usedbitmap != 0);	
	char* bitmap = instance->usedbitmap;
		
	//Init mutex 
	Mutex_Init( &instance->lock );
		
		
	
	#ifdef DEBUG_GOSFS
	Print("\tReading %d bitmap blocks.\n", GOSFS_BITMAP_BLOCKS( superblock.cntblocks )); 
	#endif 
		
		
	//Read used-blocks bitmap
	int i;
	for( i=0; i<GOSFS_BITMAP_BLOCKS( superblock.cntblocks ); i++)
	{	
		//Get the superblock
		Get_FS_Buffer(cache, 1+i, &buf);
		KASSERT( buf!=0 );
				
		//Not last block, copy full block
		if( i+1 < GOSFS_BITMAP_BLOCKS( superblock.cntblocks ) )
		{
			#ifdef DEBUG_GOSFS
			Print("\t\tCopying full block %d of %d.\n", i, GOSFS_BITMAP_BLOCKS( superblock.cntblocks )); 
			#endif 
			
			//Copy full blocks
			memcpy(&bitmap[i*GOSFS_FS_BLOCK_SIZE],
					buf->data,
					GOSFS_FS_BLOCK_SIZE );			
		}  
		else
		{
			#ifdef DEBUG_GOSFS
			Print("\t\tCopying remaining block with %d Bytes.\n", (superblock.cntblocks/8) % GOSFS_FS_BLOCK_SIZE); 
			#endif 
							
			//For the last block we only copy the remaining data
			//So we have superblock.cntblocks/8 Bytes over all
			//Therefore the last block holds cnt../8 mod blocksize
			memcpy(&bitmap[i*GOSFS_FS_BLOCK_SIZE],
					buf->data,
					(superblock.cntblocks/8) % GOSFS_FS_BLOCK_SIZE );
		}
		
		
		//Release buffer
		Release_FS_Buffer(cache, buf);
	}
	
	
	
	//So this and the next block must be used!
	KASSERT( Is_Bit_Set(instance->usedbitmap, 0) == true );
	KASSERT( Is_Bit_Set(instance->usedbitmap, 1) == true );
	
	
	//Destroy the cache which is not needed any more
	Destroy_FS_Buffer_Cache(cache);
	
	
	//Set the FS data object and the pointers to the mount point
	//operations
	mountPoint->fsData = instance;
	mountPoint->ops = &s_gosfsMountPointOps;
			
	
	return 0;	
}



static struct Filesystem_Ops s_gosfsFilesystemOps = {
    &GOSFS_Format,
    &GOSFS_Mount,
};



/* ----------------------------------------------------------------------
 * Public functions
 * ---------------------------------------------------------------------- */

void Init_GOSFS(void)
{
    Register_Filesystem("gosfs", &s_gosfsFilesystemOps);
}

