/*
 * Paging (virtual memory) support
 * Copyright (c) 2003, Jeffrey K. Hollingsworth <hollings@cs.umd.edu>
 * Copyright (c) 2003,2004 David H. Hovemeyer <daveho@cs.umd.edu>
 * $Revision: 1.55 $
 * 
 * This is free software.  You are permitted to use,
 * redistribute, and modify it as specified in the file "COPYING".
 */

#include <geekos/string.h>
#include <geekos/int.h>
#include <geekos/idt.h>
#include <geekos/kthread.h>
#include <geekos/kassert.h>
#include <geekos/screen.h>
#include <geekos/mem.h>
#include <geekos/malloc.h>
#include <geekos/gdt.h>
#include <geekos/segment.h>
#include <geekos/user.h>
#include <geekos/vfs.h>
#include <geekos/crc32.h>
#include <geekos/paging.h>


//Uncomment this to print debug messages
//#define DEBUG_PAGING

#ifdef DEBUG_PAGING
	//Uncomment this to print access bits of pages while update clock
	//#define DEBUG_PRINT_PAGEACCESSBIT
#endif


/* ----------------------------------------------------------------------
 * Public data
 * ---------------------------------------------------------------------- */

/* ----------------------------------------------------------------------
 * Private functions/data
 * ---------------------------------------------------------------------- */

#define SECTORS_PER_PAGE (PAGE_SIZE / SECTOR_SIZE)


/* global reference to kernel page dir */
pde_t *g_kernelPageDir = 0;

/** A array of indices indicating used frames on swap device */
static bool* s_usedSwapIndices = 0;
/** Amount of free space on swap device */
static int  s_cntSwapFrames = 0;

/*
 * flag to indicate if debugging paging code
 */
int debugFaults = 0;
#define Debug(args...) if (debugFaults) Print(args)


void checkPaging()
{
  unsigned long reg=0;
  __asm__ __volatile__( "movl %%cr0, %0" : "=a" (reg));
  Print("Paging on ? : %d\n", (reg & (1<<31)) != 0);
}


/*
 * Print diagnostic information for a page fault.
 */
static void Print_Fault_Info(uint_t address, faultcode_t faultCode)
{
    extern uint_t g_freePageCount;

    Print("Pid %d, Page Fault received, at address %x (%d pages free)\n",
        g_currentThread->pid, address, g_freePageCount);
    if (faultCode.protectionViolation)
        Print ("   Protection Violation, ");
    else
        Print ("   Non-present page, ");
    if (faultCode.writeFault)
        Print ("Write Fault, ");
    else
        Print ("Read Fault, ");
    if (faultCode.userModeFault)
        Print ("in User Mode\n");
    else
        Print ("in Supervisor Mode\n");
}

/*
 * Handler for page faults.
 * You should call the Install_Interrupt_Handler() function to
 * register this function as the handler for interrupt 14.
 */
/*static*/ void Page_Fault_Handler(struct Interrupt_State* state)
{
    ulong_t address;
    faultcode_t faultCode;

    KASSERT(!Interrupts_Enabled());

    /* Get the linear address that caused the page fault */
    address = Get_Page_Fault_Address();

    /* Get the fault code */
    faultCode = *((faultcode_t *) &(state->errorCode));
    
    //Update clock for page replacement algorithms
    Update_Clock( );
       
    
    
    //Page fault in superuser mode! - This should not be!!
    if( !faultCode.userModeFault )
    {
    	Print("Page fault in kernel mode while handling " \
    		  "kthread %p with pid %d!!!\n",
    		  g_currentThread, g_currentThread->pid);
    	Print("Page fault @ 0x%lx\n", address);
    	Print_Fault_Info(address, faultCode);
    	Dump_Interrupt_State(state);
    	
    	KASSERT(0);
    }    
    
    
    //Error with access rights or non present table/page
    //ATTENTION: It was the case that this bit was NOT set altough,
    //we had to expand the stack.
    if( faultCode.protectionViolation || true )
    {
    	#ifdef DEBUG_PAGING
    	Print("Kthread %p with pid %d wants to access a page and this "\
    		  "led to a page fault!\n",
    		g_currentThread, g_currentThread->pid);
    	Print("Page fault at lin addr. 0x%lx\n", address);
    	Print("Look for reason of page fault...\n");
    	#endif
    	   	
		//Indices for directory entry and table entry
		const unsigned direIdx = ((unsigned)address >> PAGE_POWER) / NUM_PAGE_DIR_ENTRIES;
		const unsigned tabeIdx = ((unsigned)address >> PAGE_POWER) % NUM_PAGE_DIR_ENTRIES;
		
		//dir entry of table
		pde_t* dire = &g_currentThread->userContext->pageDir[direIdx];
		
		//Page tables are not paged out --> therefore page table
		//did not exist yet. --> Invalid memory access by user process 
		if( dire->present == 0)
		{
			Print("Page table not present --> invalid mem access!");
			Exit(-1);
	    	KASSERT(0);
		}
		//Page table is not user mode --> therefore supervisor mode
		else if( (dire->flags & VM_USER) == 0 )
		{
			Print("Page table is not in user mode! --> "\
				"Invalid mem access!\n");
			Exit(-1);
	    	KASSERT(0);
		}
		
		//Get page table and table entry
		pte_t* table = (pte_t*) (dire->pageTableBaseAddr << PAGE_POWER);
		pte_t* tabe = & table[tabeIdx];
		
				
		
		//Page is paged out 
		if( tabe->kernelInfo & KINFO_PAGE_ON_DISK )
		{						
			//Therefore page is not present
			KASSERT( tabe->present == 0 );
			//This assert is invalid: If page is paged out, then the
			//correspondig page struct does not correspond to this entry
			//Furthermore pageBaseAddr is the index in the swap file...
			//KASSERT( page->flags & PAGE_PAGEABLE );
			
			#ifdef DEBUG_PAGING
			Print("Try to acces paged-out page.\n");
			#endif
						
			//Get the page from swap file to RAM
			Bring_In_Page( (unsigned)address & ~PAGE_MASK);
			
			return;
		}
		//Page is not paged out and not present
		else if( tabe->present == 0 )
		{
			#ifdef DEBUG_PAGING
			Print("Page is not present and not paged out.\n");
			#endif
			
			//Stack and argument block have been allocated at user
			//context creation time
			KASSERT( tabeIdx != NUM_PAGE_TABLE_ENTRIES-1);
			KASSERT( tabeIdx != NUM_PAGE_TABLE_ENTRIES-2);
			
			//Check if page above exits --> then we expand stack
			//Otherwise invalid mem access
			//ATTENTION: We already checked that page table is existing
			//So stack+argblock can not grow beyond the 4Meg limit which
			//can be allocated by a single table.
			if( faultCode.writeFault && 
					(table[tabeIdx+1].present==1 || table[tabeIdx+1].kernelInfo & KINFO_PAGE_ON_DISK ))
			{
				#ifdef DEBUG_PAGING
				Print("Fault code at writing, page above exists --> expand stack\n");
				#endif
				
    			KASSERT(!Interrupts_Enabled());
				
				//Allocate new part for stack: table table entry is &table[tabe]
				//Virtual address of page is page aligned address of 'address',
				void* stackpart = Alloc_Pageable_Page( tabe, address & ~PAGE_MASK );
				
				//No more memory -- kill process
				if( stackpart == 0)
					Exit(-1);
				
				//Clear content of page
				memset(stackpart, 0, PAGE_SIZE);
				//Reset all flags of page table entry
				memset(tabe, 0, sizeof(pte_t));
				
				//The table is now present
				tabe->present = 1;
				//We are only interessted in the higher 20 bits
				tabe->pageBaseAddr = ((unsigned)stackpart) >> PAGE_POWER;
				//user table, writeable
				tabe->flags = VM_USER | VM_WRITE;	
				
				return;
			}
			//Do not expand stack --> mem fault
			else
			{			
				Print("Invalid memory acces.\n");
				Exit(-1);
		    	KASSERT(0);
			}
		}
		//Process wants to write at not write-able page
		else if( faultCode.writeFault  &&  (tabe->flags & VM_WRITE) == 0 )
   		{
	    	Print("Kthread %p with pid %d wants to write at read-only page!\n",
	    		g_currentThread, g_currentThread->pid);
	    	Print("Page fault @ 0x%lx\n", address);
	    	
	    	//Terminate process
	    	Exit(-1);
	    	KASSERT(0);
    	}
    	//Page is not user mode --> therefore supervisor mode
		else if( (tabe->flags & VM_USER) == 0 )
		{
			Print("Page is not in user mode! --> "\
				"Invalid mem access!\n");
			Exit(-1);
	    	KASSERT(0);
		}    	
    }    
    

    /* rest of your handling code here */
    Print ("Unexpected Page Fault received\n");
    Print_Fault_Info(address, faultCode);
    Dump_Interrupt_State(state);
    /* user faults just kill the process */
    if (!faultCode.userModeFault) KASSERT(0);

    /* For now, just kill the thread/process. */
    Exit(-1);
}

/* ----------------------------------------------------------------------
 * Public functions
 * ---------------------------------------------------------------------- */


/*
 * Initialize virtual memory by building page tables
 * for the kernel and physical memory.
 */
void Init_VM(struct Boot_Info *bootInfo)
{
    /*
     * Hints:
     * - Build kernel page directory and page tables
     * - Call Enable_Paging() with the kernel page directory
     * - Install an interrupt handler for interrupt 14,
     *   page fault
     * - Do not map a page at address 0; this will help trap
     *   null pointer references
     */
	
	//Intial directory/table setup: We want to map the whole
	//phy memory into the virtual memory.
	
	
	//Guarantee that a page directory needs a page and not more
	KASSERT( sizeof(pde_t)*NUM_PAGE_DIR_ENTRIES == PAGE_SIZE );
	//Guarantee that a page table needs a page and not more
	KASSERT( sizeof(pte_t)*NUM_PAGE_TABLE_ENTRIES == PAGE_SIZE );
	
	//build kernel page directory. We use alloc page here because
	//we need 4k-aligned memory and a directory fits in a page
	//anyway.
	pde_t *pageDir = Alloc_Page();
	//set global reference
	g_kernelPageDir = pageDir;
	//constraint for running the kernel
	KASSERT ( pageDir != 0 );
	//set all fields to absent
	memset(pageDir, 0, PAGE_SIZE);

	
	//Number of page frames available. A page takes 2^PAGE_POWER bytes
	//and we have memSizeKB.2^10 bytes as phy. mem.
	const unsigned numFrames = (bootInfo->memSizeKB >> (PAGE_POWER-10));
	//round up number of page tables to map the whole physical memory
	const unsigned numTables = (numFrames+NUM_PAGE_DIR_ENTRIES-1)/NUM_PAGE_DIR_ENTRIES;
	
	Print("Activate paging with numFrames: %d, numTables: %d\n", numFrames, numTables);
	Print("\tMax phy. addr: 0x%x\n", 1024*bootInfo->memSizeKB);

	//counter of page frames (in physical memory)
	unsigned frameCounter;
		
	//Loop over whole phy memory.
	for( frameCounter=0; frameCounter<numFrames; frameCounter++ )
	{
		//Index of page table in page directory
		const unsigned tableIdx = frameCounter/NUM_PAGE_TABLE_ENTRIES;
		//Index of page in page table (of no. tableIdx)
		const unsigned pageIdx = frameCounter % NUM_PAGE_TABLE_ENTRIES;

		//Print("%d %d %d\n", frameCounter, tableIdx, pageIdx);

		//If we map the whole phy mem to VM, this must hold
		KASSERT(tableIdx < numTables);
		
		//Create new page table
		if( pageIdx == 0 )
		{
			//Create a new page table. Same reason for alloc page as above
			pte_t *pageTable = Alloc_Page();
			//if there is no mem for a page table we are lost
			KASSERT( pageTable != 0 );
			//mark all entries as absent
			memset(pageTable, 0, PAGE_SIZE);

			//Get corresponding entry in dir
			pde_t* entry = &pageDir[tableIdx];
			
			//The table is now present
			entry->present = 1;
			//We are only interessted in the higher 20 bits
			entry->pageTableBaseAddr = ((unsigned)pageTable) >> PAGE_POWER;
			//supervisor table, writeable
			entry->flags = VM_WRITE;
		}
		
		
		//Skip first entry
		//Hint from comment above: Supports pointer to zero error debugging.		
		if( frameCounter == 0 )
			continue;
		
		//Get corresponding entry in dir
		pde_t* dirEntry = &pageDir[tableIdx];
		KASSERT(dirEntry->present);
		//Get pointer to the table itself
		pte_t* table = (pte_t*) (dirEntry->pageTableBaseAddr << PAGE_POWER);	
			
		//Table entry of curr page
		pte_t* tableEntry = &table[pageIdx];
		
		//Same flags as for directory
		tableEntry->present = 1;
		tableEntry->flags = dirEntry->flags;
		tableEntry->pageBaseAddr = frameCounter;
	}
	
	
	
	//The moment when we first enable paging - god saves hrwos.
	Enable_Paging(pageDir);
	
	//install the page fault handler
	Install_Interrupt_Handler(14, &Page_Fault_Handler);
}

/**
 * Initialize paging file data structures.
 * All filesystems should be mounted before this function
 * is called, to ensure that the paging file is available.
 */
void Init_Paging(void)
{	
    Print("Initialize paging file data structures...\n");
        
     // get the paging file
    struct Paging_Device *p_dev = Get_Paging_Device();
    
    //No paging file
    if( p_dev == 0)    
    	Print("\tThere is no paging dev!\n");
          
    
    //Get dimensions... 
    const unsigned sectorsPerPage = PAGE_SIZE / SECTOR_SIZE; 
    
    //Paging device available
    if( p_dev )
    {
    	//Number of frames available on swap device
    	s_cntSwapFrames = p_dev->numSectors / sectorsPerPage;
    	 
    	//Get mem for the free indicators
	    Print("\tSpace for %d page frames.\n", s_cntSwapFrames);    
	    s_usedSwapIndices = Malloc( sizeof(bool) * s_cntSwapFrames );	    
    	KASSERT( s_usedSwapIndices != 0); 
	    
	    //Set them all to zero
    	memset(s_usedSwapIndices, 0, sizeof(bool) * s_cntSwapFrames );
    }
    else
    {
    	s_cntSwapFrames = 0;
    }        
}

/**
 * Find a free bit of disk on the paging file for this page.
 * Interrupts must be disabled.
 * @return index of free page sized chunk of disk space in
 *   the paging file, or -1 if the paging file is full
 */
int Find_Space_On_Paging_File(void)
{
    KASSERT(!Interrupts_Enabled());
    
    int idx;
        
    //Return first unused frame on swap device
    for( idx=0; idx < s_cntSwapFrames; idx++)
    	if( !s_usedSwapIndices[idx] )    	
    		return idx;
    
    //No free frame found
    return -1;
}

/**
 * Free a page-sized chunk of disk space in the paging file.
 * Interrupts must be disabled.
 * @param pagefileIndex index of the chunk of disk space
 */
void Free_Space_On_Paging_File(int pagefileIndex)
{
    KASSERT(!Interrupts_Enabled());    
    KASSERT( 0 <= pagefileIndex && pagefileIndex < s_cntSwapFrames);
    
    //Not used any more
    s_usedSwapIndices[pagefileIndex] = false;
}

/**
 * Write the contents of given page to the indicated block
 * of space in the paging file.
 * @param paddr a pointer to the physical memory of the page
 * @param vaddr virtual address where page is mapped in user memory
 * @param pagefileIndex the index of the page sized chunk of space
 *   in the paging file
 */
void Write_To_Paging_File(void *paddr, ulong_t vaddr, int pagefileIndex)
{
	#ifdef DEBUG_PAGING
	Print("Writing page to swap: paddr: %p, vaddr: %p, pagefileidx: %d\n",
		paddr, (void*)vaddr, pagefileIndex);
	#endif
	
    struct Page *page = Get_Page((ulong_t) paddr);
    KASSERT(!(page->flags & PAGE_PAGEABLE)); /* Page must be locked! */
    
    //When this function is called, ints are enabled
    KASSERT( Interrupts_Enabled() );
    
    //Check bounds and if frame is free
    KASSERT( 0 <= pagefileIndex && pagefileIndex < s_cntSwapFrames);
    KASSERT( !s_usedSwapIndices[pagefileIndex] );
    
    // get the paging file
    struct Paging_Device *p_dev = Get_Paging_Device();   
    
    //Get dimensions...  
    const unsigned startSector = p_dev->startSector+ pagefileIndex * SECTORS_PER_PAGE;
    unsigned sector;
    
    
    //Before enabeling interrupts!, indicate frame as used
    s_usedSwapIndices[pagefileIndex] = true;
    
    
    //Write all sectors to file
    for(sector=0; sector < SECTORS_PER_PAGE; sector++)
    	//Write to sector the content from paddr
    	Block_Write(p_dev->dev, startSector+sector, paddr+sector*SECTOR_SIZE);    
}

/**
 * Read the contents of the indicated block
 * of space in the paging file into the given page.
 * @param paddr a pointer to the physical memory of the page
 * @param vaddr virtual address where page will be re-mapped in
 *   user memory
 * @param pagefileIndex the index of the page sized chunk of space
 *   in the paging file
 */
void Read_From_Paging_File(void *paddr, ulong_t vaddr, int pagefileIndex)
{   
	#ifdef DEBUG_PAGING
	Print("Reading page from swap: paddr: %p, vaddr: %p, pagefileidx: %d\n",
		paddr, (void*) vaddr, pagefileIndex);
	#endif
	 
	
    struct Page *page = Get_Page((ulong_t) paddr);
    KASSERT(!(page->flags & PAGE_PAGEABLE)); /* Page must be locked! */
    
    //When this function is called, ints are enabled
    KASSERT( Interrupts_Enabled() );
    
    
    //Check bounds and if frame is free
    KASSERT( 0 <= pagefileIndex && pagefileIndex < s_cntSwapFrames);
    KASSERT( s_usedSwapIndices[pagefileIndex] );
    
    // get the paging file
    struct Paging_Device *p_dev = Get_Paging_Device();   
    
    //Get dimensions... 
    const unsigned startSector = p_dev->startSector+ pagefileIndex * SECTORS_PER_PAGE;
    unsigned sector;
    
    
    
    //Read all sectors to file
    for(sector=0; sector < SECTORS_PER_PAGE; sector++)
    	//Read from sector the content from paddr
    	Block_Read(p_dev->dev, sector+startSector, paddr+sector*SECTOR_SIZE);
    	
    	
    //Before enabeling interrupts!, indicate frame as not used
    s_usedSwapIndices[pagefileIndex] = false;    
}


/**Update clock for page replacement algorithm (Find_Page_To_Page_Out)*/
void Update_Clock()
{
	int idx;
	struct Page* page;
	
	#ifdef DEBUG_PAGING
	Print("Update clock\n");
	#endif
	
	//Run through all pages (page frames)
	for(idx=0; idx < s_numPages; idx++)
	{
		//Page is not pageable -- skip it
		if( (g_pageList[idx].flags & PAGE_PAGEABLE ) == 0)
			continue;
		//Page is not allocated -- skip it
		if( (g_pageList[idx].flags & PAGE_ALLOCATED ) == 0)
			continue;
			
		//Current pageable, allocated page
		page = &g_pageList[idx];
		//And corresponding table entry
		pte_t* pte = page->entry;
		
		//If page has been accessed resently (accessed bit set)
		//we increment the frequency indicator 'clock'
		if( page->clock < 1<<30  &&  pte->accessed)
		{
			#ifdef DEBUG_PRINT_PAGEACCESSBIT
			Print("Page %d has been accessed\n", idx);
			#endif
			page->clock++;
		}
						
		//If page has not been accessed resently (accessed bit clear)
		//we decrement the frequency indicator 'clock'
		if( page->clock > 0  &&  !pte->accessed)
		{
			#ifdef DEBUG_PRINT_PAGEACCESSBIT
			Print("Page %d has not been accessed\n", idx);
			#endif
			page->clock--;
		}			
		
		//Reset accessed bit
		pte->accessed = 0;	
	}
}


/** Bring in a page with specific virtual address */
void Bring_In_Page(unsigned virtAddr)
{	
	KASSERT( (virtAddr & PAGE_MASK) == 0);
	KASSERT( !Interrupts_Enabled() );
	
	//Indices for directory entry and table entry
	const unsigned direIdx = ((unsigned)virtAddr >> PAGE_POWER) / NUM_PAGE_DIR_ENTRIES;
	const unsigned tabeIdx = ((unsigned)virtAddr >> PAGE_POWER) % NUM_PAGE_DIR_ENTRIES;
	
	//dir entry of table
	pde_t* dire = &g_currentThread->userContext->pageDir[direIdx];
	
	//Page tables are not paged out --> therefore page table
	//did not exist yet. --> Invalid memory access by user process 
	KASSERT( dire->present == 1);
	KASSERT( dire->flags & VM_USER);
	
	
	//Get page table and table entry
	pte_t* table = (pte_t*) (dire->pageTableBaseAddr << PAGE_POWER);
	pte_t* tabe = & table[tabeIdx];
	
	
	//Index where swapped out page lies
	const unsigned swapFileIdx = tabe->pageBaseAddr; 
	
	//Allocate an empty page where we can load the data
	//From swapfile
	void* phyaddr = Alloc_Pageable_Page( tabe, virtAddr);
	tabe->pageBaseAddr = (unsigned)phyaddr >> PAGE_POWER;
	tabe->present = 1;
	//user table, writeable
	tabe->flags = VM_USER | VM_WRITE;
	
	
    struct Page *page = Get_Page((ulong_t) phyaddr);
	
	//Clear pageable bit so that page is not stolen...	
	page->flags &= ~PAGE_PAGEABLE;
		
	//Needed for I/O
	Enable_Interrupts();	
	//Read swapped out page into page frame
	Read_From_Paging_File( phyaddr, virtAddr, swapFileIdx);	
	Disable_Interrupts();
							
	//Reset pageable bit	
	page->flags |= PAGE_PAGEABLE;
	//Accesse now..
	page->clock = 10;
}
