/*
 * Pipe pseudo-filesystem
 * Copyright (c) 2004, David H. Hovemeyer <daveho@cs.umd.edu>
 * $Revision: 1.12 $
 *
 * This is free software.  You are permitted to use,
 * redistribute, and modify it as specified in the file "COPYING".
 */

#include <geekos/kassert.h>
#include <geekos/errno.h>
#include <geekos/vfs.h>
#include <geekos/malloc.h>
#include <geekos/string.h>
#include <geekos/synch.h>
#include <geekos/user.h>
#include <geekos/pipefs.h>
#include <geekos/int.h>

//#define DEBUG_PIPEFS


/* ----------------------------------------------------------------------
 * Private data and functions
 * ---------------------------------------------------------------------- */
struct Pipes_List g_pipesList;

/*
 * Read data from a pipe.
 * Returns number of bytes read, or 0 if end-of-file
 * has been reached.
 */
static int Pipe_Read(struct File *file, void *buf, ulong_t numBytes)
{
    struct Pipe *pipe = file->fsData;
    
    //Assure mutual exclusion
    Mutex_Lock(&pipe->mutex);
    
    ulong_t usedBytes = (pipe->writePos-pipe->readPos+PIPE_BUF_SIZE)%PIPE_BUF_SIZE;

	#ifdef DEBUG_PIPEFS
	Print("Buffer has %ld used bytes\n", usedBytes);
	#endif
	
    //Number of bytes read by this function
    ulong_t readBytes = MIN(numBytes, usedBytes);
    
   	if ( pipe->readPos+readBytes > PIPE_BUF_SIZE )
	{
		ulong_t finalPart = PIPE_BUF_SIZE-pipe->readPos;
		
		//Copy last part of buffer
		memcpy(buf, pipe->buffer+pipe->readPos, finalPart);
		
		//Copy remaining bytes from front of buffer
		memcpy(buf+finalPart, pipe->buffer, readBytes-finalPart);
	}
	else
	{
		//We can copy the whole part
		memcpy(buf, pipe->buffer+pipe->readPos, readBytes);
	}
	
	pipe->readPos = (pipe->readPos+readBytes)%PIPE_BUF_SIZE;
    
	

    #ifdef DEBUG_PIPEFS
    Print("Had to read %ld bytes - read %ld\n", numBytes, readBytes);
    #endif

    Mutex_Unlock(&pipe->mutex);
    
    return readBytes;
}

/*
 * Write data to pipe.
 * Returns number of bytes written.
 */
static int Pipe_Write(struct File *file, void *buf, ulong_t numBytes)
{
    struct Pipe *pipe = file->fsData;

    //Assure mutual exclusion
    Mutex_Lock(&pipe->mutex);

	//Calculate free bytes in buffer
	ulong_t usedBytes = (pipe->writePos-pipe->readPos + PIPE_BUF_SIZE)%PIPE_BUF_SIZE;
	ulong_t freeBytes = PIPE_BUF_SIZE-1-usedBytes;
	
	#ifdef DEBUG_PIPEFS
	Print("Buffer has %ld free bytes\n", freeBytes);
	#endif

    const ulong_t writtenBytes = MIN(numBytes, freeBytes);

	
	if( pipe->writePos+numBytes > PIPE_BUF_SIZE )
	{
		ulong_t finalPart = PIPE_BUF_SIZE-pipe->writePos;

		//Copy part till PIPE_BUF_SIZE-1
		memcpy(pipe->buffer+pipe->writePos, buf, finalPart);
		
		//Copy to buffer from 0 beginning
		memcpy(pipe->buffer, buf+finalPart, writtenBytes-finalPart);
	}
	else
	{
		memcpy(pipe->buffer+pipe->writePos, buf, writtenBytes);
	}

	pipe->writePos = (pipe->writePos+writtenBytes)%PIPE_BUF_SIZE;
	
	

    Mutex_Unlock(&pipe->mutex);
    
	#ifdef DEBUG_PIPEFS
	Print("Had to write %ld bytes - wrote %ld bytes\n", numBytes, writtenBytes);
	#endif
    
    return writtenBytes;
}

/*
 * Close pipe.
 */
static int Pipe_Close(struct File *file)
{
	//Make sure we don't get interrupted! otherwise access to our list might be a problem
	KASSERT ( Interrupts_Enabled() );


	//Get fsData (pipe)
	struct Pipe *pipe = (struct Pipe*)file->fsData;
	
	//Mutual exclusion while accessing members of the pipe!
	Mutex_Lock(&pipe->mutex);
	
	//Make sure we got a pipe!
	KASSERT(pipe!=0);

	//Make sure we got a pipe FD (one is readable and one is writeable!)
	KASSERT( !(file->mode&O_WRITE && file->mode&O_READ) );

	//Decrease ref counts
	KASSERT(pipe->refCount>0);
	//We only close ONE fd!! so here -1
	pipe->refCount--;
	
	//If noone refers to this pipe anymore -> free is
	if ( pipe->refCount==0 )
	{
		//Free name
		if( pipe->name != 0)
			Free( pipe->name );
		
		//Clear pipe from queue
		Remove_From_Pipes_List(&g_pipesList, pipe);

		#ifdef DEBUG_PIPEFS
		Print("Freed pipe 0x%p\n", pipe);
		#endif

		//Free the pipe
		Free(pipe);
	}
	else
	{
		//We only have to unlock an existing pipe :D
		Mutex_Unlock(&pipe->mutex);
	}

	return 0;
}

/*
 * Clone a pipe.
 */
static int Pipe_Clone(struct File *file, struct File **pClone)
{    
    //Make sure we got ONLY write XOR read descriptor
    KASSERT ( !(file->mode&O_READ && file->mode&O_WRITE) );
    
    //Allocate file and copy the original one
    *pClone = Malloc(sizeof(struct File));
    if ( *pClone == 0 )
    	return ENOMEM;
    
    //now copy the file
    memcpy(*pClone, file, sizeof(struct File));

	//Get the original pipe
    struct Pipe *pipe = (struct Pipe*)file->fsData;
    
    KASSERT(pipe!=0);

    //Increment the according refCount
	pipe->refCount++;

    return 0;
}

static struct File_Ops s_readPipeFileOps = {
    0,			/* FStat() */
    &Pipe_Read,
    0,			/* Write() */
    0,			/* Seek() */
    &Pipe_Close,
    0,			/* Read_Entry() */
    &Pipe_Clone,
};

static struct File_Ops s_writePipeFileOps = {
    0,			/* FStat() */
    0,			/* Read() */
    &Pipe_Write,
    0,			/* Seek() */
    &Pipe_Close,
    0,			/* Read_Entry() */
    &Pipe_Clone,
};

/*
 * Initialize all members of a pipe
 */
static void Init_Pipe(struct Pipe *pipe)
{
	//By default - unnamed pipe
	pipe->name = 0;

	pipe->refCount = 0;
	
	memset(pipe->buffer, 0, sizeof(pipe->buffer));

	pipe->readPos  = 0;
	pipe->writePos = 0;

	Mutex_Init(&pipe->mutex);
}

/*
 * Free allocated pipe files
 */
static void Free_Pipe_Files(struct File **pRead, struct File **pWrite)
{
	if (*pRead != 0)
	    Free(*pRead);
	if (*pWrite != 0)
	    Free(*pWrite);
}

/* 
 * Allocate Files for the passed pointers
 * In case of error a probably allocated file will be freed!
 */
static int Allocate_Pipe_Files(struct File **pRead, struct File **pWrite)
{
    struct File *read = 0, *write = 0;

    /* Allocate File objects */
    if (
    	(read = Allocate_File(&s_readPipeFileOps, 0, 0, 0, O_READ, 0)) == 0 ||
		(write = Allocate_File(&s_writePipeFileOps, 0, 0, 0, O_WRITE, 0)) == 0
	   )
	{
		Free_Pipe_Files(&read, &write);
		return ENOMEM;
    }

    *pRead  = read;
    *pWrite = write;
    
    (*pRead)->endPos  = 1 << 30;
    (*pWrite)->endPos = 1 << 30;

    return 0;
}

/* ----------------------------------------------------------------------
 * Public functions
 * ---------------------------------------------------------------------- */
 
/* Init global data for the pipefs (lists and stuff...) */
void Init_PIPEFS()
{
 	Clear_Pipes_List(&g_pipesList);
}

int Create_Pipe(struct File **pRead, struct File **pWrite)
{
    int rc = 0;
    
	//Make sure we don't get interrupted! otherwise access to our list might be a problem
	KASSERT ( !Interrupts_Enabled() );

    rc = Allocate_Pipe_Files(pRead, pWrite);
    if ( rc == 0 )
    {
		//Allocate a new pipe object
	    struct Pipe *pipe = Malloc(sizeof(struct Pipe));
	    if ( pipe == 0 )
	    {
			Free_Pipe_Files(pRead, pWrite);
			return ENOMEM;
	    }
	     
	    KASSERT(pipe!=0);
	     
	    Init_Pipe(pipe);
	    
	    //NOTE: we have 2 fds referring to this pipe
	    pipe->refCount = 2;
	
	    //Assign the allocated pipe
	    (*pRead)->fsData  = pipe;
	    (*pWrite)->fsData = pipe;

    	//If everything went fine -> append the pipe to the global pipes list
	    //Append pipe to the global PipesList
	    Add_To_Front_Of_Pipes_List(&g_pipesList, pipe);
	    
		#ifdef DEBUG_PIPEFS
		Print("Added pipe to pipes list\n");
		#endif
    }
    
    return rc;
}

/**
 * Create a new named pipe with name 'name. Read and write will point to 
 * the corresponding file structs.
 */
int Create_Named_Pipe(const char *name, struct File **pRead, struct File **pWrite)
{
	int rc = 0;
	
	//Make sure we don't get interrupted! otherwise access to our list might be a problem
	KASSERT ( !Interrupts_Enabled() );

	//Make sure the name does not exceed our MAX_PIPE_NAME
	if ( strlen(name) > MAX_PIPE_NAME )
	{
		#ifdef DEBUG_PIPEFS
		Print("Pipe name [%s] exceeds max size!\n", name);
		#endif

		return -1;
	}

	//Make sure we even got a name!
	if ( strlen(name) <= 0 )
	{
		#ifdef DEBUG_PIPEFS
		Print("Pipe name [%s] no length!\n", name);
		#endif

		return -1;
	}
	
	//If a pipe with the same name already exists: return it
	if ( !Is_Pipes_List_Empty(&g_pipesList) )
	{
		struct Pipe *curr = Get_Front_Of_Pipes_List(&g_pipesList);
		while ( curr!=0 )
		{
			if ( strcmp(name, curr->name) == 0 )
			{
				#ifdef DEBUG_PIPEFS
				Print("Found pipe name [%s] in list - allocating file!\n", name);
				#endif

				//Allocate files and assign the pipe
			    rc = Allocate_Pipe_Files(pRead, pWrite);
			    if ( rc == 0 )
			    {
			    	//NOTE: 2 more fds referring this pipe!!
				    curr->refCount += 2;

				    //Assign the allocated pipe
				    (*pRead)->fsData  = curr;
				    (*pWrite)->fsData = curr;
	
					#ifdef DEBUG_PIPEFS
					Print("Assigned pipe to fds for pipe [%s]\n", name);
					#endif
			    }
			    
			    return 0;
			}
			
			curr = Get_Next_In_Pipes_List(curr);
		}
	}

	//Create a new pipe
	rc = Create_Pipe(pRead, pWrite);
	
	//Creation worked correct
	if ( rc == 0 )
	{
		//Get reference to pipe (is in fsData...)
		struct Pipe *pipe = (*pRead)->fsData;
		
		KASSERT ( pipe != 0 );
		KASSERT ( pipe == (*pWrite)->fsData );
		
		//And set the name to the unnamed pipe
		pipe->name = Malloc( sizeof(char) * (strlen(name)+1) );
		if ( pipe->name == 0 )
		{
			#ifdef DEBUG_PIPEFS
			Print("Could not allocate mem for pipe name [%s]!\n", name);
			#endif

			Free_Pipe_Files(pRead, pWrite);
			return ENOMEM;
		}
		else
		{
			strcpy(pipe->name, name);
		}
	}

	return rc;
}
