/*
 * Synchronization primitives
 * Copyright (c) 2001,2004 David H. Hovemeyer <daveho@cs.umd.edu>
 * $Revision: 1.13 $
 * 
 * This is free software.  You are permitted to use,
 * redistribute, and modify it as specified in the file "COPYING".
 */

#include <geekos/kthread.h>
#include <geekos/int.h>
#include <geekos/kassert.h>
#include <geekos/screen.h>
#include <geekos/malloc.h>
#include <geekos/string.h>
#include <geekos/synch.h>
#include <geekos/user.h>

//global semaphores list
static struct Semaphores_List g_semaphoresList;

void Init_Semaphores()
{
	Clear_Semaphores_List(&g_semaphoresList);
}

/* semaphores api (used by syscalls...) */

/*
 * create a new semaphore id or returns an already
 * existing id if "semName" exists in the global semaphores list
 * - semName is the name of the semaphore
 * - nameLength is the length of the name (<MAX_SEM_NAME_LENGTH!!)
 * - initialCount is the count the semaphore should start with
 */
int Create_Semaphore(char *semName, int nameLength, int initialCount)
{
	static int lastSemaphoreId = 0;
	bool found = false;
	struct Semaphore *sem;

	//make sure this is called non-interruptibly!
	KASSERT(!Interrupts_Enabled());

	//have a look wether we already created a semphore with this name:

	//we have to iterate the list and search for the name
	sem = Get_Front_Of_Semaphores_List(&g_semaphoresList);
	while ( sem != 0 )
	{		
		if ( strcmp(sem->name, semName) == 0 )
		{
			found = true;
			break;
		}
		
		sem = Get_Next_In_Semaphores_List(sem);
	}

	//Couldn't find semaphore -> create new one
	if( !found )
	{
		//we need to setup a new semaphore
		sem = Malloc(sizeof(struct Semaphore));
		if ( sem == 0 )
		{
			Print("Can't get mem for semaphore\n");
			return -1;
		}

		//Initial data of a semaphore
		sem->id       = ++lastSemaphoreId;
		sem->count    = initialCount;
		sem->refCount = 0;

		//Copy name
		memset(sem->name, 0, sizeof(char)*(MAX_SEM_NAME_LENGTH + 1));
		strncpy(sem->name, semName, nameLength);

		//clear the semaphores queue
		Clear_Thread_Queue(&sem->queue);

		//add to global semaphores list:
		Add_To_Back_Of_Semaphores_List(&g_semaphoresList, sem);
	}
	
	
	
	//We now also use this semaphore
	sem->refCount++;
	

	//bugfix (what is if some process calls Create_Semaphore multiple times?)
	//we only return the semaphore ID but do not add the process to the list!

	//lets have a look wether the semaphore is already registered for this process!
	found = false;
	struct SemaphoreRefNode *tmpItem = Get_Front_Of_Semaphores_Ref_List(&g_currentThread->userContext->refSemaphoresList);
	while ( tmpItem != 0 )
	{
		if ( tmpItem->semaphore == sem )
		{
			found = true;
			break;
		}
		tmpItem = Get_Next_In_Semaphores_Ref_List(tmpItem);
	}

	if ( found )
	{
		return sem->id;
	}
	
	//this process did not already register this semaphore -> so lets do it now
	
	//Create reference node on semaphore
	struct SemaphoreRefNode *procSem = Malloc(sizeof(struct SemaphoreRefNode));
	if ( procSem == 0 )
	{
		Print("Can't get mem for semaphore reference\n");
		return -1;
	}

	procSem->semaphore = sem;

	//add semaphore to the process' semaphores list
	Add_To_Back_Of_Semaphores_Ref_List(
		&g_currentThread->userContext->refSemaphoresList,
		procSem
	);

	return sem->id;
}

/**
 * Get Semaphore-Reference by semaphore-id from list of current user context.
 * @return 0 if not found, result otherwise
 */
static struct SemaphoreRefNode *Get_Current_SemaphoreRefNode(int semId)
{
	struct SemaphoreRefNode *item;
	struct Semaphores_Ref_List *list = &g_currentThread->userContext->refSemaphoresList;

	//make sure this is called non-interruptibly!
	KASSERT(!Interrupts_Enabled());

	//find the semaphore with semId in the process semaphores list
	item = Get_Front_Of_Semaphores_Ref_List(list);
	while ( item != 0 )
	{
		//Finished search.
		if ( item->semaphore->id == semId )
		{			
			return item;
		}
		
		item = Get_Next_In_Semaphores_Ref_List(item);
	}
	
	
	return 0;
}

/*
 *
 * returns 0 on success <0 on error
 */
int Semaphore_Down(int semId)
{
	struct SemaphoreRefNode *refSem;
	struct Semaphore *sem;

	//make sure this is called non-interruptibly!
	KASSERT(!Interrupts_Enabled());

	refSem = Get_Current_SemaphoreRefNode(semId);	
	//we didn't find semId in the process semaphores list
	if ( refSem == 0 )
	{
		return -1;
	}

	//Get Semaphore from Reference...	
	sem = refSem->semaphore;

	//if the counter <= 0 -> we have to wait!
	while ( sem->count <= 0 )
	{
		Wait(&sem->queue);
	}

	sem->count--;

	return 1;
}

/*
 *
 * returns 0 on success <0 on error
 */
int Semaphore_Up(int semId)
{
	struct SemaphoreRefNode *procSem;
	struct Semaphore *sem;

	//make sure this is called non-interruptibly!
	KASSERT(!Interrupts_Enabled());

	procSem = Get_Current_SemaphoreRefNode(semId);
	if ( procSem == 0 )
	{
		//we didn't find semId in the process semaphores list
		return -1;
	}
	
	sem = procSem->semaphore;

	//increment semaphore counter
	sem->count++;

	//if counter > 0 another process may enter
	if ( sem->count > 0 )
	{
		Wake_Up_One(&sem->queue);
	}

	return 0;
}


/*
 *
 * returns 0 on success and
 * -1 on failure (semId not found in the process semaphores list)
 */
int Destroy_Semaphore(int semId)
{
	struct Semaphores_Ref_List *list = &g_currentThread->userContext->refSemaphoresList;
	struct SemaphoreRefNode *item;

	//make sure this is called non-interruptibly!
	KASSERT(!Interrupts_Enabled());

	item = Get_Current_SemaphoreRefNode(semId);
	if ( item != 0 )
	{
		//decrement reference counter
		item->semaphore->refCount--;
		
		//if the last reference was removed - free the allocated memory
		if ( item->semaphore->refCount == 0 )
		{
			Remove_From_Semaphores_List(&g_semaphoresList, item->semaphore);
			Free(item->semaphore);
		}

		//remove the item from the list
		Remove_From_Semaphores_Ref_List(list, item);
		
		//free the previously allocated item
		Free(item);

		return 0;
	}

	//we did not find semId in the process semaphores list
	return -1;
}

/*
 * list is a SemaphoreRefNodes_List which belongs i.e. to the UserContext
 * of each thread. this function removes all references to the semaphores
 * might be called by a cleanup function for a process
 */
void Free_Semaphores_From_List(struct Semaphores_Ref_List *list)
{
	struct SemaphoreRefNode *item, *nextItem;

	item = Get_Front_Of_Semaphores_Ref_List(list);
	while ( item != 0 )
	{
		//decrement reference counter
		item->semaphore->refCount--;
		
		//if the last reference was removed - free the allocated memory
		if ( item->semaphore->refCount == 0 )
		{
			Remove_From_Semaphores_List(&g_semaphoresList, item->semaphore);
			Free(item->semaphore);
		}

		//get next item - before we remove it from the list
		nextItem = Get_Next_In_Semaphores_Ref_List(item);

		//remove the item from the list
		Remove_From_Semaphores_Ref_List(list, item);
		
		//free the previously allocated item
		Free(item);
		
		item = nextItem;
	}
}

/*
 * NOTES:
 * - The GeekOS mutex and condition variable APIs are based on those
 *   in pthreads.
 * - Unlike disabling interrupts, mutexes offer NO protection against
 *   concurrent execution of interrupt handlers.  Mutexes and
 *   condition variables should only be used from kernel threads,
 *   with interrupts enabled.
 */

/* ----------------------------------------------------------------------
 * Private functions
 * ---------------------------------------------------------------------- */

/*
 * The mutex is currently locked.
 * Atomically reenable preemption and wait in the
 * mutex's wait queue.
 */
static void Mutex_Wait(struct Mutex *mutex)
{
    KASSERT(mutex->state == MUTEX_LOCKED);
    KASSERT(g_preemptionDisabled);

    Disable_Interrupts();
    g_preemptionDisabled = false;
    Wait(&mutex->waitQueue);
    g_preemptionDisabled = true;
    Enable_Interrupts();
}

/*
 * Lock given mutex.
 * Preemption must be disabled.
 */
static __inline__ void Mutex_Lock_Imp(struct Mutex* mutex)
{
    KASSERT(g_preemptionDisabled);

    /* Make sure we're not already holding the mutex */
    KASSERT(!IS_HELD(mutex));

    /* Wait until the mutex is in an unlocked state */
    while (mutex->state == MUTEX_LOCKED) {
	Mutex_Wait(mutex);
    }

    /* Now it's ours! */
    mutex->state = MUTEX_LOCKED;
    mutex->owner = g_currentThread;
}

/*
 * Unlock given mutex.
 * Preemption must be disabled.
 */
static __inline__ void Mutex_Unlock_Imp(struct Mutex* mutex)
{
    KASSERT(g_preemptionDisabled);

    /* Make sure mutex was actually acquired by this thread. */
    KASSERT(IS_HELD(mutex));

    /* Unlock the mutex. */
    mutex->state = MUTEX_UNLOCKED;
    mutex->owner = 0;

    /*
     * If there are threads waiting to acquire the mutex,
     * wake one of them up.  Note that it is legal to inspect
     * the queue with interrupts enabled because preemption
     * is disabled, and therefore we know that no thread can
     * concurrently add itself to the queue.
     */
    if (!Is_Thread_Queue_Empty(&mutex->waitQueue)) {
	Disable_Interrupts();
	Wake_Up_One(&mutex->waitQueue);
	Enable_Interrupts();
    }
}

/* ----------------------------------------------------------------------
 * Public functions
 * ---------------------------------------------------------------------- */

/*
 * Initialize given mutex.
 */
void Mutex_Init(struct Mutex* mutex)
{
    mutex->state = MUTEX_UNLOCKED;
    mutex->owner = 0;
    Clear_Thread_Queue(&mutex->waitQueue);
}

/*
 * Lock given mutex.
 */
void Mutex_Lock(struct Mutex* mutex)
{
    KASSERT(Interrupts_Enabled());

    g_preemptionDisabled = true;
    Mutex_Lock_Imp(mutex);
    g_preemptionDisabled = false;
}

/*
 * Unlock given mutex.
 */
void Mutex_Unlock(struct Mutex* mutex)
{
    KASSERT(Interrupts_Enabled());

    g_preemptionDisabled = true;
    Mutex_Unlock_Imp(mutex);
    g_preemptionDisabled = false;
}

/*
 * Initialize given condition.
 */
void Cond_Init(struct Condition* cond)
{
    Clear_Thread_Queue(&cond->waitQueue);
}

/*
 * Wait on given condition (protected by given mutex).
 */
void Cond_Wait(struct Condition* cond, struct Mutex* mutex)
{
    KASSERT(Interrupts_Enabled());

    /* Ensure mutex is held. */
    KASSERT(IS_HELD(mutex));

    /* Turn off scheduling. */
    g_preemptionDisabled = true;

    /*
     * Release the mutex, but leave preemption disabled.
     * No other threads will be able to run before this thread
     * is able to wait.  Therefore, this thread will not
     * miss the eventual notification on the condition.
     */
    Mutex_Unlock_Imp(mutex);

    /*
     * Atomically reenable preemption and wait in the condition wait queue.
     * Other threads can run while this thread is waiting,
     * and eventually one of them will call Cond_Signal() or Cond_Broadcast()
     * to wake up this thread.
     * On wakeup, disable preemption again.
     */
    Disable_Interrupts();
    g_preemptionDisabled = false;
    Wait(&cond->waitQueue);
    g_preemptionDisabled = true;
    Enable_Interrupts();

    /* Reacquire the mutex. */
    Mutex_Lock_Imp(mutex);

    /* Turn scheduling back on. */
    g_preemptionDisabled = false;
}

/*
 * Wake up one thread waiting on the given condition.
 * The mutex guarding the condition should be held!
 */
void Cond_Signal(struct Condition* cond)
{
    KASSERT(Interrupts_Enabled());
    Disable_Interrupts();  /* prevent scheduling */
    Wake_Up_One(&cond->waitQueue);
    Enable_Interrupts();  /* resume scheduling */
}

/*
 * Wake up all threads waiting on the given condition.
 * The mutex guarding the condition should be held!
 */
void Cond_Broadcast(struct Condition* cond)
{
    KASSERT(Interrupts_Enabled());
    Disable_Interrupts();  /* prevent scheduling */
    Wake_Up(&cond->waitQueue);
    Enable_Interrupts();  /* resume scheduling */
}
