/*
 * System call handlers
 * Copyright (c) 2003, Jeffrey K. Hollingsworth <hollings@cs.umd.edu>
 * Copyright (c) 2003,2004 David Hovemeyer <daveho@cs.umd.edu>
 * $Revision: 1.59 $
 * 
 * This is free software.  You are permitted to use,
 * redistribute, and modify it as specified in the file "COPYING".
 */

#include <geekos/syscall.h>
#include <geekos/errno.h>
#include <geekos/kthread.h>
#include <geekos/int.h>
#include <geekos/elf.h>
#include <geekos/malloc.h>
#include <geekos/screen.h>
#include <geekos/keyboard.h>
#include <geekos/string.h>
#include <geekos/user.h>
#include <geekos/timer.h>
#include <geekos/vfs.h>
#include <geekos/synch.h>
#include <geekos/pipefs.h>



//Uncomment this to print debug messages
//#define DEBUG_SYSCALL


/*
 * Null system call.
 * Does nothing except immediately return control back
 * to the interrupted user program.
 * Params:
 *  state - processor registers from user mode
 *
 * Returns:
 *   always returns the value 0 (zero)
 */
static int Sys_Null(struct Interrupt_State* state)
{
    return 0;
}

/*
 * Exit system call.
 * The interrupted user process is terminated.
 * Params:
 *   state->ebx - process exit code
 * Returns:
 *   Never returns to user mode!
 */
static int Sys_Exit(struct Interrupt_State* state)
{	
	//Exit the current thread with the specific exit code
	Exit(state->ebx);
	
	//We never get here.
	KASSERT(false);
	return 0;
}

/*
 * Print a string to the console.
 * Params:
 *   state->ebx - user pointer of string to be printed
 *   state->ecx - number of characters to print
 * Returns: 0 if successful, -1 if not
 */
static int Sys_PrintString(struct Interrupt_State* state)
{	
	//For security reasons: copy data to \0-terminated buffer:
	//User has explicit length info, maybe \0 terminator is not
	//in *ebx
	
	//Make space for the text buffer
	char* txtbuf = Malloc( sizeof(char) * (state->ecx+1));
	
	//Couldn't get enough mem
	if( txtbuf == 0 )
		return -1;
		
		
	//Set all to terminating char
	memset(txtbuf, 0, sizeof(char) * (state->ecx+1));
	//And copy data
	if( !Copy_From_User( txtbuf, state->ebx, sizeof(char) * (state->ecx)) )
	{
		Free(txtbuf);
		return -1;
	}  
	
	//Print text
	Print(txtbuf);		
	
	//Free our text buffer
	Free(txtbuf);
	
	//Success
	return 0;
}

/*
 * Get a single key press from the console.
 * Suspends the user process until a key press is available.
 * Params:
 *   state - processor registers from user mode
 * Returns: the key code
 */
static int Sys_GetKey(struct Interrupt_State* state)
{	
	//Wait for key waits, until key is available
    return (int) Wait_For_Key();
}

/*
 * Set the current text attributes.
 * Params:
 *   state->ebx - character attributes to use
 * Returns: always returns 0
 */
static int Sys_SetAttr(struct Interrupt_State* state)
{	
	//Set color and so on...	
    Set_Current_Attr(state->ebx);
    return 0;
}

/*
 * Get the current cursor position.
 * Params:
 *   state->ebx - pointer to user int where row value should be stored
 *   state->ecx - pointer to user int where column value should be stored
 * Returns: 0 if successful, -1 otherwise
 */
static int Sys_GetCursor(struct Interrupt_State* state)
{	
	int row, col;
	
	//Get current cursor
	Get_Cursor(&row, &col);
	
	//Save the cursor positions
	if( !Copy_To_User( state->ebx, &row, sizeof(int)) )
		return -1;
	if( !Copy_To_User( state->ecx, &col, sizeof(int)) )
		return -1;
	
	//Can there be an error, I don't think so
    return 0;
}

/*
 * Set the current cursor position.
 * Params:
 *   state->ebx - new row value
 *   state->ecx - new column value
 * Returns: 0 if successful, -1 otherwise
 */
static int Sys_PutCursor(struct Interrupt_State* state)
{	
	//No comment.
	return Put_Cursor(state->ebx, state->ecx) ? 0 : -1;
}

/*
 * Create a new user process.
 * Params:
 *   state->ebx - user address of name of executable
 *   state->ecx - length of executable name
 *   state->edx - user address of command string
 *   state->esi - length of command string
 *   state->edi == stdin file descriptor in low 16 bits, stdout file descriptor in high 16 bits
 * Returns: pid of process if successful, error code (< 0) otherwise
 */
static int Sys_Spawn(struct Interrupt_State* state)
{	
	//For security reasons: copy data to \0-terminated buffer:
	//User has explicit length info, maybe \0 terminator is not
	//in *ebx.
	//Same holds for edx,esi
	
	/*
     * NOTE: here is how to get the file descriptors for
     * the child's stdin and stdout.
     * int stdinFd = (int) (state->edi & 0xffff);
     * int stdoutFd = (int) (state->edi >> 16);
     */
	
	const int stdinFd = (int) (state->edi & 0xffff);
    const int stdoutFd = (int) (state->edi >> 16);
    
            
    
    //Illegal file descriptors
    if( stdinFd < 0  ||  stdoutFd < 0 )
    {
    	#ifdef DEBUG_SYSCALL
		Print("Sys_Spawn: Illegal file descriptors\n");
		#endif
		
		return EINVALID;
    }
    
    //Stdin, Stdout we want to use
    struct File* stdinOrg = g_currentThread->userContext->openFiles[stdinFd];
    struct File* stdoutOrg = g_currentThread->userContext->openFiles[stdoutFd];
    
    //Lookup file descriptors
    if( stdinOrg==0 || stdoutOrg==0)
 	{	
 		#ifdef DEBUG_SYSCALL
		Print("Sys_Spawn: Unknown file descriptor.\n");
		#endif
		
		return EINVALID;
 	}
    
		
	//Strings of spawn parameter
	char* program = Malloc( sizeof(char) * (state->ecx+1));
	char* command = Malloc( sizeof(char) * (state->esi+1));
	
	
	//Not enough space - ok then we couldn't load executable anyway.
	if( program==0 || command==0)
	{
		#ifdef DEBUG_SYSCALL
		Print("Sys_Spawn: No mem\n");
		#endif
		
		if( program != 0)
			Free(program);
		if( command != 0)
			Free(command);
					
		return ENOMEM;
	}
	
		
	//Set all to terminating chars
	memset(program, 0, sizeof(char) * (state->ecx+1));
	memset(command, 0, sizeof(char) * (state->esi+1));
		
	//Copy program string
	if( !Copy_From_User(program, state->ebx,  sizeof(char) * (state->ecx)) )
	{
		#ifdef DEBUG_SYSCALL
		Print("Sys_Spawn: copy from user\n");
		#endif
		
		Free(program);
		Free(command);
		return -1;		
	}
		
	//Copy command string
	if( !Copy_From_User(command, state->edx,  sizeof(char) * (state->esi)) )
	{
		#ifdef DEBUG_SYSCALL
		Print("Sys_Spawn: copy from user\n");
		#endif
		
		Free(program);
		Free(command);
		return -1;		
	}
	    
   
	#ifdef DEBUG_SYSCALL
	Print("Calling spawn for %s with fdin %d, fdout: %d\n", command, stdinFd, stdoutFd);
	#endif
	
		
	//Needed for mutex
	Enable_Interrupts();	
	
	struct Kernel_Thread* kthread;
	
	
	//Clone the files
	struct File* stdin;	
	struct File* stdout;
	
	
	Clone_File(stdinOrg, &stdin);
	Clone_File(stdoutOrg, &stdout);
	
		
	//Spawn thread
	int pid =  Spawn(program, command, stdin, stdout, &kthread );
	
	//Error in spawn
	if( pid < 0)
	{
		//Close the files...
		Close(stdin);
		Close(stdout);
	}
	
    //Restore state
    Disable_Interrupts();
    
    #ifdef DEBUG_SYSCALL
    Print("Ending spawn with pid %d\n", pid);
    #endif
    		
	//Free mem
	Free(program);
	Free(command);
	
	
	
	return pid;
}

/*
 * Wait for a process to exit.
 * Params:
 *   state->ebx - pid of process to wait for
 * Returns: the exit code of the process,
 *   or error code (< 0) on error
 */
static int Sys_Wait(struct Interrupt_State* state)
{	
	//Get the thread with this pid
    struct Kernel_Thread* towaitfor =  Lookup_Thread((int)state->ebx);
    
    //So such thread
    if( towaitfor == 0 )
    	return -1;
    
    //Needed for mutex	
    Enable_Interrupts();    	
    //Now, wait for him
    Join(towaitfor);
    //Restore state
    Disable_Interrupts();
    
    //Success
    return 0;    
}

/*
 * Get pid (process id) of current thread.
 * Params:
 *   state - processor registers from user mode
 * Returns: the pid of the current thread
 */
static int Sys_GetPID(struct Interrupt_State* state)
{	
	//Return our pid
    return g_currentThread->pid;
}

/*
 * Get the quantum.
 * Params:
 *   state - processor registers from user mode
 * Returns: the current quantum each process may use
 *          (i.e before being scheduled)
 */
static int Sys_GetQuantum(struct Interrupt_State* state)
{
	return g_Quantum;
}

/*
 * Get the scheduling policy.
 * Params:
 *   state - processor registers from user mode
 * Returns: the current scheduling policy
 */
static int Sys_GetSchedulingPolicy(struct Interrupt_State* state)
{
	return g_schedulingPolicy;
}

/*
 * Set the scheduling policy.
 * Params:
 *   state->ebx - policy,
 *   state->ecx - number of ticks in quantum
 * Returns: 0 if successful, -1 otherwise
 */
static int Sys_SetSchedulingPolicy(struct Interrupt_State* state)
{
	//see if we got a valid policy request

	//set the quantum:
	//just check to be greater 0 for now -> should there be an upper limit too?
	if ( state->ecx > 0 )
	{
		g_Quantum = state->ecx;
	}
	else
	{
		//should we return a SPECIAL error code? define INVALID_QUANTUM...
		return EUNSUPPORTED;
	}

	if ( state->ebx == SCHED_POL_RR )
	{
		g_schedulingPolicy = SCHED_POL_RR;
		return 0;
	}
	else if ( state->ebx == SCHED_POL_ML )
	{
		g_schedulingPolicy = SCHED_POL_ML;
		return 0;
	}
	
	return EUNSUPPORTED;
}

/*
 * Get the time of day.
 * Params:
 *   state - processor registers from user mode
 *
 * Returns: value of the g_numTicks global variable
 */
static int Sys_GetTimeOfDay(struct Interrupt_State* state)
{
	return g_numTicks;
}

/*
 * Create a semaphore.
 * Params:
 *   state->ebx - user address of name of semaphore
 *   state->ecx - length of semaphore name
 *   state->edx - initial semaphore count
 * Returns: the global semaphore id
 */
static int Sys_CreateSemaphore(struct Interrupt_State* state)
{
	//The name of the semaphore, terminated with \0
	char semaphoreName[MAX_SEM_NAME_LENGTH + 1];	
	memset(semaphoreName, 0, sizeof(char)*(MAX_SEM_NAME_LENGTH + 1));
		
	//Length of name
	const int len = state->ecx;

	//too long semaphore name...
	if ( len > MAX_SEM_NAME_LENGTH )
	{
		//too long name given - we only support MAX_SEM_NAME_LENGTH
		return -1;
	}
	
	//Copy name of semaphore
	if ( !Copy_From_User(semaphoreName, state->ebx, sizeof(char)*len) )
	{
		//could not copy -> return some invalid identifier
		return -1;
	}
	
	//see if we already registered a semaphore with this name
	return Create_Semaphore(semaphoreName, sizeof(char)*len, state->edx);
}

/*
 * Acquire a semaphore.
 * Assume that the process has permission to access the semaphore,
 * the call will block until the semaphore count is >= 0.
 * Params:
 *   state->ebx - the semaphore id
 *
 * Returns: 0 if successful, error code (< 0) if unsuccessful
 */
static int Sys_P(struct Interrupt_State* state)
{
	return Semaphore_Down(state->ebx);
}

/*
 * Release a semaphore.
 * Params:
 *   state->ebx - the semaphore id
 *
 * Returns: 0 if successful, error code (< 0) if unsuccessful
 */
static int Sys_V(struct Interrupt_State* state)
{
	return Semaphore_Up(state->ebx);
}

/*
 * Destroy a semaphore.
 * Params:
 *   state->ebx - the semaphore id
 *
 * Returns: 0 if successful, error code (< 0) if unsuccessful
 */
static int Sys_DestroySemaphore(struct Interrupt_State* state)
{
    return Destroy_Semaphore(state->ebx);
}

/*
 * Mount a filesystem.
 * Params:
 * state->ebx - contains a pointer to the Mount_Syscall_Args structure
 *   which contains the block device name, mount prefix,
 *   and filesystem type
 *
 * Returns:
 *   0 if successful, error code if unsuccessful
 */
static int Sys_Mount(struct Interrupt_State *state)
{
    int rc = 0;
    struct VFS_Mount_Request *args = 0;

    /* Allocate space for VFS_Mount_Request struct. */
    if ((args = (struct VFS_Mount_Request *) Malloc(sizeof(struct VFS_Mount_Request))) == 0) {
	rc = ENOMEM;
	goto done;
    }

    /* Copy the mount arguments structure from user space. */
    if (!Copy_From_User(args, state->ebx, sizeof(struct VFS_Mount_Request))) {
	rc = EINVALID;
	goto done;
    }

    /*
     * Hint: use devname, prefix, and fstype from the args structure
     * and invoke the Mount() VFS function.  You will need to check
     * to make sure they are correctly nul-terminated.
     */
     
    //Set terminating character
    args->devname[BLOCKDEV_MAX_NAME_LEN] = 0;
    args->prefix[VFS_MAX_PATH_LEN] = 0;
    args->fstype[VFS_MAX_FS_NAME_LEN] = 0;
     
     
	//Needed for mutex
	Enable_Interrupts();	
	
	//Mount cmd from VFS
	rc =  Mount(args->devname, args->prefix, args->fstype);
	
	Disable_Interrupts();		
	

done:
    if (args != 0) Free(args);
    return rc;
}

/*
 * Open a file.
 * Params:
 *   state->ebx - address of user string containing path of file to open
 *   state->ecx - length of path
 *   state->edx - file mode flags
 *
 * Returns: a file descriptor (>= 0) if successful,
 *   or an error code (< 0) if unsuccessful
 */
static int Sys_Open(struct Interrupt_State *state)
{
	//For security reasons: copy data to \0-terminated buffer:
	//User has explicit length info, maybe \0 terminator is not
	//in *ebx.
			
	const int namelen = state->ecx;
	const int flags = state->edx;
	
	//Length of device string not valid
	if( namelen < 0 )
	{
		#ifdef DEBUG_SYSCALL
		Print("Sys_Format: Invalid string length\n");
		#endif
			
		return -1;
	}	
		
	//Strings of spawn parameter
	char* name = Malloc( sizeof(char) * (namelen+1));
	if( name == 0)
		return ENOMEM;		
	//Set all to terminating chars
	memset(name, 0, sizeof(char) * (namelen+1));
		
		
	//Copy devname string
	if( !Copy_From_User(name, state->ebx,  sizeof(char) * (namelen)) )
	{
		#ifdef DEBUG_SYSCALL
		Print("Sys_Format: copy from user\n");
		#endif
				
		Free(name);
		return -1;		
	}	
	
	
	//Needed for mutex
	Enable_Interrupts();
	
	int ret = EMFILE;	
	
	//Try to find an empty slot in file table...
	int i;
	for( i=0; i<USER_MAX_FILES; i++)
	{
		struct File** pfile = &(g_currentThread->userContext->openFiles[i]);
		
		//If used, skip
		if( *pfile != 0 )
			continue;
			
		//Open file and break slot searching
		ret = Open(name, flags, pfile);
		
		//If open worked, return value is the fd
		if( ret == 0 )
			ret = i;
		
		#ifdef DEBUG_SYSCALL
		Print("Open call set fd %d to addr %p.\n", i, *pfile);
		#endif
		
		break;
	}
	
	Disable_Interrupts();		
	
	//Free the mem
	Free(name);
		
	return ret;
	
	
}

/*
 * Open a directory.
 * Params:
 *   state->ebx - address of user string containing path of directory to open
 *   state->ecx - length of path
 *
 * Returns: a file descriptor (>= 0) if successful,
 *   or an error code (< 0) if unsuccessful
 */
static int Sys_OpenDirectory(struct Interrupt_State *state)
{
	//For security reasons: copy data to \0-terminated buffer:
	//User has explicit length info, maybe \0 terminator is not
	//in *ebx.
	const int pathlen = state->ecx;
	
	//Length of device string not valid
	if( pathlen < 0 )
	{
		#ifdef DEBUG_SYSCALL
		Print("Sys_OpenDirectory: Invalid path length\n");
		#endif
			
		return -1;
	}	
		
	//Strings of spawn parameter
	char *path = Malloc( sizeof(char) * (pathlen+1));
	if( path == 0)
		return ENOMEM;
	//Set all to terminating chars
	memset(path, 0, sizeof(char) * (pathlen+1));


	//Copy path string
	if( !Copy_From_User(path, state->ebx,  sizeof(char) * (pathlen)) )
	{
		#ifdef DEBUG_SYSCALL
		Print("Sys_OpenDirectory: copy from user\n");
		#endif 
				
		Free(path);
		return -1;		
	}	
	
	
	//Needed for mutex
	Enable_Interrupts();
	
	int ret = EMFILE;	
	
	//Try to find an empty slot in file table...
	int i;
	for( i=0; i<USER_MAX_FILES; i++)
	{
		struct File** pdir = &(g_currentThread->userContext->openFiles[i]);
		
		//If used, skip
		if( *pdir != 0 )
			continue;

		//Open directory and break slot searching
		ret = Open_Directory(path, pdir);
		
		//If open worked, return value is the fd
		if( ret == 0 )
			ret = i;
		
		#ifdef DEBUG_SYSCALL
		Print("Open directory call set fd %d to addr %p.\n", i, *pdir);
		#endif
		
		break;
	}
	
	Disable_Interrupts();		
	
	//Free the mem
	Free(path);
		
	return ret;
}

/*
 * Close an open file or directory.
 * Params:
 *   state->ebx - file descriptor of the open file or directory
 * Returns: 0 if successful, or an error code (< 0) if unsuccessful
 */
static int Sys_Close(struct Interrupt_State *state)
{
	//The files fd which should be closed
	const int fd = state->ebx;
	
	#ifdef DEBUG_SYSCALL
	Print("Try to close fd %d\n", fd);
	#endif
	
	//Invalid file descriptor
	if( fd < 0  ||  fd >= USER_MAX_FILES )
	{
		#ifdef DEBUG_SYSCALL
		Print("Error: FD %d out of range.\n", fd);
		#endif
		
		return ENOTFOUND;
	}
	
	//The filedescriptor is not used
	if( g_currentThread->userContext->openFiles[fd] == 0 )
	{
		#ifdef DEBUG_SYSCALL
		Print("Error: FD %d not used. Addr %p\n", 
			fd, g_currentThread->userContext->openFiles[fd]);
		#endif
		
		return ENOTFOUND;
	}
	
	
	//Needed for mutex
	Enable_Interrupts();
	
	//Call the VFS close function
	int ret = Close( g_currentThread->userContext->openFiles[fd] );
	
	Disable_Interrupts();
	
	//Clear entry, no matter if close failed
	g_currentThread->userContext->openFiles[fd] = 0;
	
	return ret;	
}

/*
 * Delete a file.
 * Params:
 *   state->ebx - address of user string containing path to delete
 *   state->ecx - length of path
 *
 * Returns: 0 if successful, error code (< 0) if unsuccessful
 */
static int Sys_Delete(struct Interrupt_State *state)
{
    //For security reasons: copy data to \0-terminated buffer:
	//User has explicit length info, maybe \0 terminator is not
	//in *ebx.
			
	const int namelen = state->ecx;
	
	//Length of device string not valid
	if( namelen < 0 )
	{
		#ifdef DEBUG_SYSCALL
		Print("Sys_Format: Invalid string length\n");
		#endif
			
		return -1;
	}	
		
	//Strings of spawn parameter
	char* name = Malloc( sizeof(char) * (namelen+1));
	if( name == 0)
		return ENOMEM;		
	//Set all to terminating chars
	memset(name, 0, sizeof(char) * (namelen+1));
		
		
	//Copy devname string
	if( !Copy_From_User(name, state->ebx,  sizeof(char) * (namelen)) )
	{
		#ifdef DEBUG_SYSCALL
		Print("Sys_Format: copy from user\n");
		#endif
				
		Free(name);
		return -1;		
	}	
	
	
	//Needed for mutex
	Enable_Interrupts();
	//Call delete function
	int ret = Delete(name);
	Disable_Interrupts();
	
	return ret;
}

/*
 * Read from an open file.
 * Params:
 *   state->ebx - file descriptor to read from
 *   state->ecx - user address of buffer to read into
 *   state->edx - number of bytes to read
 *
 * Returns: number of bytes read, 0 if end of file,
 *   or error code (< 0) on error
 */
static int Sys_Read(struct Interrupt_State *state)
{
    const int fd = state->ebx;
    int numbytes = state->edx;
    
    #ifdef DEBUG_SYSCALL
    Print("Try to read from file, %d bytes\n", numbytes);
    #endif
    
    //Invalid file descriptor
	if( fd < 0  ||  fd >= USER_MAX_FILES )
	{
		#ifdef DEBUG_SYSCALL
		Print("Error: FD %d out of range.\n", fd);
		#endif
		
		return ENOTFOUND;
	}
	
	//The filedescriptor is not used
	if( g_currentThread->userContext->openFiles[fd] == 0 )
	{
		#ifdef DEBUG_SYSCALL
		Print("Error: FD %d not used. Addr %p\n", 
			fd, g_currentThread->userContext->openFiles[fd]);
		#endif
			
		return ENOTFOUND;
	}
	
	//Invalid size of buffer
	if( numbytes < 0 )
	{
		#ifdef DEBUG_SYSCALL
		Print("Error: Negative buffer size %d\n", numbytes);
		#endif
		
		return ENOTFOUND;
	}
	
	
	
	//Nothing to do...
	if( numbytes == 0)
		return 0;
	
	//Get mem for buffer
	void* buffer = Malloc( numbytes);
	
	
	//No free mem
	if( buffer == 0)
	{
		#ifdef DEBUG_SYSCALL
		Print("Error: No free mem for buffer of size %d.\n", numbytes);
		#endif
		
		return ENOMEM;
	}
	


	//The file structure
	struct File* file = g_currentThread->userContext->openFiles[fd];
	
	
	//Truncate number of bytes so that at maximum to file-size is read
	numbytes = MIN(numbytes, file->endPos - file->filePos);
	
	#ifdef DEBUG_SYSCALL
	Print("Files size: %ld, pos: %ld\n", file->endPos, file->filePos);
	#endif
	
	//for mutex
	Enable_Interrupts();	
	//Call the VFS read function
	int ret = Read( file, buffer, numbytes );	
	Disable_Interrupts();
	
	//Set file pos to new value
	if( ret >= 0 )
		file->filePos += ret;
		
	
	//Copy buffer to user from kernel
	if( !Copy_To_User(state->ecx, buffer, numbytes) )
	{
		#ifdef DEBUG_SYSCALL
		Print("Error: Copying data from user to kernel failed.\n");
		#endif
		
		Free(buffer);
		return -1;	
	}
	
	//Free the buffer
	Free(buffer);
	
	return ret;
}

/*
 * Read a directory entry from an open directory handle.
 * Params:
 *   state->ebx - file descriptor of the directory
 *   state->ecx - user address of struct VFS_Dir_Entry to copy entry into
 * Returns: 0 if successful, error code (< 0) if unsuccessful
 */
static int Sys_ReadEntry(struct Interrupt_State *state)
{
	//Make sure the passed file descriptor has been opened by the process!
    const int fd = state->ebx;

    //Invalid file descriptor
	if( fd < 0  ||  fd >= USER_MAX_FILES )
	{
		#ifdef DEBUG_SYSCALL
		Print("Error: FD %d out of range.\n", fd);
		#endif
		
		return ENOTFOUND;
	}
	
	//The filedescriptor is not used
	if ( g_currentThread->userContext->openFiles[fd] == 0 )
	{
		#ifdef DEBUG_SYSCALL
		Print("Error: FD %d not used. Addr %p\n", 
			fd, g_currentThread->userContext->openFiles[fd]);
		#endif
			
		return ENOTFOUND;
	}

	struct VFS_Dir_Entry entry;
	memset(&entry, 0, sizeof(struct VFS_Dir_Entry));
	
	//We need interrupts for mutex
	Enable_Interrupts();
	int ret = Read_Entry(g_currentThread->userContext->openFiles[fd], &entry);	
	Disable_Interrupts();
	
	if ( ret == 0 )
	{
		//increment the filePos after a successful read
		g_currentThread->userContext->openFiles[fd]->filePos++;

		//Copy the result into the user space
		if ( !Copy_To_User(state->ecx, &entry, sizeof(struct VFS_Dir_Entry)) )
		{
			#ifdef DEBUG_SYSCALL
			Print("Sys_ReadEntry: Error[Could not copy result to user]\n");
			#endif
			return ENOMEM;
		}
	}

	return ret;
}

/*
 * Write to an open file.
 * Params:
 *   state->ebx - file descriptor to write to
 *   state->ecx - user address of buffer get data to write from
 *   state->edx - number of bytes to write
 *
 * Returns: number of bytes written,
 *   or error code (< 0) on error
 */
static int Sys_Write(struct Interrupt_State *state)
{
    const int fd = state->ebx;
    const int numbytes = state->edx;
       
    #ifdef DEBUG_SYSCALL     
    Print("Try to write to file\n");
    #endif
    	  
    
    //Invalid file descriptor
	if( fd < 0  ||  fd >= USER_MAX_FILES )
	{
		#ifdef DEBUG_SYSCALL
		Print("Error: FD %d out of range.\n", fd);
		#endif
		
		return ENOTFOUND;
	}
	
	//The filedescriptor is not used
	if( g_currentThread->userContext->openFiles[fd] == 0 )
	{
		#ifdef DEBUG_SYSCALL
		Print("Error: FD %d not used. Addr %p\n", 
			fd, g_currentThread->userContext->openFiles[fd]);
		#endif	
			
		return ENOTFOUND;
	}
	
	//Invalid size of buffer
	if( numbytes < 0 )
	{
		#ifdef DEBUG_SYSCALL
		Print("Error: Negative buffer size %d\n", numbytes);
		#endif
		
		return ENOTFOUND;
	}
	
		
	//Get mem for buffer
	void* buffer = Malloc( numbytes);
	
	//No free mem
	if( buffer == 0)
	{
		#ifdef DEBUG_SYSCALL
		Print("Error: No free mem for buffer of size %d.\n", numbytes);
		#endif
		
		return ENOMEM;
	}
	
	//Copy buffer from user to kernel
	if( !Copy_From_User(buffer, state->ecx, numbytes) )
	{	
		#ifdef DEBUG_SYSCALL
		Print("Error: Copying data from user to kernel failed.\n");
		#endif
		
		Free(buffer);
		return -1;	
	}
	
		
	//The file we modify
	struct File* file = g_currentThread->userContext->openFiles[fd];
		
	//For Mutex
	Enable_Interrupts();		
	//Call the VFS write function
	int ret = Write( g_currentThread->userContext->openFiles[fd], buffer, numbytes );	
	Disable_Interrupts();
	
	
	//Set file pos to new value
	if( ret >= 0 )
	{
		file->filePos += ret;
		file->endPos = MAX( file->endPos, file->filePos);
	}
		
	
	//Free the buffer
	Free(buffer);
	
	return ret;
	
}

/*
 * Get file metadata.
 * Params:
 *   state->ebx - address of user string containing path of file
 *   state->ecx - length of path
 *   state->edx - user address of struct VFS_File_Stat object to store metadata in
 *
 * Returns: 0 if successful, error code (< 0) if unsuccessful
 */
static int Sys_Stat(struct Interrupt_State *state)
{
    //For security reasons: copy data to \0-terminated buffer:
	//User has explicit length info, maybe \0 terminator is not
	//in *ebx.
			
	const int namelen = state->ecx;
	
	//Length of device string not valid
	if( namelen < 0 )
	{
		#ifdef DEBUG_SYSCALL
		Print("Sys_Stat: Invalid string length\n");
		#endif
			
		return -1;
	}	
		
	//Strings of spawn parameter
	char* name = Malloc( sizeof(char) * (namelen+1));
	if( name == 0 )
		return ENOMEM;		
	//Set all to terminating chars
	memset(name, 0, sizeof(char) * (namelen+1));
		
		
	//Copy devname string
	if( !Copy_From_User(name, state->ebx,  sizeof(char) * (namelen)) )
	{
		#ifdef DEBUG_SYSCALL
		Print("Sys_Stat: Error while copy from user\n");
		#endif
				
		Free(name);
		return -1;		
	}	
	
	
	//The stat structure which will be filled
	struct VFS_File_Stat stat;
	
	
	//Needed for mutex
	Enable_Interrupts();
	//Call the stat
	int ret = Stat(name, &stat);	
	Disable_Interrupts();
	
	
	//Copy stat to user
	if( !Copy_To_User(state->edx, &stat, sizeof(struct VFS_File_Stat)) )
	{
		#ifdef DEBUG_SYSCALL
		Print("Sys_Stat: Error while copy to user\n");
		#endif
		
		Free(name);
		return -1;
	}		
	
	//Free the mem
	Free(name);
		
	return ret;
	
}

/*
 * Get metadata of an open file.
 * Params:
 *   state->ebx - file descriptor to get metadata for
 *   state->ecx - user address of struct VFS_File_Stat object to store metadata in
 *
 * Returns: 0 if successful, error code (< 0) if unsuccessful
 */
static int Sys_FStat(struct Interrupt_State *state)
{
    const int fd = state->ebx;
    
    #ifdef DEBUG_SYSCALL
    Print("Try to fstat\n");
    #endif
    
    //Invalid file descriptor
	if( fd < 0  ||  fd >= USER_MAX_FILES )
	{
		#ifdef DEBUG_SYSCALL
		Print("Error: FD %d out of range.\n", fd);
		#endif
		
		return ENOTFOUND;
	}
	
	//The filedescriptor is not used
	if( g_currentThread->userContext->openFiles[fd] == 0 )
	{
		#ifdef DEBUG_SYSCALL
		Print("Error: FD %d not used. Addr %p\n", 
			fd, g_currentThread->userContext->openFiles[fd]);
		#endif
			
		return ENOTFOUND;
	}
	
	
	
	//The stat structure which will be filled
	struct VFS_File_Stat stat;
		
	//Call the VFS FSTAT function
	int ret = FStat( g_currentThread->userContext->openFiles[fd], &stat);
	
	//Error while fstatc
	if( ret < 0 )
	{
		#ifdef DEBUG_SYSCALL
		Print("Error: In FStat impl\n");
		#endif
		
		return ret;
	}
	
	
	//Copy stat to user
	if( !Copy_To_User(state->ecx, &stat, sizeof(struct VFS_File_Stat)) )
		return -1;
	
	return 0;
}

/*
 * Change the access position in a file
 * Params:
 *   state->ebx - file descriptor 
 *   state->ecx - position in file
 *
 * Returns: 0 if successful, error code (< 0) if unsuccessful
 */
static int Sys_Seek(struct Interrupt_State *state)
{
	const int fd = state->ebx;
	const int pos = state->ecx;

    #ifdef DEBUG_SYSCALL
    Print("Try to seek in file\n");
    #endif
    
    //Invalid file descriptor
	if( fd < 0  ||  fd >= USER_MAX_FILES )
	{
		#ifdef DEBUG_SYSCALL
		Print("Error: FD %d out of range.\n", fd);
		#endif
		
		return ENOTFOUND;
	}
	
	//Invalid position
	if( pos < 0 )
	{
		#ifdef DEBUG_SYSCALL
		Print("Error: Invalid position %d.\n", pos);
		#endif
		
		return EINVALID;
	}
	
	//The filedescriptor is not used
	if( g_currentThread->userContext->openFiles[fd] == 0 )
	{
		#ifdef DEBUG_SYSCALL
		Print("Error: FD %d not used. Addr %p\n", 
			fd, g_currentThread->userContext->openFiles[fd]);
		#endif
			
		return ENOTFOUND;
	}	
	
	
	//For mutex
	Enable_Interrupts();		
	//Call the VFS seek function
	int ret = Seek( g_currentThread->userContext->openFiles[fd], pos );	
	Disable_Interrupts();
	
	
	//Error while fstatc
	if( ret < 0 )
	{
		#ifdef DEBUG_SYSCALL
		Print("Error: In seek impl\n");
		#endif
		
		return ret;
	}

	return 0;
}

/*
 * Create directory
 * Params:
 *   state->ebx - address of user string containing path of new directory
 *   state->ecx - length of path
 *
 * Returns: 0 if successful, error code (< 0) if unsuccessful
 */
static int Sys_CreateDir(struct Interrupt_State *state)
{
	char *path = 0;
    const int pathlength = state->ecx;

  	path = Malloc( sizeof(char) * (pathlength+1) );
  	if ( path == 0 )
  	{
  		#ifdef DEBUG_SYSCALL
  		Print("Sys_CreateDir: could not allocate path\n");
  		#endif
  		
  		return ENOMEM;
  	}
  	
  	memset(path, 0,  sizeof(char) * (pathlength+1) );

  	if ( !Copy_From_User(path, state->ebx,  sizeof(char) * (pathlength+1) ) )
  	{
  		#ifdef DEBUG_SYSCALL
  		Print("Sys_CreateDir: cpy from user failed - path could not be copied\n");
  		#endif
  		
  		Free(path);
  		return ENOMEM;
  	}
  		
	//Needed for mutex
	Enable_Interrupts();
	
  	int ret = Create_Directory(path);
  	Free(path);
  	
	//Needed for mutex
	Disable_Interrupts();
  	
  	return ret;
}

/*
 * Flush filesystem buffers
 * Params: none 
 * Returns: 0 if successful, error code (< 0) if unsuccessful
 */
static int Sys_Sync(struct Interrupt_State *state)
{
	//Just call sync
    return Sync();
}
/*
 * Format a device
 * Params:
 *   state->ebx - address of user string containing device to format
 *   state->ecx - length of device name string
 *   state->edx - address of user string containing filesystem type 
 *   state->esi - length of filesystem type string

 * Returns: 0 if successful, error code (< 0) if unsuccessful
 */
static int Sys_Format(struct Interrupt_State *state)
{
	//For security reasons: copy data to \0-terminated buffer:
	//User has explicit length info, maybe \0 terminator is not
	//in *ebx.
	
		
	const int devnamelen = state->ecx;
	const int fstypelen = state->esi;
	
	//Length of device string not valid
	if( devnamelen < 0 || fstypelen < 0 )
	{
		#ifdef DEBUG_SYSCALL
		Print("Sys_Format: Invalid string length\n");
		#endif
			
		return -1;
	}	
		
	//Strings of spawn parameter
	char* devname = Malloc( sizeof(char) * (devnamelen+1));
	char* fstype = Malloc( sizeof(char) * (fstypelen+1));
	if( fstype == 0  || devname==0)
		return ENOMEM;		
		
	//Set all to terminating chars
	memset(devname, 0, sizeof(char) * (devnamelen+1));
	memset(fstype, 0, sizeof(char) * (fstypelen+1));
	
	
		
	//Copy devname string
	if( !Copy_From_User(devname, state->ebx,  sizeof(char) * (devnamelen)) )
	{
		#ifdef DEBUG_SYSCALL
		Print("Sys_Format: copy from user\n");
		#endif
				
		Free(devname);
		Free(fstype);
		return -1;		
	}	
	//Copy fstype string
	if( !Copy_From_User(fstype, state->edx,  sizeof(char) * (fstypelen)) )
	{
		#ifdef DEBUG_SYSCALL
		Print("Sys_Format: copy from user\n");
		#endif
				
		Free(devname);
		Free(fstype);
		return -1;		
	}
	
	
	//Needed for mutex
	Enable_Interrupts();	
	
	//Format corresponding strings
	int ret =  Format(devname, fstype);
	
	Disable_Interrupts();		
	
	//Free the mem
	Free(devname);
	Free(fstype);
		
	return ret;
}

/*
 * Create a pipe
 * Params:
 *   state->ebx - address of read file descriptor
 *   state->ecx - address of write file desciptor

 * Returns: 0 if successful, error code (< 0) if unsuccessful
 */
static int Sys_CreatePipe(struct Interrupt_State *state)
{
	struct File **fdRead=0, **fdWrite=0;
	
	
	//Try to find an empty slot in file table for the read fd...
	int readFdIdx = 0;
	for( readFdIdx=0; readFdIdx<USER_MAX_FILES; readFdIdx++)
	{
		struct File** pfile = &(g_currentThread->userContext->openFiles[readFdIdx]);
		
		//If used, skip
		if( *pfile != 0 )
			continue;

		//Slot is unused -> use it...
		fdRead = pfile;
		break;
	}
	
	//No empy slot was found - descriptor table is full
	if ( fdRead == 0 )
		return EMFILE;

	
	

	//Try to find an empty slot in file table for the writer fd...
	int writeFdIdx = 0;
	for( writeFdIdx=0; writeFdIdx<USER_MAX_FILES; writeFdIdx++)
	{
		struct File** pfile = &(g_currentThread->userContext->openFiles[writeFdIdx]);
		
		//If used, skip
		if( *pfile != 0 || writeFdIdx==readFdIdx )
			continue;

		//Slot is unused -> use it...
		fdWrite = pfile;
		break;
	}
	
	//No empy slot was found - descriptor table is full
	if ( fdWrite == 0 )		
		return EMFILE;

	//Initialize the read/write fds 
	int rc = Create_Pipe(fdRead, fdWrite);
	if ( rc < 0 )
	{
		//Garuantee that fds are zero
		*fdRead = 0;
		*fdWrite = 0;
		return rc;
	}
	
	//Copy the result fds back to the user
	if ( !Copy_To_User(state->ebx, &readFdIdx, sizeof(int)) )
	{
		Close(*fdRead);
		Close(*fdWrite);
		return ENOMEM;
	}

	//Copy the result fds back to the user
	if ( !Copy_To_User(state->ecx, &writeFdIdx, sizeof(int)) )
	{
		Close(*fdRead);
		Close(*fdWrite);
		return ENOMEM;
	}

    return rc;
}

/*
 * Create a named pipe
 * Params:
 *   state->ebx - name of pipe
 *   state->ecx - length of pipe name
 *   state->edx - address of read file descriptor
 *   state->esi - address of write file desciptor

 * Returns: 0 if successful, error code (< 0) if unsuccessful
 */
static int Sys_CreateNamedPipe(struct Interrupt_State *state)
{
	const int namelen = state->ecx;
	
	if ( namelen <= 0 )
	{
		#ifdef DEBUG_SYSCALL
		Print("Sys_CreateNamedPipe: Invalid pipe name length\n");
		#endif

		return -1;
	}
	
	char *name = Malloc( sizeof(char) * (namelen+1) );
	if ( name == 0 )
	{
		#ifdef DEBUG_SYSCALL
		Print("Sys_CreateNamedPipe: Could not allocate memory for pipe name\n");
		#endif

		return ENOMEM;
	}
	
	//Set all to terminating chars
	memset(name, 0, sizeof(char) * (namelen+1));
	
	//Copy name from user space
	if ( !Copy_From_User(name, state->ebx, sizeof(char) * (namelen)) )
	{
		#ifdef DEBUG_SYSCALL
		Print("Sys_CreateNamedPipe: Could not copy pipe name from user space\n");
		#endif
		
		Free(name);

		return ENOMEM;
	}
	
	//Same code as CraetPipe
	struct File **fdRead=0, **fdWrite=0;
	
	
	//Try to find an empty slot in file table for the read fd...
	int readFdIdx = 0;
	for( readFdIdx=0; readFdIdx<USER_MAX_FILES; readFdIdx++)
	{
		struct File** pfile = &(g_currentThread->userContext->openFiles[readFdIdx]);
		
		//If used, skip
		if( *pfile != 0 )
			continue;

		//Slot is unused -> use it...
		fdRead = pfile;
		break;
	}
	
	//No empy slot was found - descriptor table is full
	if ( fdRead == 0 )
		return EMFILE;

	
	//Try to find an empty slot in file table for the writer fd...
	int writeFdIdx = 0;
	for( writeFdIdx=0; writeFdIdx<USER_MAX_FILES; writeFdIdx++)
	{
		struct File** pfile = &(g_currentThread->userContext->openFiles[writeFdIdx]);
		
		//If used or same slot as read fd, skip
		if( *pfile != 0 || writeFdIdx==readFdIdx )
			continue;

		//Slot is unused -> use it...
		fdWrite = pfile;
		break;
	}
	
	//No empy slot was found - descriptor table is full
	if ( fdWrite == 0 )		
		return EMFILE;

	//Initialize the read/write fds 
	int rc = Create_Named_Pipe(name, fdRead, fdWrite);
	if ( rc < 0 )
	{
		//Garuantee that fds are zero
		*fdRead  = 0;
		*fdWrite = 0;
		
		#ifdef DEBUG_SYSCALL
		Print("Sys_CreateNamedPipe: Create_Named_Pipe failed\n");
		#endif

		return rc;
	}
	
	//Copy the result fds back to the user
	if ( !Copy_To_User(state->edx, &readFdIdx, sizeof(int)) )
	{
		#ifdef DEBUG_SYSCALL
		Print("Sys_CreateNamedPipe: Copy_To_User for readFd failed\n");
		#endif

		Close(*fdRead);
		Close(*fdWrite);
		return ENOMEM;
	}

	//Copy the result fds back to the user
	if ( !Copy_To_User(state->esi, &writeFdIdx, sizeof(int)) )
	{
		#ifdef DEBUG_SYSCALL
		Print("Sys_CreateNamedPipe: Copy_To_User for writeFd failed\n");
		#endif

		Close(*fdRead);
		Close(*fdWrite);
		return ENOMEM;
	}

    return rc;
}

/*
 * Global table of system call handler functions.
 */
const Syscall g_syscallTable[] = {
    Sys_Null,
    Sys_Exit,
    Sys_PrintString,
    Sys_GetKey,
    Sys_SetAttr,
    Sys_GetCursor,
    Sys_PutCursor,
    Sys_Spawn,
    Sys_Wait,
    Sys_GetPID,
    /* Scheduling and semaphore system calls. */
	Sys_GetQuantum,
	Sys_GetSchedulingPolicy,
    Sys_SetSchedulingPolicy,
    Sys_GetTimeOfDay,
    Sys_CreateSemaphore,
    Sys_P,
    Sys_V,
    Sys_DestroySemaphore,
    /* File I/O system calls. */
    Sys_Mount,
    Sys_Open,
    Sys_OpenDirectory,
    Sys_Close,
    Sys_Delete,
    Sys_Read,
    Sys_ReadEntry,
    Sys_Write,
    Sys_Stat,
    Sys_FStat,
    Sys_Seek,
    Sys_CreateDir,
    Sys_Sync,
    Sys_Format,
    /* Pipe system calls. */
    Sys_CreatePipe,
    Sys_CreateNamedPipe,
};

/*
 * Number of system calls implemented.
 */
const int g_numSyscalls = sizeof(g_syscallTable) / sizeof(Syscall);
