/*
 * Common user mode functions
 * Copyright (c) 2001,2003,2004 David H. Hovemeyer <daveho@cs.umd.edu>
 * $Revision: 1.50 $
 * 
 * This is free software.  You are permitted to use,
 * redistribute, and modify it as specified in the file "COPYING".
 */

#include <geekos/errno.h>
#include <geekos/ktypes.h>
#include <geekos/kassert.h>
#include <geekos/int.h>
#include <geekos/mem.h>
#include <geekos/malloc.h>
#include <geekos/kthread.h>
#include <geekos/vfs.h>
#include <geekos/tss.h>
#include <geekos/user.h>

//Uncomment this to print debug messages
//#define DEBUG_USER

/*
 * This module contains common functions for implementation of user
 * mode processes.
 */

/*
 * Associate the given user context with a kernel thread.
 * This makes the thread a user process.
 */
void Attach_User_Context(struct Kernel_Thread* kthread, struct User_Context* context)
{
    KASSERT(context != 0);
    kthread->userContext = context;

    Disable_Interrupts();

    /*
     * We don't actually allow multiple threads
     * to share a user context (yet)
     */
    KASSERT(context->refCount == 0);

    ++context->refCount;
    Enable_Interrupts();
}

/*
 * If the given thread has a user context, detach it
 * and destroy it.  This is called when a thread is
 * being destroyed.
 */
void Detach_User_Context(struct Kernel_Thread* kthread)
{
    struct User_Context* old = kthread->userContext;

    kthread->userContext = 0;

    if (old != 0) {
		int refCount;
	
		Disable_Interrupts();
			--old->refCount;
		refCount = old->refCount;
		Enable_Interrupts();
	
		/*Print("User context refcount == %d\n", refCount);*/
			if (refCount == 0)
			{
				Destroy_User_Context(old);
			}
    }
}

/*
 * Spawn a user process.
 * Params:
 *   program - the full path of the program executable file
 *   command - the command, including name of program and arguments
 *   stdInput - File to be Cloned as the new process's stdin
 *   stdOutput - File to be Cloned as the new process's stdout
 *   pThread - reference to Kernel_Thread pointer where a pointer to
 *     the newly created user mode thread (process) should be
 *     stored. If 0, this will be skipped.
 * Returns:
 *   The process id (pid) of the new process, or an error code
 *   if the process couldn't be created.  Note that this function
 *   should return ENOTFOUND if the reason for failure is that
 *   the executable file doesn't exist.
 */
int Spawn(const char *program, const char *command,
    struct File *stdInput, struct File *stdOutput,
    struct Kernel_Thread **pThread)
{
    /*
     * Hints:
     * - Call Read_Fully() to load the entire executable into a memory buffer
     * - Call Parse_ELF_Executable() to verify that the executable is
     *   valid, and to populate an Exe_Format data structure describing
     *   how the executable should be loaded
     * - Call Load_User_Program() to create a User_Context with the loaded
     *   program
     * - Call Start_User_Thread() with the new User_Context
     *
     * If all goes well, store the pointer to the new thread in
     * pThread and return 0.  Otherwise, return an error code.
     */
          
	//Return codes
	int ret;
    
    //Exe-file content 
	char* exeFileData = 0;
	//Length of exe-file
	ulong_t exeFileLength = 0;
	//Structure of ELF file after parsing
	struct Exe_Format exeFormat;

	
#ifdef DEBUG_USER
	Print("Spawning from thread %p with cmd: %s\n", g_currentThread, command); 
#endif
	
     
    //Load program data in exefiledata with length exefilelength
    ret = Read_Fully(program, (void**) &exeFileData, &exeFileLength);
	if( ret != 0)
	{
		#ifdef DEBUG_USER
		Print("Read_Fully failed to read %s from disk [%d]\n", program, ret);
		#endif
		
		//Only free if existant!
		if(exeFileData)
			Free(exeFileData);	 
		return ret;
    }
       
#ifdef DEBUG_USER
	Print("\tParse ELF file.\n"); 
#endif 
        
    //Parsing ELF binary data
    ret = Parse_ELF_Executable(exeFileData, exeFileLength, &exeFormat);
	if( ret != 0 )
	{
		Print("Parse_ELF_Executable failed: %d\n", ret);
		
		Free(exeFileData);	 
		return ENOEXEC;
    }
    
    //The user context
    struct User_Context* pUserContext=0;

   
#ifdef DEBUG_USER
	Print("\tLoad user context and copy data.\n"); 
#endif 
    
    //Load the usrContext, set up virtual mem of usrContext and
    //copy exe-file data to mem.
    ret = Load_User_Program(exeFileData, exeFileLength, &exeFormat, command, &pUserContext);    
    if( ret != 0 )
    {
    	#ifdef DEBUG_USER
    	Print("Load_User_Program failed: %d\n", ret);
    	#endif
    	
    	Free(exeFileData);
    	return ret;
    }
    
    
    //Exe-file content unused
	Free(exeFileData);
	
	
	
	//Set stdin and stdout file descr.
	KASSERT( USER_MAX_FILES >= 2);
	pUserContext->openFiles[USER_STDIN_FD] = stdInput;
	pUserContext->openFiles[USER_STDOUT_FD] = stdOutput;
	
	
        
#ifdef DEBUG_USER
	Print("\tStart User Thread.\n"); 
#endif 
    
    //Start the kernel thread, standalone
    *pThread = Start_User_Thread(pUserContext, 0);    
    if( *pThread == 0)
    {
    	Print("Start_User_Thread failed.\n");

    	Destroy_User_Context(pUserContext);
    	Free(exeFileData);
    	return -1;
    }
            

#ifdef DEBUG_USER
	Print("Spawned thread %p from with cmd: %s\n", *pThread, command); 
#endif

	
    //Everything went ok, return pid
	if( *pThread != 0)
		return (*pThread)->pid;
		
    return -1;
}

/*
 * If the given thread has a User_Context,
 * switch to its memory space.
 *
 * Params:
 *   kthread - the thread that is about to execute
 *   state - saved processor registers describing the state when
 *      the thread was interrupted
 */
void Switch_To_User_Context(struct Kernel_Thread* kthread, struct Interrupt_State* state)
{
    /*
     * Hint: Before executing in user mode, you will need to call
     * the Set_Kernel_Stack_Pointer() and Switch_To_Address_Space()
     * functions.
     */
     
     
    if( kthread->userContext == 0 )
    	return;
    				 
	
	//Set esp to user processes stack pointer
	//The kernel stack pointer is not the one in our user space
	Set_Kernel_Stack_Pointer((ulong_t)kthread->stackPage + PAGE_SIZE);	   
	
	//Switch to user context
	Switch_To_Address_Space(kthread->userContext);
}
