/*
 * Segmentation-based user mode implementation
 * Copyright (c) 2001,2003 David H. Hovemeyer <daveho@cs.umd.edu>
 * $Revision: 1.23 $
 * 
 * This is free software.  You are permitted to use,
 * redistribute, and modify it as specified in the file "COPYING".
 */

#include <geekos/ktypes.h>
#include <geekos/kassert.h>
#include <geekos/defs.h>
#include <geekos/mem.h>
#include <geekos/string.h>
#include <geekos/malloc.h>
#include <geekos/int.h>
#include <geekos/gdt.h>
#include <geekos/segment.h>
#include <geekos/tss.h>
#include <geekos/kthread.h>
#include <geekos/argblock.h>
#include <geekos/user.h>
#include <geekos/errno.h>
#include <geekos/synch.h>



//Uncomment this to print debug messages
//#define DEBUG_USERSEG


/* ----------------------------------------------------------------------
 * Variables
 * ---------------------------------------------------------------------- */

#define DEFAULT_USER_STACK_SIZE 8192


/* ----------------------------------------------------------------------
 * Private functions
 * ---------------------------------------------------------------------- */


/*
 * Create a new user context of given virtual memory size
 * @return 0 for error, otherwise the pointer to user context
 */
static struct User_Context* Create_User_Context(ulong_t size)
{	
	//Get mem for the user context structure
	struct User_Context* usrContext = Malloc( sizeof(struct User_Context) );
	//Space for virtual mem of context
	void* memory = Malloc(size);
	
	
	//If there is no space for one of both...
	if(memory==0 || usrContext==0)
	{
		//...free the other
		if(memory!=0)
			Free(memory);
			
		//...free the other
		if(usrContext!=0)
			Free(usrContext);
			
		return 0;		
	}
	
	//For secure reasons: init all to zero
	memset(usrContext, 0, sizeof(struct User_Context));
	memset(memory, 0, size);

	Clear_Semaphores_Ref_List(&userContext->refSemaphoresList);

	//Set virtual mem of user context
	usrContext->memory = memory;
	usrContext->size = size;	
	
	return usrContext;	
}


static bool Validate_User_Memory(struct User_Context* userContext,
    ulong_t userAddr, ulong_t bufSize)
{
    ulong_t avail;

    if (userAddr >= userContext->size)
        return false;

    avail = userContext->size - userAddr;
    if (bufSize > avail)
        return false;

    return true;
}

/* ----------------------------------------------------------------------
 * Public functions
 * ---------------------------------------------------------------------- */

/*
 * Destroy a User_Context object, including all memory
 * and other resources allocated within it.
 */
void Destroy_User_Context(struct User_Context* userContext)
{
    /*
     * Hints:
     * - you need to free the memory allocated for the user process
     * - don't forget to free the segment descriptor allocated
     *   for the process's LDT
     */
     
     //Delete user space mem
     Free( userContext->memory );
     //Free the segment descriptor
     Free_Segment_Descriptor( userContext->ldtDescriptor );

	 //Free semaphores this process might refer to
	 Free_Semaphores_From_List(&userContext->refSemaphoresList);

     //Destroy the user context object itself
     //ATTENTION: Should we free userContext now!?
     Free( userContext );
    
}

/*
 * Load a user executable into memory by creating a User_Context
 * data structure.
 * Params:
 * exeFileData - a buffer containing the executable to load
 * exeFileLength - number of bytes in exeFileData
 * exeFormat - parsed ELF segment information describing how to
 *   load the executable's text and data segments, and the
 *   code entry point address
 * command - string containing the complete command to be executed:
 *   this should be used to create the argument block for the
 *   process
 * pUserContext - reference to the pointer where the User_Context
 *   should be stored
 *
 * Returns:
 *   0 if successful, or an error code (< 0) if unsuccessful
 */
int Load_User_Program(char *exeFileData, ulong_t exeFileLength,
    struct Exe_Format *exeFormat, const char *command,
    struct User_Context **pUserContext)
{
    /*
     * Hints:
     * - Determine where in memory each executable segment will be placed
     * - Determine size of argument block and where in memory it will
     *   be placed
     * - Copy each executable segment into memory
     * - Format argument block in memory
     * - In the created User_Context object, set code entry point
     *   address, argument block address, and initial kernel stack pointer
     *   address
     */
     
	
	//Maximum virtual address required
	ulong_t maxaddr = 0;
	int i;
	
	
	//Determine the maximum virtual address in all segements
	for(i=0; i<exeFormat->numSegments; i++)
	{
		//Max. virtual address in segment
		const ulong_t segmaxaddr = exeFormat->segmentList[i].startAddress + 
						exeFormat->segmentList[i].sizeInMemory;
		
		//Recalc new max virtual addr
		if(maxaddr < segmaxaddr)
				maxaddr = segmaxaddr;
	} 
	
	
	//Get size and number of arguments
	ulong_t argblocksize;
	unsigned numargs;
	Get_Argument_Block_Size(command, &numargs, &argblocksize);
	
		
	//Round to next page-size-modulo memory size
	maxaddr = Round_Up_To_Page(maxaddr);
	//Make space for stack
	maxaddr += DEFAULT_USER_STACK_SIZE;
	//Make additional space for arguments
	maxaddr += argblocksize;
	
#ifdef DEBUG_USERSEG
	Print("\tGet 0x%lx bytes for user space\n", maxaddr);
#endif
	
	//Make User context
	*pUserContext = Create_User_Context(maxaddr);	
	//Uups, out of mem
	if(*pUserContext == 0)
		return ENOMEM;
		


	//Argument block is located above the stack	
	(*pUserContext)->argBlockAddr = maxaddr-argblocksize;		
	//The stack is right below the argument block. A push decrements esp and after that
	//the value is saved at esp. So we do not need to decrement argBlockAddr.
	(*pUserContext)->stackPointerAddr = maxaddr-argblocksize;	
	
	//Set entry point in user mem
	(*pUserContext)->entryAddr = exeFormat->entryAddr;

#ifdef DEBUG_USERSEG
	Print("\t%s has %d arguments\n", command, numargs);
	Print("\tArgument block at 0x%lx + %ld.\n", (*pUserContext)->argBlockAddr, argblocksize);
	Print("\tUserspace is located at %p.\n", (*pUserContext)->memory);
	Print("\tStack begins at %p.\n", (void*) (*pUserContext)->stackPointerAddr);
	Print("\tEntry point at %p, phy: %p.\n", (void*) (*pUserContext)->entryAddr, 
			(*pUserContext)->memory+(*pUserContext)->entryAddr);
#endif
	
	//Save argument block in user space mem
	Format_Argument_Block((*pUserContext)->memory+(*pUserContext)->argBlockAddr, numargs,
			(*pUserContext)->argBlockAddr, command);
			
			
	//Copy segements to virtual memory of context
	for(i=0; i<exeFormat->numSegments; i++)
	{
		//The segment to load
		struct Exe_Segment* seg = &exeFormat->segmentList[i];

#ifdef DEBUG_USERSEG
		Print("\tCopy segment with 0x%lx Bytes to 0x%lx.\n",  seg->lengthInFile, seg->startAddress);
#endif

		//Copy data of process
		memcpy( (*pUserContext)->memory + seg->startAddress,
				exeFileData + seg->offsetInFile,
				seg->lengthInFile);
	} 
	
	
	
	//There must be at least two entries in LDT, code and data
	KASSERT( NUM_USER_LDT_ENTRIES >= 2);	
	
	//Create a segment descriptor (in the GDT) which describes the segment
	//where the LDT is located. See intel docs, sysprog-a, p101
	(*pUserContext)->ldtDescriptor = Allocate_Segment_Descriptor();
	//GDT full
	if((*pUserContext)->ldtDescriptor  == 0)
	{
		Print("No free space in GDT.\n");
		return ENOMEM;
	}	
	
	//Create segment descriptor of the LDT-segment in GDT
	//It seems clear to me that this segment must have kernel privileges
	//otherwise user could move his segment to modify kernel space or
	//other users data/program.
	(*pUserContext)->ldtSelector =  Selector(KERNEL_PRIVILEGE, true, 
				Get_Descriptor_Index((*pUserContext)->ldtDescriptor));

#ifdef DEBUG_USERSEG
	Print("Got LDT selector 0x%lx for LDT table at %p\n", (unsigned long) (*pUserContext)->ldtSelector, (*pUserContext)->ldt);
#endif
			
	//LDT lies in LDT-segment, set members appropriately
	Init_LDT_Descriptor((*pUserContext)->ldtDescriptor,
			(*pUserContext)->ldt,
			 NUM_USER_LDT_ENTRIES);
			
	//Descriptors of code and data segment in LDT
	struct Segment_Descriptor* codeDesc = &(*pUserContext)->ldt[0];
	struct Segment_Descriptor* dataDesc = &(*pUserContext)->ldt[1];
	
	
	//Clear this two structures
	memset(codeDesc, 0, sizeof(struct Segment_Descriptor));
	memset(dataDesc, 0, sizeof(struct Segment_Descriptor));	
	
	//Get the text and data segment for initializing segments in memory.
	//We align data and code segment at the base of user space mem and both
	//are as big as the whole user space mem is.  
	for(i=0; i<exeFormat->numSegments; i++)
	{
		//Current segment
		struct Exe_Segment* exeseg = &exeFormat->segmentList[i];
						
		
		//Executeable segment in ELF file --> we found code segment
		if((exeseg->protFlags & PF_X) != 0)
		{	
			//For security reasons: Code segment is only where text segment is (Starting from 0, of course)
			//So a jmp or call can never jump out of this segment 
			const int pagenmb = Round_Up_To_Page(exeseg->startAddress + exeseg->sizeInMemory) >> PAGE_POWER;
							
			Init_Code_Segment_Descriptor(codeDesc,
				(ulong_t)(*pUserContext)->memory,
				pagenmb,
				USER_PRIVILEGE);

#ifdef DEBUG_USERSEG
			Print("Init code segment at %p with %d pages.\n", (*pUserContext)->memory, pagenmb);
#endif
		}
		//Nonexecuteable segment in ELF file --> we found text segment
		else
		{
			//Data segment goes over the whole user space memory. It also includes the stack
			//And the argument block. This is needed when the main function of user-space-progs
			//gets the command line options and so on from space over the stack.
			const int pagenmb = Round_Up_To_Page(maxaddr) >> PAGE_POWER;
			
			Init_Data_Segment_Descriptor(dataDesc,
				(ulong_t)(*pUserContext)->memory, 
				pagenmb,
				USER_PRIVILEGE);

#ifdef DEBUG_USERSEG
			Print("Init data segment at %p with %d pages.\n", (*pUserContext)->memory, pagenmb);
#endif
		}		
	} 
	
	//One of both was not set!
	if( codeDesc->type == 0 || dataDesc->type == 0)
	{
		Print("Code or data segment not initialized.\n");
		return ENOEXEC;
	} 
	
	
	//Create segment descriptor of code and data segment
	(*pUserContext)->csSelector =  Selector(USER_PRIVILEGE, false, 0);
	(*pUserContext)->dsSelector =  Selector(USER_PRIVILEGE, false, 1);	

#ifdef DEBUG_USERSEG
	Print("Created code and data selector: 0x%lx, 0x%lx\n",
			(unsigned long) (*pUserContext)->csSelector, (unsigned long) (*pUserContext)->dsSelector);
#endif

    //Everything went ok
    return 0;
}

/*
 * Copy data from user memory into a kernel buffer.
 * Params:
 * destInKernel - address of kernel buffer
 * srcInUser - address of user buffer
 * bufSize - number of bytes to copy
 *
 * Returns:
 *   true if successful, false if user buffer is invalid (i.e.,
 *   doesn't correspond to memory the process has a right to
 *   access)
 */
bool Copy_From_User(void* destInKernel, ulong_t srcInUser, ulong_t bufSize)
{
    /*
     * Hints:
     * - the User_Context of the current process can be found
     *   from g_currentThread->userContext
     * - the user address is an index relative to the chunk
     *   of memory you allocated for it
     * - make sure the user buffer lies entirely in memory belonging
     *   to the process
     */
     
    
     
    //Is enough space for this chunk of mem
	if( Validate_User_Memory(g_currentThread->userContext, srcInUser, bufSize) )	
		//Only copy data, finished
		memcpy(destInKernel, (void*) (g_currentThread->userContext->memory + srcInUser), bufSize);
	
	else	
	{		
    	Print("Error Copy from user:: kernel: %p, srcinuser: %lx, size: %lx\n", destInKernel, srcInUser, bufSize); 
		return false;
	}
	
	
	return true;    
}

/*
 * Copy data from kernel memory into a user buffer.
 * Params:
 * destInUser - address of user buffer
 * srcInKernel - address of kernel buffer
 * bufSize - number of bytes to copy
 *
 * Returns:
 *   true if successful, false if user buffer is invalid (i.e.,
 *   doesn't correspond to memory the process has a right to
 *   access)
 */
bool Copy_To_User(ulong_t destInUser, void* srcInKernel, ulong_t bufSize)
{
    /*
     * Hints: same as for Copy_From_User()
     */
     
    //Is enough space for this chunk of mem
	if( Validate_User_Memory(g_currentThread->userContext, destInUser, bufSize) )	
		//Only copy data, finished
		memcpy((void*) (g_currentThread->userContext->memory + destInUser), srcInKernel, bufSize);
	
	else	
	{		
    	Print("Error Copy to user:: kernel: %p, srcinuser: 0x%lx, size: %lx\n", srcInKernel, destInUser, bufSize); 
		return false;
	}
	
	
	return true;   
}

/*
 * Switch to user address space belonging to given
 * User_Context object.
 * Params:
 * userContext - the User_Context
 */
void Switch_To_Address_Space(struct User_Context *userContext)
{
    /*
     * Hint: you will need to use the lldt assembly language instruction
     * to load the process's LDT by specifying its LDT selector.
     */
     
          
	//I think this is all we have to do: At a rescedule in lowlevel.asm
	//the seg-descriptors are loaded via RestoreRegisters. The stack pointer
	//is reset to the new thread which should be run and therefore his
	//registers (including the seg-descr-registers) are loaded.
	//We only have to guarantee that the ldt is set correctly, so that
	//restoring seg-desc-registers from stack works proper.
	Load_LDTR((unsigned long int)userContext->ldtSelector);	
}

