/*
 * Paging-based user mode implementation
 * Copyright (c) 2003,2004 David H. Hovemeyer <daveho@cs.umd.edu>
 * $Revision: 1.50 $
 * 
 * This is free software.  You are permitted to use,
 * redistribute, and modify it as specified in the file "COPYING".
 */

#include <geekos/int.h>
#include <geekos/mem.h>
#include <geekos/gdt.h>
#include <geekos/paging.h>
#include <geekos/malloc.h>
#include <geekos/string.h>
#include <geekos/argblock.h>
#include <geekos/kthread.h>
#include <geekos/range.h>
#include <geekos/vfs.h>
#include <geekos/user.h>
#include <geekos/errno.h>

/* ----------------------------------------------------------------------
 * Private functions
 * ---------------------------------------------------------------------- */
 
//Uncomment this to enable debug messages for this file
//#define DEBUG_USERVM
 
 
static struct User_Context* Create_User_Context();
static bool Init_User_Context_Memory(struct User_Context* context, char *exeFileData, ulong_t exeFileLength,struct Exe_Format *exeFormat);
static bool Copy_Page(pde_t* dir, void* virtDest, void* src, unsigned size);
 

// TODO: Add private functions
/* ----------------------------------------------------------------------
 * Public functions
 * ---------------------------------------------------------------------- */

/*
 * Destroy a User_Context object, including all memory
 * and other resources allocated within it.
 */
void Destroy_User_Context(struct User_Context* context)
{
    /*
     * Hints:
     * - Free all pages, page tables, and page directory for
     *   the process (interrupts must be disabled while you do this,
     *   otherwise those pages could be stolen by other processes)
     * - Free semaphores, files, and other resources used
     *   by the process
     */
     
	KASSERT( context != 0 );
	
	#ifdef DEBUG_USERVM
	Print("Desroying user account\n");
	#endif		

	//Close all open files
	int i;
	for(i=0; i<USER_MAX_FILES; i++)
	{
		//Current file to check if open
		struct File* file = context->openFiles[i];
		
		//File exists -- close it
		if( file != 0 )
		{				
			#ifdef DEBUG_USERVM
			Print("\tClean file %d\n", i);
			#endif		
			
			//Call the VFS close function
			int ret = Close( file );
			
			//Error while closing
			if( ret != 0)
			{
				#ifdef DEBUG_USERVM
				Print("Error closing file %d, %s\n", i);
				#endif 
			}
		}
		
		//Mark as closed
		context->openFiles[i] = 0;
	}
	
	
	//Start CS
	bool iflag = Begin_Int_Atomic();
	
	//Free semaphores this process might refer to
	Free_Semaphores_From_List(&context->refSemaphoresList);
	
	
	if( context->pageDir != 0)
	{	
		//Starting directory index
		unsigned direidx = (USER_MEM_BASE >> PAGE_POWER) / NUM_PAGE_DIR_ENTRIES;
		KASSERT( (USER_MEM_BASE >> PAGE_POWER) % NUM_PAGE_DIR_ENTRIES == 0);
			
		//Run through directory indices and free everything
		for(; direidx < NUM_PAGE_DIR_ENTRIES; direidx++)
		{
			//Page table does not exist -- skip
			if( context->pageDir[direidx].pageTableBaseAddr == 0)
				continue;
				
			//So table must be present
			KASSERT( context->pageDir[direidx].present == 1 );
				
			//Get page table
			pte_t* table = (pte_t*) (context->pageDir[direidx].pageTableBaseAddr << PAGE_POWER);
			
			
			//Run through page table and free all pages
			unsigned tabeidx;
			for( tabeidx=0; tabeidx < NUM_PAGE_TABLE_ENTRIES; tabeidx++)
			{
				//Page not allocated -- skip
				if( table[tabeidx].pageBaseAddr == 0)
					continue;
				
				//Page is on disc	
				if( table[tabeidx].kernelInfo & KINFO_PAGE_ON_DISK )
					//Free the page frame on swap file
					Free_Space_On_Paging_File(table[tabeidx].pageBaseAddr);
					
				//Page is not on disc
				else
				{
					KASSERT( table[tabeidx].present == 1 );
					//Free the page in RAM
					Free_Page( (void*) (table[tabeidx].pageBaseAddr << PAGE_POWER) );
				}
			}
			
			
			//Free the page table itself
			Free_Page(table);
		}
	      
	 
	 	//Free page of page directory    
	    Free_Page( context->pageDir );
	} 	
          
     
    //Free the segment descriptor
    if( context->ldtDescriptor != 0)
    	Free_Segment_Descriptor( context->ldtDescriptor );
    	

    //Destroy the user context object itself
    //ATTENTION: Should we free userContext now!?
    Free( context );
     
    //End CS
    End_Int_Atomic(iflag);
}

/*
 * Load a user executable into memory by creating a User_Context
 * data structure.
 * Params:
 * exeFileData - a buffer containing the executable to load
 * exeFileLength - number of bytes in exeFileData
 * exeFormat - parsed ELF segment information describing how to
 *   load the executable's text and data segments, and the
 *   code entry point address
 * command - string containing the complete command to be executed:
 *   this should be used to create the argument block for the
 *   process
 * pUserContext - reference to the pointer where the User_Context
 *   should be stored
 *
 * Returns:
 *   0 if successful, or an error code (< 0) if unsuccessful
 */
int Load_User_Program(char *exeFileData, ulong_t exeFileLength,
    struct Exe_Format *exeFormat, const char *command,
    struct User_Context **pUserContext)
{
    /*
     * Hints:
     * - This will be similar to the same function in userseg.c
     * - Determine space requirements for code, data, argument block,
     *   and stack
     * - Allocate pages for above, map them into user address
     *   space (allocating page directory and page tables as needed)
     * - Fill in initial stack pointer, argument block address,
     *   and code entry point fields in User_Context
     */
    
    
    
	//Update clock of pages --> used that our pages are not 
	//swapped out immediately
	Update_Clock(); 
     
	//Create user context, page tables and pages are created on demand later
	(*pUserContext) = Create_User_Context();	
	if( *pUserContext == 0)
	{
		Print("No free space for user context struct!\n");
		return ENOMEM;
	}			
	
	#ifdef DEBUG_USERVM
	Print("Created user context %p\n", *pUserContext);
	#endif	     		
	
	
	
	//Setup argument block page and stack
	/////////////////////////////////////	
	
	//Get size and number of arguments
	ulong_t argblocksize;
	unsigned numargs;
	Get_Argument_Block_Size(command, &numargs, &argblocksize);
	if( argblocksize >= PAGE_SIZE )
	{
		Print("Argblock does not fit in a page.\n");
		Destroy_User_Context(*pUserContext);
		return ENOMEM;
	}
	
	//Get mem for argblock
	#ifdef DEBUG_USERVM
	Print("Get 0x%lx bytes for argument block\n", argblocksize);
	#endif
	
	char* argblock = Malloc(argblocksize);
	KASSERT(argblock != 0);	
	//Save argument block in tmp memory
	Format_Argument_Block(argblock, numargs, USER_ARGBLOCK_BASE-USER_MEM_BASE, command);
	
	//Copy argument block to user mem
	bool ret = Copy_Page((*pUserContext)->pageDir, (void*)(USER_ARGBLOCK_BASE), 
							(void*)(argblock), argblocksize);
							
	Free(argblock);	
							
	//No mem available for argblock
	if( ret == false)
	{
		#ifdef DEBUG_USERVM
		Print("No free page for argument block\n");
		#endif
		Destroy_User_Context(*pUserContext);
		return ENOMEM;
	}				

	
	//A empty stack
	char stack[1];
	stack[0] = '\0';
	
	//Make stack page
	ret = Copy_Page((*pUserContext)->pageDir, (void*)(USER_STACK_BASE), (void*)(stack), 1);
	
	//No mem available for stack
	if( ret == false)
	{
		#ifdef DEBUG_USERVM
		Print("No free space for stack\n");
		#endif
		Destroy_User_Context(*pUserContext);
		return ENOMEM;
	}
	
	
	
	
				
	//Copy exe file data into user context
	//////////////////////////////////////
	
	if( !Init_User_Context_Memory((*pUserContext), exeFileData, exeFileLength, exeFormat) )
	{
		#ifdef DEBUG_USERVM
		Print("No free space for exe data\n");
		#endif
		Destroy_User_Context(*pUserContext);
		return ENOMEM;
	}								
	
	
	
	//Setup user context struct
	///////////////////////////	

	//Set addresses of stack and argument block	
	(*pUserContext)->argBlockAddr = USER_ARGBLOCK_BASE-USER_MEM_BASE;	
	(*pUserContext)->stackPointerAddr = USER_STACK_BASE-USER_MEM_BASE + PAGE_SIZE-1;		
	//Set entry point in user mem
	(*pUserContext)->entryAddr = exeFormat->entryAddr;
	
	#ifdef DEBUG_USERVM
	Print("Loaded user program:\n");
	Print("\tArgumentblock at 0x%lx\n", (*pUserContext)->argBlockAddr);
	Print("\tStack at 0x%lx\n", (*pUserContext)->stackPointerAddr);
	#endif
	
	
	
	
	//Setup LDT and descriptors
	////////////////////////////	
	
	
	//There must be at least two entries in LDT, code and data
	KASSERT( NUM_USER_LDT_ENTRIES >= 2);	
	
	//Create a segment descriptor (in the GDT) which describes the segment
	//where the LDT is located. See intel docs, sysprog-a, p101
	(*pUserContext)->ldtDescriptor = Allocate_Segment_Descriptor();
	//GDT full
	if((*pUserContext)->ldtDescriptor  == 0)
	{
		Print("No free space in GDT.\n");
		Destroy_User_Context(*pUserContext);
		return ENOMEM;
	}	
	
	//Create segment descriptor of the LDT-segment in GDT
	//It seems clear to me that this segment must have kernel privileges
	//otherwise user could move his segment to modify kernel space or
	//other users data/program.
	(*pUserContext)->ldtSelector =  Selector(KERNEL_PRIVILEGE, true, 
				Get_Descriptor_Index((*pUserContext)->ldtDescriptor));

	#ifdef DEBUG_USERVM
	Print("Got LDT selector 0x%lx for LDT table at %p\n", (unsigned long) (*pUserContext)->ldtSelector, (*pUserContext)->ldt);
	#endif
			
	//LDT lies in LDT-segment, set members appropriately
	Init_LDT_Descriptor((*pUserContext)->ldtDescriptor,
			(*pUserContext)->ldt,
			 NUM_USER_LDT_ENTRIES);
			
	//Descriptors of code and data segment in LDT
	struct Segment_Descriptor* codeDesc = &(*pUserContext)->ldt[0];
	struct Segment_Descriptor* dataDesc = &(*pUserContext)->ldt[1];
	
	
	//Clear this two structures
	memset(codeDesc, 0, sizeof(struct Segment_Descriptor));
	memset(dataDesc, 0, sizeof(struct Segment_Descriptor));	
	
	//Get the text and data segment for initializing segments in memory.
	//We align data and code segment at the base of user space mem.
	int i;
	for(i=0; i<exeFormat->numSegments; i++)
	{
		//Current segment
		struct Exe_Segment* exeseg = &exeFormat->segmentList[i];
						
		
		//Executeable segment in ELF file --> we found code segment
		if((exeseg->protFlags & PF_X) != 0)
		{	
			//For security reasons: Code segment is only where text segment is (Starting from 0, of course)
			//So a jmp or call can never jump out of this segment 
			const int pagenmb = Round_Up_To_Page(exeseg->startAddress + exeseg->sizeInMemory) >> PAGE_POWER;
							
			Init_Code_Segment_Descriptor(codeDesc,
				USER_MEM_BASE,
				pagenmb,
				USER_PRIVILEGE);

			#ifdef DEBUG_USERVM
			Print("Init code segment at %p with 0x%x pages.\n", (void*) USER_MEM_BASE, pagenmb);
			#endif
		}
		//Nonexecuteable segment in ELF file --> we found text segment
		else
		{
			//Data segment goes over the whole user space memory. It also includes the stack
			//And the argument block. This is needed when the main function of user-space-progs
			//gets the command line options and so on from space over the stack.
			const int pagenmb = ((unsigned)0-USER_MEM_BASE) >> PAGE_POWER;
			
			Init_Data_Segment_Descriptor(dataDesc,
				USER_MEM_BASE, 
				pagenmb,
				USER_PRIVILEGE);

			#ifdef DEBUG_USERVM
			Print("Init data segment at %p with 0x%x pages.\n", (void*)USER_MEM_BASE, pagenmb);
			#endif
		}		
	} 
	
	//One of both was not set!
	if( codeDesc->type == 0 || dataDesc->type == 0)
	{
		Print("Code or data segment not initialized.\n");
		Destroy_User_Context(*pUserContext);
		return ENOEXEC;
	} 
	
	
	//Create segment descriptor of code and data segment
	(*pUserContext)->csSelector =  Selector(USER_PRIVILEGE, false, 0);
	(*pUserContext)->dsSelector =  Selector(USER_PRIVILEGE, false, 1);	

	#ifdef DEBUG_USERVM
	Print("Created code and data selector: 0x%lx, 0x%lx\n",
			(unsigned long) (*pUserContext)->csSelector, (unsigned long) (*pUserContext)->dsSelector);
	#endif

    //Everything went ok
    return 0;
}

/*
 * Copy data from user buffer into kernel buffer.
 * Returns true if successful, false otherwise.
 */
bool Copy_From_User(void* destInKernel, ulong_t srcInUser, ulong_t numBytes)
{
    /*
     * Hints:
     * - Make sure that user page is part of a valid region
     *   of memory
     * - Remember that you need to add 0x80000000 to user addresses
     *   to convert them to kernel addresses, because of how the
     *   user code and data segments are defined
     * - User pages may need to be paged in from disk before being accessed.
     * - Before you touch (read or write) any data in a user
     *   page, **disable the PAGE_PAGEABLE bit**.
     *
     * Be very careful with race conditions in reading a page from disk.
     * Kernel code must always assume that if the struct Page for
     * a page of memory has the PAGE_PAGEABLE bit set,
     * IT CAN BE STOLEN AT ANY TIME.  The only exception is if
     * interrupts are disabled; because no other process can run,
     * the page is guaranteed not to be stolen.
     */
     
    
    //Add the base of user mem.
	srcInUser += USER_MEM_BASE;

	
	//Check if srcInUser is in user space (could be violated if
	//srcInUser has an overflow
	if( srcInUser < USER_MEM_BASE )
	{
		Print(" Invalid memory access: User address in kernel space!\n");
		KASSERT(0);
		return false;
	}
	
	//Check if srcInUser is in user space (could be violated if
	//srcInUser has an overflow
	if( srcInUser+numBytes < USER_MEM_BASE )
	{
		Print(" Invalid memory access: User address in kernel space!\n");
		KASSERT(0);
		return false;
	}
	
	//Check if kernel address is in user space
	if( (unsigned) destInKernel >= USER_MEM_BASE )
	{
		Print(" Invalid memory access: Kernel address in user space!\n");
		KASSERT(0);
		return false;
	}
	
	
	
	//Get page alligned address of src in user space
	const unsigned srcPageAddr = srcInUser & ~PAGE_MASK;
	//A address counter which is always page alligned
	unsigned srcAddr;
	
	//Indicate an invalid access
	bool invalidAccess = false;
	
	
	//Guarantee that all pages of user process are in RAM
	//Clear the pageable bit so that the page does not get lost by other process
	//when copying from it
	for( srcAddr=srcPageAddr; 0<srcAddr && srcAddr < srcInUser+numBytes; srcAddr+=PAGE_SIZE)
	{
		//Indices for directory entry and table entry
		const unsigned direIdx = ((unsigned)srcAddr >> PAGE_POWER) / NUM_PAGE_DIR_ENTRIES;
		const unsigned tabeIdx = ((unsigned)srcAddr >> PAGE_POWER) % NUM_PAGE_DIR_ENTRIES;

		
		//Get directory entry of page table
		pde_t* dire = &g_currentThread->userContext->pageDir[direIdx];
		
		//Table is not present --> report error
		if( dire->present == 0)
		{
			invalidAccess = true;
			break;
		}
		
				
		//Get page table and table entry
		pte_t* table = (pte_t*) (dire->pageTableBaseAddr << PAGE_POWER);
		pte_t* tabe = &table[tabeIdx];
		
		//Page is not present --> report error
		if( tabe->present == 0)
		{
			invalidAccess = true;
			break;
		}
		
		// Page is on disc
		if( tabe->kernelInfo & KINFO_PAGE_ON_DISK )
		{
			//QUESTION: This page must also be available later
			//sooo: loading page must be in CS, or page must
			//get locked or something like that. But I guess
			//giving that in CS would lock the kernel...
			#ifdef DEBUG_USERVM
			Print("Page is on disc - get it from there!\n");
			#endif
			
			Bring_In_Page( tabe->pageBaseAddr << PAGE_POWER );
		}
		
		
		//Begin critical section
		bool iflag = Begin_Int_Atomic();
		
		//Get page by (phy.) base addr
		struct Page* page = Get_Page( tabe->pageBaseAddr << PAGE_POWER );
		//Clear pageable bit
		page->flags &= ~PAGE_PAGEABLE;		
		
		//End critical section
		End_Int_Atomic(iflag);		
	}
	
	
	//Everything went good - copy data
	if( !invalidAccess )
		memcpy(destInKernel, (void*)srcInUser, numBytes);
	
	
	
	//Reset the pageable bit of all pages in user space
	for( srcAddr=srcPageAddr;  0<srcAddr && srcAddr < srcInUser+numBytes; srcAddr+=PAGE_SIZE)
	{
		//Indices for directory entry and table entry
		const unsigned direIdx = ((unsigned)srcAddr >> PAGE_POWER) / NUM_PAGE_DIR_ENTRIES;
		const unsigned tabeIdx = ((unsigned)srcAddr >> PAGE_POWER) % NUM_PAGE_DIR_ENTRIES;
		
		
		//Get directory entry of page table
		pde_t* dire = &g_currentThread->userContext->pageDir[direIdx];
		
		//There must be the table -- we accessed it just before!
		KASSERT( invalidAccess || dire->present == 1 );		
		if( dire->present == 0)
			continue;
			
					
		//Get page table and table entry
		pte_t* table = (pte_t*) (dire->pageTableBaseAddr << PAGE_POWER);
		pte_t* tabe = &table[tabeIdx];
		
		
		//Page is not present -- we accessed it just before!
		KASSERT( invalidAccess || tabe->present == 1 );
		if( tabe->present == 0)
			continue;
		
				
		//Begin critical section
		bool iflag = Begin_Int_Atomic();
		
		//Get page by base addr
		struct Page* page = Get_Page( tabe->pageBaseAddr  << PAGE_POWER );
		//Clear pageable bit
		page->flags |= PAGE_PAGEABLE;		
		
		//End critical section
		End_Int_Atomic(iflag);
	}
     
        
         
    return !invalidAccess;
}

/*
 * Copy data from kernel buffer into user buffer.
 * Returns true if successful, false otherwise.
 */
bool Copy_To_User(ulong_t destInUser, void* srcInKernel, ulong_t numBytes)
{
    /*
     * Hints:
     * - Same as for Copy_From_User()
     * - Also, make sure the memory is mapped into the user
     *   address space with write permission enabled
     */
     
	
     
    //Add the base of user mem.
	destInUser += USER_MEM_BASE;
		
	//Check if srcInUser is in user space (could be violated if
	//srcInUser has an overflow
	if( destInUser < USER_MEM_BASE )
	{
		Print(" Invalid memory access: User address in kernel space!\n");
		KASSERT(0);
		return false;
	}
	
	//Check if srcInUser is in user space (could be violated if
	//srcInUser has an overflow
	if( destInUser+numBytes < USER_MEM_BASE )
	{
		Print(" Invalid memory access: User address in kernel space!\n");
		KASSERT(0);
		return false;
	}
	
	//Check if kernel address is in user space
	if( (unsigned) srcInKernel >= USER_MEM_BASE )
	{
		Print(" Invalid memory access: Kernel address in user space!\n");
		KASSERT(0);
		return false;
	}
	
	
	
	//Get page alligned address of src in user space
	const unsigned destPageAddr = destInUser & ~PAGE_MASK;
	//A address counter which is always page alligned
	unsigned destAddr;
	
	//Indicate an invalid access
	bool invalidAccess = false;
	
	
	//Guarantee that all pages of user process are in RAM
	//Clear the pageable bit so that the page does not get lost by other process
	//when copying from it
	for( destAddr=destPageAddr; 0<destAddr && destAddr < destInUser+numBytes; destAddr+=PAGE_SIZE)
	{
		//Indices for directory entry and table entry
		const unsigned direIdx = ((unsigned)destAddr >> PAGE_POWER) / NUM_PAGE_DIR_ENTRIES;
		const unsigned tabeIdx = ((unsigned)destAddr >> PAGE_POWER) % NUM_PAGE_DIR_ENTRIES;
		
		//Get directory entry of page table
		pde_t* dire = &g_currentThread->userContext->pageDir[direIdx];
		
		//Table is not present --> report error
		if( dire->present == 0)
		{
			#ifdef DEBUG_USERVM
			Print("Error: Table not present: %d, destAddr: 0x%x\n", tabeIdx, destAddr);
			#endif
			invalidAccess = true;
			break;
		}
		
		//Get page table and table entry
		pte_t* table = (pte_t*) (dire->pageTableBaseAddr << PAGE_POWER);
		pte_t* tabe = &table[tabeIdx];
		
		//Page is not present --> report error
		if( tabe->present == 0)
		{
			#ifdef DEBUG_USERVM
			Print("Error: Table entry not present: %d, destAddr: 0x%x\n", tabeIdx, destAddr);
			#endif
			invalidAccess = true;
			break;
		}
		
		
		//Begin critical section
		bool iflag = Begin_Int_Atomic();
		
		// Page is on disc
		while( tabe->kernelInfo & KINFO_PAGE_ON_DISK )
		{
			//QUESTION: This page must also be available later
			//sooo: loading page must be in CS, or page must
			//get locked or something like that. But I guess
			//giving that in CS would lock the kernel...
			#ifdef DEBUG_USERVM
			Print("Page is on disc - get it from there!\n");
			#endif
			
			Bring_In_Page( tabe->pageBaseAddr << PAGE_POWER );
		}
		
		KASSERT( !Interrupts_Enabled() );				
		//Page must be in RAM now
		KASSERT( (tabe->kernelInfo & KINFO_PAGE_ON_DISK) == 0 );
		
		//Get page by base addr
		struct Page* page = Get_Page( tabe->pageBaseAddr  << PAGE_POWER );
		//Clear pageable bit
		page->flags &= ~PAGE_PAGEABLE;		
		
		//End critical section
		End_Int_Atomic(iflag);		
	}
	
	
	//Everything went good - copy data
	if( !invalidAccess )
		memcpy((void*)destInUser, srcInKernel, numBytes);	
	
	
	//Reset the pageable bit of all pages in user space
	for( destAddr=destPageAddr; 0<destAddr && destAddr < destInUser+numBytes; destAddr+=PAGE_SIZE)
	{
		//Indices for directory entry and table entry
		const unsigned direIdx = ((unsigned)destAddr >> PAGE_POWER) / NUM_PAGE_DIR_ENTRIES;
		const unsigned tabeIdx = ((unsigned)destAddr >> PAGE_POWER) % NUM_PAGE_DIR_ENTRIES;
		
		//Get directory entry of page table
		pde_t* dire = &g_currentThread->userContext->pageDir[direIdx];
		
		//There must be the table -- we accessed it just before!
		KASSERT( invalidAccess || dire->present == 1 );		
		if( dire->present == 0)
			continue;
		
		//Get page table and table entry
		pte_t* table = (pte_t*) (dire->pageTableBaseAddr << PAGE_POWER);
		pte_t* tabe = &table[tabeIdx];
		
		//Page is not present -- we accessed it just before!
		KASSERT( invalidAccess || tabe->present == 1 );
		if( tabe->present == 0)
			continue;
		
		//Begin critical section
		bool iflag = Begin_Int_Atomic();
		
		//Get page by base addr
		struct Page* page = Get_Page( tabe->pageBaseAddr  << PAGE_POWER );
		//Clear pageable bit
		page->flags |= PAGE_PAGEABLE;		
		
		//End critical section
		End_Int_Atomic(iflag);		
	}
     
         
    return !invalidAccess;
}

/*
 * Switch to user address space.
 */
void Switch_To_Address_Space(struct User_Context *userContext)
{
    /*
     * - If you are still using an LDT to define your user code and data
     *   segments, switch to the process's LDT
     * - 
     */
     
    KASSERT( userContext!=0 );
     
	//Set the paging directory if this user process
	KASSERT( userContext->pageDir != 0);
	Set_PDBR( userContext->pageDir );
          
	//I think this is all we have to do: At a rescedule in lowlevel.asm
	//the seg-descriptors are loaded via RestoreRegisters. The stack pointer
	//is reset to the new thread which should be run and therefore his
	//registers (including the seg-descr-registers) are loaded.
	//We only have to guarantee that the ldt is set correctly, so that
	//restoring seg-desc-registers from stack works proper.
	Load_LDTR((unsigned long int)userContext->ldtSelector);	
}


/*
 * Create a new user context of given virtual memory size. This means
 * creating a user context object, the page directory
 * @param size The size of memory which is needed from 0x80000000 (.text, .data and argblock)
 * @return 0 for error, otherwise the pointer to user context
 */
struct User_Context* Create_User_Context()
{	
	//Get mem for the user context structure
	struct User_Context* usrContext = Malloc( sizeof(struct User_Context) );
		
	//If there is no space for one of both...
	if(usrContext==0)			
		return 0;	
	//For secure reasons: init all to zero
	memset(usrContext, 0, sizeof(struct User_Context));
	
	
	//Reset semaphore reference list for this user context
	Clear_Semaphores_Ref_List(& usrContext->refSemaphoresList);
	
	//Allocated page for the page directory. We use a non-pageable page
	//Because page directory should not be swapped
	usrContext->pageDir = Alloc_Page();
	if( usrContext->pageDir == 0)
		return 0;
	
		
	//Reset all open file slots to unused
	int i;
	for(i=0; i<USER_MAX_FILES; i++)
		usrContext->openFiles[i] = 0;
		
	
	//Copy the kernel page directory as basis for user space directory
	memcpy( usrContext->pageDir, g_kernelPageDir, PAGE_SIZE);
		
	return usrContext;	
}



/** Copies exefile data to the corresponding pages*/
bool Init_User_Context_Memory(struct User_Context* context, char *exeFileData, 
			ulong_t exeFileLength, struct Exe_Format *exeFormat)
{
	//ELF segment index
	int segIdx;
	//Maximal virtual address
	unsigned maxvaddr = 0;
		
	//Copy all (ELF) segments
	for(segIdx=0; segIdx < exeFormat->numSegments; segIdx++)
	{
		//Get ELF segment
		struct Exe_Segment* elfseg = &exeFormat->segmentList[segIdx];
		
		//Recalc max virt addr
		if(maxvaddr < elfseg->startAddress +  elfseg->sizeInMemory)
			maxvaddr = elfseg->startAddress +  elfseg->sizeInMemory;
	}
	
	
	//Get temporal mem for loading segments	
	KASSERT(maxvaddr > 0);
	void* tmp = Malloc( maxvaddr );	
	if(tmp == 0)
	{
		#ifdef DEBUG_USERVM
		Print("Malloc has not enough space for exe data.\n");
		#endif
		return false;
	}
	
	
	//Copy all (ELF) segments
	for(segIdx=0; segIdx < exeFormat->numSegments; segIdx++)
	{
		//Get ELF segment
		struct Exe_Segment* elfseg = &exeFormat->segmentList[segIdx];
		//Copy corresponding segment in mem
		memcpy( tmp + elfseg->startAddress, exeFileData + elfseg->offsetInFile, elfseg->lengthInFile);
	}
	
	
		
	//Start addr of user space
	unsigned userOffset;
	//Copy page per page the whole memory
	for(userOffset = 0;  userOffset < maxvaddr; userOffset += PAGE_SIZE)
	{
		//Copy the page
		bool ret = Copy_Page(context->pageDir, (void*)(userOffset+USER_MEM_BASE), 
							(void*)(tmp + userOffset), PAGE_SIZE);
			
		//No mem for page
		if( ret == false)
		{
			//Free mem for copying pages
			Free(tmp);
			return false;
		}
	} 
	
	//Free mem for copying pages
	Free(tmp);	
	return true;		
}


/**
 * Determine directory and table entry by virtDest address. Create table if necessary. 
 * Copy size-many bytes from src to virtDest.
 */
bool Copy_Page(pde_t* dir, void* virtDest, void* src, unsigned size)
{
	#ifdef DEBUG_USERVM
	Print("Copy page to virtDest %p\n", virtDest);
	#endif
	
	//Only copy a whole page, not more
	KASSERT(size <= PAGE_SIZE);
	//Prove that virtDest is aligned on page base addresses
	KASSERT( IS_PAGE_ALIGNED_ADDR(virtDest) );
	//Only works for user space
	KASSERT( virtDest >= (void*) USER_MEM_BASE );
	
	
	//Indices for directory entry and table entry
	const unsigned direIdx = ((unsigned)virtDest >> PAGE_POWER) / NUM_PAGE_DIR_ENTRIES;
	const unsigned tabeIdx = ((unsigned)virtDest >> PAGE_POWER) % NUM_PAGE_DIR_ENTRIES;
	
	//Check boundaries
	KASSERT( NUM_PAGE_DIR_ENTRIES/2 <= direIdx && direIdx < NUM_PAGE_DIR_ENTRIES);
	KASSERT( 0<= tabeIdx && tabeIdx < NUM_PAGE_TABLE_ENTRIES);
	

	//Table is not present --> create a table	
	if( dir[direIdx].present == 0)
	{
		//Create a new page table.
		pte_t *pageTable = (pte_t*) Alloc_Page();		
		if( pageTable == 0 )
		{
			#ifdef DEBUG_USERVM
			Print("Could not alloc page for page table\n");
			#endif
			return false;
		}
			
		//mark all entries as absent
		memset(pageTable, 0, PAGE_SIZE);
		
		//The table is now present
		dir[direIdx].present = 1;
		//We are only interessted in the higher 20 bits
		dir[direIdx].pageTableBaseAddr = ((unsigned)pageTable) >> PAGE_POWER;
		//user table, writeable
		dir[direIdx].flags = VM_USER | VM_WRITE;	
	}
	
	
	//This is the page table we use
	KASSERT( dir[direIdx].present == 1);
	pte_t* pageTable = (pte_t*)(dir[direIdx].pageTableBaseAddr << PAGE_POWER);
	
	//Page we want to access is not present --> alloc a pageable page
	if( pageTable[tabeIdx].present == 0)
	{
		//Interrupts are not allowed to be enabled. It could be that
		//page is paged out between alloc and setting present bit to 1
    	//KASSERT(!Interrupts_Enabled());    
    	bool iflag = Begin_Int_Atomic();
		
		//Allocate a pageable page
		void* page = Alloc_Pageable_Page(  &pageTable[tabeIdx], (unsigned)virtDest );		
		if( page == 0)
		{
			#ifdef DEBUG_USERVM
			Print("Could not alloc page.\n");
			#endif
    		End_Int_Atomic(iflag);
			return false;
		}
			
		//Clear content of page
		memset(page, 0, PAGE_SIZE);
		
		//The table is now present
		pageTable[tabeIdx].present = 1;
		//We are only interessted in the higher 20 bits
		pageTable[tabeIdx].pageBaseAddr = ((unsigned)page) >> PAGE_POWER;
		//user table, writeable
		pageTable[tabeIdx].flags = VM_USER | VM_WRITE;
				
		//Set clock
		Get_Page((unsigned)page)->clock = 1;
		
		
    	End_Int_Atomic(iflag);		
	}
	
	
	
	//Now, we have oure page allocated, copy data
	KASSERT( pageTable[tabeIdx].present == 1);
	//Get phy. mem where page is located
	void* phyDest = (void*)(pageTable[tabeIdx].pageBaseAddr << PAGE_POWER);
	KASSERT(phyDest != 0);
	
	//Finally copy fucking data
	memcpy(phyDest, src, size);
	
	return true;	
}
