/*
 * Implementation of one philosopher for the phitable
 * which is our implementation of the dining philosopher problem
 */

#include <conio.h>
#include <string.h>
#include <process.h>
#include <fileio.h>
#include <sema.h> 

int main(int argc, char **argv)
{
	//We expect: MutexName, PhilosopherPipeName, Iterations, PhilosopherId, PhilosphLeftId, PhilosophRightId
	if ( argc != 8 )
	{
		Print("Please use:\n");
		Print("   %s <MutexName> <ReadPipeName> <WritPipeName> <Iterations> <PhilosopherId> <PhilosphLeftId> <PhilosophRightId>\n", argv[0]);
		Exit(1);
	}

	//Treat the CLI params
	const char *readPipeName  = argv[2];
	const char *writePipeName = argv[3];
	int ctIterations     = atoi(argv[4]);
	int philId           = atoi(argv[5]);
	int leftPhilId       = atoi(argv[6]);
	int rightPhilId      = atoi(argv[7]);

	//Connect to the mutex semaphore (with dummy initial value...)
	int mutex = Create_Semaphore(argv[1], -1);
	if ( mutex < 0 )
	{
		Print("Could not create semaphore for philosoph [%d] : err %d\n", philId, mutex);
		Exit(1);
	}
	
	//Create a (read) pipe for the communication with the phitable
	int readFd=0, writeFd=0, dummyFd=0;
	
	int pipeRet = Create_Named_Pipe(readPipeName, &readFd, &dummyFd);
	if ( pipeRet < 0 )
	{
		Print("Could not create pipe for philosoph %d : err %d\n", philId, pipeRet);
		Exit(1);
	}
	
	//Close writeFd
	Close(dummyFd);

	pipeRet = Create_Named_Pipe(writePipeName, &dummyFd, &writeFd);
	if ( pipeRet < 0 )
	{
		Print("Could not create pipe for philosoph %d : err %d\n", philId, pipeRet);
		Exit(1);
	}
	
	//Close readFd
	Close(dummyFd);

	Print("Philosopher %d started\n", philId);
	
	//Yeah, now loop N iterations and request forks :D
	int idx;
	for ( idx=0; idx<ctIterations; idx++ )
	{
		//Yes, busy looping...
		bool done = false;
		
		do
		{
			//Try to send a fork acquire...

			//Send request for left and right forks!
			char acquire[32];
			memset(acquire, 0, sizeof(acquire));
			snprintf(acquire, sizeof(acquire), "Acquire %d %d ", leftPhilId, rightPhilId);

			//Lock semaphore before sending messages
			P(mutex);
			int writtenBytes = Write(writeFd, acquire, strlen(acquire));
			V(mutex); //Don't forget to unlock the mutex

			if ( writtenBytes != strlen(acquire) )
			{
				Print("Could not send acquire...\n");
				Exit(1);
			}

			//Try to read an answer
			bool read = false;
	
			do
			{
				//read answer to acquire
				char buffer[32];
				memset(buffer, 0, sizeof(buffer));

				//Lock semaphore before reading messages
				P(mutex);
				int readBytes = Read(readFd, buffer, sizeof(buffer));
				//Unlock semaphore
				V(mutex);

				if (  readBytes > 0 )
				{
					read = true;

					//Did we get 200 OK?
					if ( strncmp("200", buffer, 3) == 0 )
					{
						done = true;
					}
					else
					{
						Print("Philosopher %d cannot eat...\n", philId);
					}
				}
				else if ( readBytes < 0 )
				{
					Print("Error while reading...\n");
					Exit(1);
				}
			}
			while ( !read );
		}
		while ( !done );

		//Now we can eat!
		Print("Philosopher %d got a piece\n", philId);
		
		//Now we have to release the previously acquired forks

		//Here we don't get a return value!!
		
		//Send request for left and right forks!
		char release[32];
		memset(release, 0, sizeof(release));
		snprintf(release, sizeof(release), "Release %d %d ", leftPhilId, rightPhilId);

		//Lock semaphore before sending messages
		P(mutex);
		int writtenBytes = Write(writeFd, release, strlen(release));
		//Unlock semaphore
		V(mutex);
		
		if ( writtenBytes != strlen(release) )
		{
			Print("Could not send release for philosopher %d!!! Terrible error ;(\n", philId);
		}
		
		//Try to read an answer
		bool read = false;

		do
		{
			//read answer to acquire
			char buffer[32];
			memset(buffer, 0, sizeof(buffer));

			//Lock semaphore before reading messages
			P(mutex);
			int readBytes = Read(readFd, buffer, sizeof(buffer));
			//Unlock semaphore
			V(mutex);

			if (  readBytes > 0 )
			{
				read = true;

				//Did we get 200 OK?
				if ( strncmp("200", buffer, 3) != 0 )
				{
					Print("Philosopher %d did not release correctly...\n", philId);
				}
			}
			else if ( readBytes < 0 )
			{
				Print("Error while reading...\n");
				Exit(1);
			}
		}
		while ( !read );
	}

	//Close file descriptors	
	Close(readFd);
	Close(writeFd);

	//Release the mutex
	Destroy_Semaphore(mutex);
	
	Print("Philosopher %d finished\n", philId);
		
	return 0;
}
