/*
 * This is the startup file for the dining philosopher problem
 * it requires the number of philosophers to start and the number of
 * iterations to loop as command line arguments
 * 
 * 
 * The protocol works as follows:
 * 
 * Philosoph: Acquire <ID of Philosopher Left> <ID of Philosopher Right>
 * Phitable answers with: [200 OK|400 NOT OK]
 * Philosoph: Release <ID of Philosopher Left> <ID of Philosopher Right>
 * Phitable answers with: [200 OK|400 NOT OK]
 */

#include <conio.h>
#include <process.h>
#include <string.h>
#include <fileio.h>
#include <sema.h>

/*
 * For communicating with our philosophs
 * we need read/write descriptors for the pipe...
 * and of course a pid to wait for...
 */
enum State
{
	UNLOCKED,
	LOCKED
};

struct PhilCommunication
{
	int pid;
	int readFd;
	int writeFd;
	int mutex;
	int ctIterations;
	enum State state;
};

/*
 * Parse two Philosopher ids from a Acquire/Release request...
 * expected value in buffer (starting at startIdx)
 *   <id1> <id2>
 */
void ParsePhilosopherIds(char *buffer, int startIdx, int *leftPhilId, int *rightPhilId)
{
	//Left philosoph id starts at char 8
	//till the next space
	short idx = startIdx;
	short len = 0;
	char philId[4];

	do
	{
		len++;
	}
	while ( buffer[idx+len] != ' ' );
						
	memset(philId, 0, sizeof(philId));
	memcpy(philId, &buffer[idx], MIN(sizeof(philId)-1, len));
	(*leftPhilId) = atoi(philId);
	
	//Skip space
	idx += len+1;
	
	len = 0;
	do
	{
		len++;
	}
	while ( buffer[idx+len] != '\0' );
	
	memset(philId, 0, sizeof(philId));
	memcpy(philId, &buffer[idx], MIN(sizeof(philId)-1, len));
 	(*rightPhilId) = atoi(philId);
}

int main(int argc, char ** argv)
{
	//Get settings from command line
	//We get chars from the CLI and extract '0' from the character we get
	//and we will receive the int
	if ( argc != 3 )
	{
		Print("Please use:\n");
		Print("   %s <number of philosophers to create> <number of iterations>\n", argv[0]);
		Exit(1);
	}

	int ctPhilosophers = atoi(argv[1]);
	int ctIterations   = atoi(argv[2]);

	Print("Starting %d philosphers each doing %d iterations\n", ctPhilosophers, ctIterations);

	//Array with pipes we use to communicate with the philosophers
	struct PhilCommunication philosophers[ctPhilosophers];

	int philIdx;
	for ( philIdx=0; philIdx<ctPhilosophers; philIdx++ )
	{
		//We need two pipes!
		//One pipe to write from philosoph to phitable (philosoph writes, phitable reads)
		//And another pipe to read from phitable (philosoph reads, phitable writes)
		//So the names are from the view of a philosoph
		char readPipeName[32];
		memset(readPipeName, 0, sizeof(readPipeName));
		snprintf(readPipeName, sizeof(readPipeName), "PipeWriteToPhitable%d", philIdx);

		char writePipeName[32];
		memset(writePipeName, 0, sizeof(writePipeName));
		snprintf(writePipeName, sizeof(writePipeName), "PipeReadFromPhitable%d", philIdx);

		int dummyFd = 0;

		//Create a mutex semaphore, for communication with the philosophers
		char mutexName[32];
		memset(mutexName, 0, sizeof(mutexName));
		snprintf(mutexName, sizeof(mutexName), "MutexPhil%d", philIdx);
		
		philosophers[philIdx].mutex = Create_Semaphore(mutexName, 1);
		if ( philosophers[philIdx].mutex < 0 )
		{
			Print("Could not create mutex for Phil%d : err %d\n", philIdx, philosophers[philIdx].mutex);
			//We cannot simply exit here because of our childs
			continue;
		}
	
		//Create a (read) pipe for the communication with this philosoph
		int pipeRet = Create_Named_Pipe(readPipeName, &philosophers[philIdx].readFd, &dummyFd);
		if ( pipeRet < 0 )
		{
			Print("Could not create read pipe for Phil%d : err %d\n", philIdx, pipeRet);
			//We cannot simply exit here because of our childs
			continue;
		}
		
		//Close writeFd
		Close(dummyFd);

		pipeRet = Create_Named_Pipe(writePipeName, &dummyFd, &philosophers[philIdx].writeFd);
		if ( pipeRet < 0 )
		{
			Print("Could not create write pipe for Phil%d : err %d\n", philIdx, pipeRet);
			//We cannot simply exit here because of our childs
			continue;
		}
		
		//Close readFd
		Close(dummyFd);

		//We have to malloc a command for each philosoph
		//We pass: MutexName, ReadPipeName, WritePipeName, Iterations, PhilosopherId, PhilosphLeftId, PhilosophRightId 
		char command[256];
		memset(command, 0, sizeof(command));
	
		//Get ids (indexes) of philosophers left and right of the current one
		int philIdLeft  = (ctPhilosophers+philIdx-1)%ctPhilosophers;
		int philIdRight = (philIdx+1)%ctPhilosophers;

		snprintf(command, sizeof(command), "/c/phil.exe %s %s %s %d %d %d %d", mutexName, writePipeName, readPipeName, ctIterations, philIdx, philIdLeft, philIdRight);
		
		//Initialize the struct
		philosophers[philIdx].pid          = Spawn_Program("/c/phil.exe", command, 0 ,1);
		philosophers[philIdx].state        = UNLOCKED;
		philosophers[philIdx].ctIterations = 0;
		if ( philosophers[philIdx].pid <= 0 )
		{
			Print("Failed to spawn philosoph %d\n", philIdx);
		}
	}

	//Now communicate with the philosophers till all are finished...
	while ( 1 )
	{
		int ctPhilosophersFinished = 0;
		for ( philIdx=0; philIdx<ctPhilosophers; philIdx++ )
		{
			//this philosopher is done already
			if ( philosophers[philIdx].ctIterations == ctIterations )
			{
				ctPhilosophersFinished++;
				continue;
			}
			
			//Try to read from slot
			char buffer[32];
			memset(buffer, 0, sizeof(buffer));

			P(philosophers[philIdx].mutex);
			int readBytes = Read(philosophers[philIdx].readFd, buffer, sizeof(buffer));
			V(philosophers[philIdx].mutex);

			if ( readBytes < 0 )
			{
				Print("Error while reading from philosoph %d\n", philIdx);
			}
			if ( readBytes > 0 )
			{
				//We got some request. Try to parse it and answer...
				if ( strncmp("Acquire", buffer, 7) == 0 )
				{
					int leftPhilId=0;
					int rightPhilId=0;

					ParsePhilosopherIds(buffer, 8, &leftPhilId, &rightPhilId);
					
					//Try to acquire these forks:
					char reply[10];
					memset(reply, 0, sizeof(reply));

//Print("acquire [%d=%d] [%d=%d]\n", leftPhilId, philosophers[leftPhilId].state, rightPhilId, philosophers[rightPhilId].state);
					if ( philosophers[leftPhilId].state == UNLOCKED && philosophers[rightPhilId].state == UNLOCKED )
					{
						//They are free - Lock them
						philosophers[leftPhilId].state  = LOCKED;
						philosophers[rightPhilId].state = LOCKED;
						snprintf(reply, sizeof(reply), "200 OK");
					} 
					else
					{
						snprintf(reply, sizeof(reply), "400 NOT OK");
					}

					P(philosophers[philIdx].mutex);
					int writtenBytes = Write(philosophers[philIdx].writeFd, reply, strlen(reply));
					V(philosophers[philIdx].mutex);

					if ( writtenBytes != strlen(reply) )
					{
						Print("Could not send acquire reply to philosoph %d\n", philIdx);
					}
				}
				else if ( strncmp("Release", buffer, 7) == 0 )
				{
					int leftPhilId=0;
					int rightPhilId=0;

					ParsePhilosopherIds(buffer, 8, &leftPhilId, &rightPhilId);

					//Try to release these forks:
					char reply[10];
					memset(reply, 0, sizeof(reply));

//Print("release [%d=%d] [%d=%d]\n", leftPhilId, philosophers[leftPhilId].state, rightPhilId, philosophers[rightPhilId].state);
					if ( philosophers[leftPhilId].state == LOCKED && philosophers[rightPhilId].state == LOCKED )
					{
						//They are free - Lock them
						philosophers[leftPhilId].state  = UNLOCKED;
						philosophers[rightPhilId].state = UNLOCKED;
						snprintf(reply, sizeof(reply), "200 OK");

						//Increase iterations...
						philosophers[philIdx].ctIterations++;
					} 
					else
					{
						//Someone tried to release non locked slots...
						snprintf(reply, sizeof(reply), "400 NOT OK");
					}

					P(philosophers[philIdx].mutex);
					int writtenBytes = Write(philosophers[philIdx].writeFd, reply, strlen(reply));
					V(philosophers[philIdx].mutex);

					if ( writtenBytes != strlen(reply) )
					{
						Print("Could not send release reply to philosoph %d\n", philIdx);
					}
				}
			}
		}
		
		//If all philosophers iterated all iterations - we are done
		if ( ctPhilosophersFinished == ctPhilosophers )
		{
			break;
		}
	}
	
	//Close file descriptors
	for ( philIdx=0; philIdx<ctPhilosophers; philIdx++ )
	{
		Close(philosophers[philIdx].readFd);
		Close(philosophers[philIdx].writeFd);
		//Destroy semaphore
		Destroy_Semaphore(philosophers[philIdx].mutex);
		//Because all processes are finished - we wait for the termination
		//Just to make sure... 
		Wait(philosophers[philIdx].pid);
	}

	return 0;
}
